#include "ArchiveHelper.h"

#include <archive.h>
#include <archive_entry.h>
#include <fcntl.h>
#include <iostream>
#include <sys/stat.h>

#include "CSVDataImporter.h"
#include "DataImporter.h"
#include "Exception.h"

#include <QDataStream>
#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QFileInfo>
#include <QObject>
#include <QProcess>
#include <QString>
#include <QStringList>

QVariant ArchiveHelper::exportData(QWidget *parent, DatabaseContext *context, ExportCallStrategy *exportCallStrategy) {
	QStringList fileFormats;
	fileFormats << QObject::trUtf8("ZIP-Archive *.zip");
	fileFormats << QObject::trUtf8("CSV-Dateien *.csv");
	return exportData(parent, context, exportCallStrategy, fileFormats);
}

QVariant ArchiveHelper::exportData(QWidget *parent, DatabaseContext *context, ExportCallStrategy *exportCallStrategy, QStringList fileFormats) {
	QDir defaultLocation = context->getImpExpFileDialogLocation();

	QString formatString = "";
	for (int n = 0; n < fileFormats.size(); n++) {
		formatString += fileFormats.at(n);
		if (n < fileFormats.size() - 1) {
			formatString += ";;";
		}
	}

	QString fileName = QFileDialog::getSaveFileName(parent, QObject::trUtf8("Dateiname fuer Export waehlen"), QString(), formatString);

	context->log("Archive", "ArchiveHelper", 6, "Chose file " + fileName + " for export.");

	if (fileName.length() > 0) {
		QDateTime oldDateTime = QDateTime::currentDateTime();

		QFile *file = new QFile(fileName);
		QFileInfo fileInfo = QFileInfo(*file);
		context->setImpExpFileDialogLocation(fileInfo.absoluteDir());
		delete file;

		QString fileNameWithoutPath = QFileInfo(fileName).fileName();
		if (fileNameWithoutPath.endsWith(".csv")) {
			context->log("Archive", "ArchiveHelper", 4, "Exporting into csv file");
			exportCallStrategy->exportData(context, fileName);

			QDateTime newDateTime = QDateTime::currentDateTime();
			return QVariant(oldDateTime.secsTo(newDateTime));
		} else if (fileNameWithoutPath.endsWith(".zip")) {
			context->log("Archive", "ArchiveHelper", 4, "Exporting into zip file");

			QDir tempDir = ArchiveHelper::createArchiveTempDirectoryForFile(context, fileName);
			QString fileNameWithoutSuffix = fileNameWithoutPath.mid(0, fileNameWithoutPath.indexOf("."));
			fileNameWithoutSuffix = ArchiveHelper::convertFileNameForOutput(fileNameWithoutSuffix);
			QString tempFileName = tempDir.filePath(fileNameWithoutSuffix + ".csv");
			try {
				context->log("Archive", "ArchiveHelper", 4, "Exporting csv data and images into temporary directory");
				exportCallStrategy->exportData(context, tempFileName);
			} catch (Exception e) {
				QString longText = QObject::trUtf8("Fehler beim Exportieren in die Datei %1: \n\n").arg(fileName) + e.getLongText()
								 + QObject::trUtf8("\n\nDaher konnten die Daten NICHT exportiert werden."
									   "Bitte beheben Sie den Fehler und versuchen Sie es nochmals.");
				throw Exception(e.getEventLogCategory(), e.getEventLogLocation(), e.getEventLogLevel(), e.getShortText(), longText);
				return QVariant(0);
			}

			QDir outputDir = QFileInfo(fileName).dir();
			ArchiveHelper::writeArchive(context, tempDir, outputDir, fileNameWithoutSuffix, ArchiveHelper::ZIP, true);

			QDateTime newDateTime = QDateTime::currentDateTime();
			return QVariant(oldDateTime.secsTo(newDateTime));
	  } else {
		  QString shortText = QObject::trUtf8("Dateiformat nicht unterstuetzt");
		  QString longText = QObject::trUtf8("Es kann nur in die Dateiformate csv und zip exportiert werden.  Bitte waehlen Sie eine Datei "
											  "aus, die entweder die Endung .csv oder die Endung .zip hat.");
		  throw Exception("Archive", "ArchiveHelper", shortText, longText);
		  return QVariant(0);
	  }
	} else {
		return QVariant();
	}
}

QDir ArchiveHelper::createArchiveTempDirectoryForFile(DatabaseContext *context, QString fileName, QString prefix) {
	QFileInfo fileInfo = QFileInfo(fileName);
	// dir() cuts the last part of the path, if you pass a directory to it (e.g. bar/foo/ => bar/)
	QDir directory = fileInfo.dir();
	if (fileInfo.isDir()) {
		bool success = directory.cdUp();
		if (!success) {
			QString title = QObject::trUtf8("Konnte nicht auf Verzeichnis zugreifen.");
			QString text = QObject::trUtf8("Der Versuch auf das Basisverzeichnis von %1 zuzugreifen ist gescheitert. \n\n"
								  "Daher konnte die gewuenschte Aktion nicht ausgefuehrt werden.").arg(directory.absolutePath());
			throw Exception("Archive", "ArchiveHelper", title, text);
		}
	}
	return createArchiveTempDirectory(context, directory, prefix);
}

QDir ArchiveHelper::createArchiveTempDirectory(DatabaseContext *context, QDir baseDir, QString prefix) {
	QDateTime dateTime = QDateTime::currentDateTime();
	QString suffix = dateTime.toString("yyyyMMdd_hhmmss");
	QString tempDirName = prefix + suffix;
    bool success = baseDir.mkdir(tempDirName);
	context->log("Archive", "ArchiveHelper", 4, "Created temporary directory " + tempDirName + " in base directory " + baseDir.absolutePath());

	QDir tempDir = baseDir;
	success &= tempDir.cd(tempDirName);

	if (!success) {
		QString title = QObject::trUtf8("Konnte temporaeres Verzeichnis nicht schreiben");
		QString text = QObject::trUtf8("Der Versuch, im Verzeichnis %1 ein temporaeres Verzeichnis %2 anzulegen und darauf zuzugreifen ist gescheitert. \n\n"
							  "Daher konnte die gewuenschte Aktion nicht ausgefuehrt werden.").arg(baseDir.absolutePath()).arg(tempDirName);
		throw Exception("Archive", "ArchiveHelper", title, text);
	}

	return tempDir;
}

void ArchiveHelper::writeArchive(DatabaseContext *context, QDir archiveTempDir, QDir outputDir, QString outputFileNamePrefix,
							  ArchiveHelper::ARCHIVE_TYPE archiveType, bool deleteTempDirectory) {

	context->log("Archive", "ArchiveHelper", 7, "Starting to write archive ");
	context->log("Archive", "ArchiveHelper", 5, "==> archiveTempDir = " + archiveTempDir.absolutePath());
	context->log("Archive", "ArchiveHelper", 5, "==> outputDir = " + outputDir.absolutePath());
	context->log("Archive", "ArchiveHelper", 5, "==> outputFileNamePrefix = " + outputFileNamePrefix);
	context->log("Archive", "ArchiveHelper", 5, "==> deleteTempDirectory = " + (deleteTempDirectory ? QString("true") : QString("false")));

	QString outputFileName = getArchiveName(context, outputFileNamePrefix, archiveType);
	context->log("Archive", "ArchiveHelper", 5, "Computed outputFileName = " + outputFileName);
	deleteOldArchive(context, outputDir, outputFileName);

	QString outputFilePath = outputDir.filePath(outputFileName);
	std::string outputFilePathString = outputFilePath.toStdString();
	context->log("Archive", "ArchiveHelper", 5, "Computed outputFilePath (as std::string) = " + QString(outputFilePathString.c_str()));
	context->log("Archive", "ArchiveHelper", 3, "INFO: archive_failed constant is " + QString::number(ARCHIVE_FAILED));

    char buff[8192];
	int retCode;

    struct archive *a = prepareArchive(context, outputFilePathString, archiveType);

	context->log("Archive", "ArchiveHelper", 6, "Starting to process entries");
	QStringList inputFiles = archiveTempDir.entryList(QStringList(), QDir::Files);
	for (int n = 0; n < inputFiles.size(); n++) {
		QString inputFile = inputFiles.at(n);
		std::string inputFileAsString = inputFile.toStdString();
		QString absoluteInputFilePath = archiveTempDir.absoluteFilePath(inputFile);
		std::string absoluteInputFilePathAsString = absoluteInputFilePath.toStdString();

		context->log("Archive", "ArchiveHelper", 4, "Processing file (as std::string) " + QString(inputFileAsString.c_str()));
		context->log("Archive", "ArchiveHelper", 4, "==> absoluteInputFilePath (as std::string) is " + QString(absoluteInputFilePathAsString.c_str()));

		QFile *file = new QFile(absoluteInputFilePath);

		struct archive_entry *entry = prepareArchiveEntry(context, a, inputFileAsString, absoluteInputFilePathAsString);

		retCode = archive_write_header(a, entry);
		processReturnCode(context, "archive_write_header", retCode, a, 4, 8);

		if (file->open(QIODevice::ReadOnly)) {
			QDataStream stream(file);
			while (!stream.atEnd()) {
				int length = stream.readRawData((char*)&buff, sizeof(buff));
				context->log("Archive", "ArchiveHelper", 2, "Writing " + QString::number(length) + " bytes of data.");
				retCode = archive_write_data(a, buff, length);
				processReturnCode(context, "archive_write_data (wrote " + QString::number(retCode) + " bytes of data)", length != retCode, a, 2, 8);
			}
		} else {
			QString shortText = QObject::trUtf8("Kann Datei nicht öffnen.");
			QString longText = QObject::trUtf8("Die temporäre Datei %1 konnte leider nicht geöffnet werden.")
				.arg(absoluteInputFilePath);
			throw Exception(shortText, longText);
		}

		file->close();
		delete file;

		archive_entry_free(entry);
	}

    retCode = archive_write_close(a);
	processReturnCode(context, "archive_write_close", retCode, a, 4, 8);
	retCode = archive_write_finish(a);

	if (deleteTempDirectory) {
		deleteTempDir(context, archiveTempDir);
		context->log("Archive", "ArchiveHelper", 5, "Deleted temp directory " + archiveTempDir.absolutePath());
	}
}

bool ArchiveHelper::isArchive(QString filePathName) {
	QFileInfo fileInfo = QFileInfo(filePathName);
	QString fileName = fileInfo.fileName();
	return fileName.endsWith(".zip");
}

QString ArchiveHelper::prepareForArchiveImport(DatabaseContext *context, QDir archiveTempDir, QString filePathName) {
	QFileInfo fileInfo = QFileInfo(filePathName);
	QString fileName = fileInfo.fileName();

	QFileInfo archiveDirFileInfo = QFileInfo(archiveTempDir.absolutePath());
	// returns parent directory in case of directories
	QDir parentDir = archiveDirFileInfo.dir();
	context->log("Archive", "ArchiveHelper", 2, "Parent directory of temp dir is " + parentDir.absolutePath());

	QString newFilePath = archiveTempDir.filePath(fileName);
	context->log("Archive", "ArchiveHelper", 4, "Making a temporary copy of " + filePathName + " at " + newFilePath);

	uncompressArchive(context, archiveTempDir, filePathName);

	context->log("Archive", "ArchiveHelper", 4, "Starting to find csv import file in uncompressed temporary files.");
	bool found = false;
	QString foundFileName;
	QStringList fileNames = archiveTempDir.entryList(QDir::Files);
	for (int n = 0; n < fileNames.size(); n++) {
		QString fileName = fileNames.at(n);
		context->log("Archive", "ArchiveHelper", 1, "==> Inspecting file " + fileName);
		if (fileName.endsWith(".csv")) {
			if (found) {
				QString title = QObject::trUtf8("Mehrere csv-Dateien");
				QString text = QObject::trUtf8("Das zu importierende Archiv enthaelt mehrere csv-Dateien. "
											   "Das ist nicht erlaubt.");
				throw Exception("Archive", "ArchiveHelper", title, text);
			}
			foundFileName = fileName;
			found = true;
		}
	}

	if (!found) {
		QString title = QObject::trUtf8("Keine csv-Datei gefunden");
		QString text = QObject::trUtf8("Das zu importierende Archiv enthaelt anscheinend keine csv-Datei.");
		throw Exception("Archive", "ArchiveHelper", title, text);
	}

	QString foundFilePath = archiveTempDir.absoluteFilePath(foundFileName);

	context->log("Archive", "ArchiveHelper", 4, "csv file of extracted archive now has temporary path " + foundFilePath);

	std::cout << "Found file " << qPrintable(foundFileName) << std::endl;
	std::cout << "Found file path " << qPrintable(foundFilePath) << std::endl;

	return foundFilePath;
}

void ArchiveHelper::uncompressArchive(DatabaseContext *context, QDir archiveTempDir, QString filePathName) {
	int retCode;

	ArchiveHelper::ARCHIVE_TYPE archiveType = getArchiveType(context, filePathName);
	struct archive *a = archive_read_new();
	setArchiveMode(context, a, archiveType);

	std::string filePathNameString = filePathName.toStdString();
	retCode = archive_read_open_filename(a, filePathNameString.c_str(), 10240);
	processReturnCode(context, "archive_read_open_file", retCode, a, 4, 8);

	struct archive_entry *entry;
	while (true) {
		retCode = archive_read_next_header(a, &entry);
		if (retCode == ARCHIVE_EOF) {
			context->log("Archive", "ArchiveHelper", 7, "Finished reading headers");
			break;
		}
		processReturnCode(context, "archive_read_next_header", retCode, a, 4, 8);

		const char* entryFileNameChar = archive_entry_pathname(entry);
		QString entryFileName = QString(entryFileNameChar);
		QString entryFilePath = archiveTempDir.filePath(entryFileName);

		QFile *outFile = new QFile(entryFilePath);

		if (!outFile->open(QIODevice::WriteOnly)) {
			QString longMessage = QObject::trUtf8("Kann die Datei %1 nicht zum Schreiben oeffnen.").arg(entryFileName);
			throw Exception(QObject::trUtf8("Datei nicht schreibbar"), longMessage);
		}

		QDataStream *stream = new QDataStream(outFile);

		const void *buff;
		size_t size;
		off_t offset;
		while (true) {
			retCode = archive_read_data_block(a, &buff, &size, &offset);
			if (retCode == ARCHIVE_EOF) {
				context->log("Archive", "ArchiveHelper", 2, "No more data to read for entry "
																+ entryFilePath);
				break;
			}
			processReturnCode(context, "archive_read_data_block", retCode, a, 2, 8);
			context->log("Archive", "ArchiveHelper", 2, "Read " + QString::number(size)
												+ " bytes at offset " + QString::number(offset));

			int bytesWritten = stream->writeRawData((const char*)buff, size);
			context->log("Archive", "ArchiveHelper", 2, "Wrote " + QString::number(bytesWritten) + " bytes.");
		}

		outFile->close();
		delete outFile;
		delete stream;
	}

	retCode = archive_read_finish(a);
	context->log("Archive", "Archive_Helper", 4, "Finished uncompressing.");
}

void ArchiveHelper::deleteTempDir(DatabaseContext *context, QDir archiveTempDir) {
	bool success = true;
	QStringList fileNames = archiveTempDir.entryList(QDir::Files);
	for (int n = 0; n < fileNames.size(); n++) {
		QString fileName = fileNames.at(n);
		context->log("Archive", "ArchiveHelper", 2, "Deleting file " + fileName + " in directory "
													+ archiveTempDir.absolutePath());
		bool successHere = archiveTempDir.remove(fileName);
		if (!successHere) {
			context->log("Archive", "ArchiveHelper", 9, "Deleting file " + fileName + " in directory "
														+ archiveTempDir.absolutePath() + " failed.");
			success = false;
		}
	}

	QFileInfo archiveDirFileInfo = QFileInfo(archiveTempDir.absolutePath());
	QString archiveDirName = archiveTempDir.dirName();

	// returns parent directory in case of directories
	QDir parentDir = archiveDirFileInfo.dir();
	context->log("Archive", "ArchiveHelper", 4, "Deleting temporary directory " + archiveDirName + " at " + parentDir.absolutePath());
	bool successHere = parentDir.rmdir(archiveDirName);
	if (!successHere) {
		context->log("Archive", "ArchiveHelper", 9, "Deleting temporary directory " + archiveDirName + " at "
							+ parentDir.absolutePath() + " failed.");
		success = false;
	}

	if (!success) {
		QString title = QObject::trUtf8("Loeschen der temporaeren Dateien fehlgeschlagen.");
		QString text = QObject::trUtf8("Waehrend des Exports wurde das temporaere Verzeichnis %1 angelegt. "
									   "Das Loeschen desselben ist fehlgeschlagen.").arg(archiveTempDir.absolutePath());
		throw Exception("Archive", "ArchiveHelper", title, text);
	}
}

QString ArchiveHelper::getArchiveName(DatabaseContext *context, QString prefix, ArchiveHelper::ARCHIVE_TYPE archiveType) {
	if (archiveType == ArchiveHelper::ZIP) {
		return prefix + ".zip";
	} else {
		QString shortText = QObject::trUtf8("Interner Fehler");
		QString longText = QObject::trUtf8("Der Export in das Archivformat %1 wird nicht unterstuetzt.").arg(archiveType);
		throw Exception("Archive", "ArchiveHelper", shortText, longText);
	}
}

ArchiveHelper::ARCHIVE_TYPE ArchiveHelper::getArchiveType(DatabaseContext *context, QString fileName) {
	if (fileName.endsWith(".zip")) {
		return ArchiveHelper::ZIP;
	} else {
		QString shortText = QObject::trUtf8("Falsches Dateiformat");
		QString longText = QObject::trUtf8("Es koennen nur csv-Dateien mit der Endung .csv sowie ZIP-Dateien mit der Endung .zip importiert werden.  "
										   "Die Datei %1 hat somit keine unterstuetzte Endung.").arg(fileName);
		throw Exception("Archive", "ArchiveHelper", shortText, longText);
	}
}

void ArchiveHelper::setArchiveMode(DatabaseContext *context, struct archive *a, ArchiveHelper::ARCHIVE_TYPE archiveType) {
	if (archiveType == ArchiveHelper::ZIP) {
		int retCode = archive_read_support_format_zip(a);
		processReturnCode(context, "archive_read_support_format_zip", retCode, a, 4, 8);
	} else {
		QString shortText = QObject::trUtf8("Interner Fehler");
		QString longText = QObject::trUtf8("Der Import aus dem Archivformat %1 wird nicht unterstuetzt.").arg(archiveType);
		throw Exception("Archive", "ArchiveHelper", shortText, longText);
	}
}

void ArchiveHelper::deleteOldArchive(DatabaseContext *context, QDir dir, QString fileName) {
	if (dir.exists(fileName)) {
		context->log("Archive", "ArchiveHelper", 4, "Deleting old archive file " + fileName + " at " + dir.absolutePath());
		bool success = dir.remove(fileName);
		if (!success) {
			QString shortText = QObject::trUtf8("Fehler beim Ueberschreiben");
			QString longText = QObject::trUtf8("Das bereits existierende Archiv %1 konnte nicht ueberschrieben werden.");
			throw Exception("Archive", "ArchiveHelper", shortText, longText);
		} else {
			context->log("Archive", "ArchiveHelper", 5, "Deleting old archive " + fileName + " in directory " + dir.absolutePath());
		}
	} else {
		context->log("Archive", "ArchiveHelper", 4, "There is no old archive " + fileName + " to delete in directory " + dir.absolutePath());
	}
}

struct archive* ArchiveHelper::prepareArchive(DatabaseContext *context, std::string outputFilePath, ArchiveHelper::ARCHIVE_TYPE archiveType) {
	int retCode;

    struct archive *a = archive_write_new();
	processReturnCode(context, "archive_write_new", a, 4, 8);

    //retCode = archive_write_set_compression_none(a);
	//processReturnCode(context, "archive_write_set_compression", retCode, a, 4, 8);

	if (archiveType == ArchiveHelper::ZIP) {
		retCode = archive_write_set_format_zip(a);
		processReturnCode(context, "archive_write_set_format_zip", retCode, a, 4, 8);
	} else {
		QString shortText = QObject::trUtf8("Interner Fehler");
		QString longText = QObject::trUtf8("Der Export in das Archivformat %1 wird nicht unterstuetzt.").arg(archiveType);
		throw Exception("Archive", "ArchiveHelper", shortText, longText);
	}

    retCode = archive_write_open_filename(a, outputFilePath.c_str());
	processReturnCode(context, "archive_write_open_filename", retCode, a, 4, 8);

	return a;
}

struct archive_entry* ArchiveHelper::prepareArchiveEntry(DatabaseContext *context, struct archive *a, std::string inputFile, std::string absoluteInputFile) {
	struct archive *metaArchive = archive_read_disk_new();
	processReturnCode(context, "archive_read_disk_new", metaArchive, 4, 8);

	struct archive_entry *entry = archive_entry_new();
	archive_entry_set_pathname(entry, absoluteInputFile.c_str());
	processReturnCode(context, "archive_entry_set_pathname", metaArchive, 4, 8);

	int retCode = archive_read_disk_entry_from_file(metaArchive, entry, -1, NULL);
	processReturnCode(context, "archive_read_disk_entry_from_file", retCode, metaArchive, 4, 8);

	// store non-absolute path name in archive
	archive_entry_set_pathname(entry, inputFile.c_str());
	processReturnCode(context, "archive_entry_set_pathname", metaArchive, 4, 8);

	archive_read_finish(metaArchive);
	return entry;
}

void ArchiveHelper::processReturnCode(DatabaseContext *context, QString title, int code, struct archive *a, int okLevel, int failLevel) {
	int level = (code == 0 ? okLevel : failLevel);
	QString text = "Executed " + title + "; returned code is " + QString::number(code) + "; errCode = " + QString::number(archive_errno(a)) + "; errMsg = " + QString(archive_error_string(a));
	context->log("Archive", "ArchiveHelper", level, text);

	if (code != ARCHIVE_OK) {
		QString title = QObject::trUtf8("Fehler");
		QString text = QObject::trUtf8("Beim Umgang mit dem Archiv ist folgender Fehler aufgetreten: %1\n\n Der Rueckgabewert war %2, der Fehlercode %3.")
						 .arg(QString(archive_error_string(a))).arg(code).arg(QString::number(archive_errno(a)));
		throw Exception("Archive", "ArchiveHelper", title, text);
	}
}

void ArchiveHelper::processReturnCode(DatabaseContext *context, QString title, struct archive *a, int okLevel, int failLevel) {
	int errCode = archive_errno(a);
	int level = (errCode == 0 ? okLevel : failLevel);
	context->log("Archive", "ArchiveHelper", level, "Executed " + title + "; errCode = " + QString::number(errCode) + "; errMsg = " + archive_error_string(a));

	if (errCode != ARCHIVE_OK) {
		QString title = QObject::trUtf8("Fehler");
		QString text = QObject::trUtf8("Beim Umgang mit dem Archiv ist folgender Fehler aufgetreten: %1\n\n Der Fehlercode war %2.")
						 .arg(QString(archive_error_string(a))).arg(QString::number(archive_errno(a)));
		throw Exception("Archive", "ArchiveHelper", title, text);
	}
}

QString ArchiveHelper::convertFileNameForOutput(QString fileName) {
	QString s = fileName;
//	s.replace(QObject::trUtf8(" "), QObject::trUtf8(""));
	s.replace(QObject::trUtf8("ä"), QObject::trUtf8("ae"));
	s.replace(QObject::trUtf8("ö"), QObject::trUtf8("oe"));
	s.replace(QObject::trUtf8("ü"), QObject::trUtf8("ue"));
	s.replace(QObject::trUtf8("Ä"), QObject::trUtf8("Ae"));
	s.replace(QObject::trUtf8("Ö"), QObject::trUtf8("Oe"));
	s.replace(QObject::trUtf8("Ü"), QObject::trUtf8("Ue"));
	s.replace(QObject::trUtf8("ß"), QObject::trUtf8("ss"));
	return s;
}
