#ifndef ATTRIBUTEDATA_H
#define ATTRIBUTEDATA_H

#include "DataTuple.h"

#include <QVariant>

/** Instances of class AttributeData store the contents of ForeignKeyGraph
 *  nodes during import.  This can be
 *     - the value of some attribute ("value AttributeData")
 *     - a reference to some tuple on the database, identified
 *       by its primary key ("foreign key AttributeData")
 *     - a pointer to some DataTuple ("pointer AttributeData")
 *     - the fact that we have a null value or reference
 *  E.g. if we have paths a.b and a.c in an import file, then initially
 *  we have AttributeData instances with values for a.b and a.c.  Lateron,
 *  the "a" instance with values a.b and a.c will be resolved, and the
 *  value AttributeData instances for a.b and a.c will be replaced by a
 *  pointer or foreign key AttributeData instance.
 */
class AttributeData {

public:
	/** Constructs a null AttributeData instance.
     */
	AttributeData();

	AttributeData(AttributeData *attributeData);

	/** Constructs a value AttributeData instance with the given value.
     *  @param value some value
	 */
	AttributeData(QVariant value);

	/** Makes this AttributeData instance a foreign key AttributeData instance
     *  with the given foreign key.
     *  @param foreignKeyToDB the foreign key
     */
	void setForeignKeyToDB(QVariant foreignKeyToDB);

	/** Makes this AttributeData instance a pointer AttributeData instance
     *  with the given pointer to a DataTuple.
     *  @param foreignKeyToDataTuple pointer to a DataTuple
     */
	void setForeignKeyToDataTuple(DataTuple *foreignKeyToDataTuple);

	/** Makes this AttributeData instance a value AttributeData instance
     *  with the given value.
     *  @param value some value
	 */
	void setValue(QVariant value);

	/** Sets this AttributeData to be null.
	 */
	void setNull();

	/** Returns wether this AttributeData instance is a foreign key AttributeData
	 *  instance.
	 *  @return wether this AttributeData instance is a foreign key AttributeData
	 *  instance.
	 */
	bool isForeignKeyToDB();

	/** Returns wether this AttributeData instance is a pointer AttributeData
	 *  instance.
	 *  @return wether this AttributeData instance is a pointer AttributeData
	 *  instance.
	 */
	bool isForeignKeyToDataTuple();

	/** Returns wether this AttributeData instance is a value AttributeData
	 *  instance.
	 *  @return wether this AttributeData instance is a value AttributeData
	 *  instance.
	 */
	bool isValue();

	/** Returns wether this is a null AttributeData.
	 *  @return wether this is a null AttributeData.
	 */
	bool isNull();

	/** Returns the foreign key of some foreign key AttributeData instance.
	 *  If this is no foreign key AttributeData instance, the return value
     *  is undefined.
	 *  @return the foreign key of some foreign key AttributeData instance.
	 */
	QVariant getForeignKeyToDB();

	/** Returns the pointer of some pointer AttributeData instance.
	 *  If this is no pointer AttributeData instance, the return value
     *  is undefined.
	 *  @return the pointer of some foreign key AttributeData instance.
	 */
	DataTuple *getForeignKeyToDataTuple();

	/** Returns the value of some value AttributeData instance.
	 *  If this is no value AttributeData instance, the return value
     *  is undefined.
	 *  @return the value of some foreign key AttributeData instance.
	 */
	QVariant getValue();

	/** Returns wether this AttributeData instance is invalid.
	 *  @return wether this AttributeData instance is invalid.
	 */
//	bool isInvalid();

private:
	bool isAForeignKeyToDB;
	bool isAForeignKeyToDataTuple;
	bool isAValue;
	bool isANullValue;

	QVariant foreignKeyToDB;
	DataTuple *foreignKeyToDataTuple;
	QVariant value;
};
#endif
