#include "DataTuple.h"

#include <QStringList>

#include <iostream>

#include "AttributeData.h"

DataTuple::DataTuple(QString entityDBName, int lineNumber, int tupleId) {
	this->pathKeyToAttributeData = std::map<QString, AttributeData*>();
	this->originalAttributeData = std::map<QString, AttributeData*>();
	this->entityDBName = entityDBName;
	this->lineNumber = lineNumber;
	this->tupleId = tupleId;
	this->primaryKey = QVariant();
	this->pkMatchPassed = false;
	this->duplicateCheckPassed = false;
}

DataTuple::DataTuple(QString entityDBName, std::map<QString, QVariant> dbNameToValue, QVariant primaryKey) {
	this->entityDBName = entityDBName;
	this->lineNumber = 0;
	this->tupleId = 0;
	this->primaryKey = primaryKey;
	this->pathKeyToAttributeData = std::map<QString, AttributeData*>();
	this->originalAttributeData = std::map<QString, AttributeData*>();
	for (std::map<QString, QVariant>::const_iterator it = dbNameToValue.begin(); it != dbNameToValue.end(); it++) {
		setValue(it->first, it->second);
	}
	this->pkMatchPassed = false;
	this->duplicateCheckPassed = false;
}

int DataTuple::getLineNumber() {
	return lineNumber;
}

int DataTuple::getTupleId() {
	return tupleId;
}

QString DataTuple::getEntityDBName() {
	return entityDBName;
}

void DataTuple::setValue(QString pathKey, QVariant value) {
	AttributeData *attributeData = new AttributeData(value);
	pathKeyToAttributeData[pathKey] = attributeData;
}

bool DataTuple::hasAttributeData(QString pathKey) {
	return pathKeyToAttributeData.find(pathKey) != pathKeyToAttributeData.end();
}

AttributeData *DataTuple::getAttributeData(QString pathKey) {
	if (pathKeyToAttributeData.find(pathKey) != pathKeyToAttributeData.end()) {
		return pathKeyToAttributeData[pathKey];
	} else {
		return NULL;
	}
}

void DataTuple::saveAttributeData() {
	this->originalAttributeData = std::map<QString, AttributeData*>();
	for (std::map<QString, AttributeData*>::const_iterator it = pathKeyToAttributeData.begin(); it != pathKeyToAttributeData.end(); it++) {
		QString pathKey = it->first;
		AttributeData *oldAttributeData = it->second;
		AttributeData *newAttributeData = new AttributeData(oldAttributeData);
		this->originalAttributeData[pathKey] = newAttributeData;
	}
}

std::map<QString, QString> DataTuple::getOriginalPathKeyToData() {
	std::map<QString, QString> originalPathKeyToData = std::map<QString, QString>();
	for (std::map<QString, AttributeData*>::const_iterator it = originalAttributeData.begin(); it != originalAttributeData.end(); it++) {
		QString pathKey = it->first;
		AttributeData *attributeData = it->second;
		if (attributeData->isValue()) {
			originalPathKeyToData[pathKey] = attributeData->getValue().toString();
		}
	}
	return originalPathKeyToData;
}

void DataTuple::setForeignKeyToDB(QString pathKey, QVariant foreignKey) {
	AttributeData *attributeData = new AttributeData();
	attributeData->setForeignKeyToDB(foreignKey);
	pathKeyToAttributeData[pathKey] = attributeData;
}

void DataTuple::setForeignKeyToDataTuple(QString pathKey, DataTuple* dataTuple) {
	AttributeData *attributeData = new AttributeData();
	attributeData->setForeignKeyToDataTuple(dataTuple);
	pathKeyToAttributeData[pathKey] = attributeData;
}

void DataTuple::setNull(QString pathKey) {
	AttributeData *attributeData = new AttributeData();
	pathKeyToAttributeData[pathKey] = attributeData;
}

void DataTuple::deleteAttributeData(QString pathKey) {
	AttributeData *attributeData = pathKeyToAttributeData[pathKey];
	pathKeyToAttributeData.erase(pathKey);
	delete attributeData;
}

std::set<QString> DataTuple::getAllPathKeys() {
	std::set<QString> pathKeys = std::set<QString>();
	for (std::map<QString, AttributeData*>::const_iterator it = pathKeyToAttributeData.begin();
		it != pathKeyToAttributeData.end(); it++) {
		QString pathKey = it->first;
		pathKeys.insert(pathKey);
	}
	return pathKeys;
}

void DataTuple::setPrimaryKey(QVariant primaryKey) {
	this->primaryKey = primaryKey;
}

QVariant DataTuple::getPrimaryKey() {
	return primaryKey;
}

void DataTuple::setPKMatchPassed(bool pkMatchPassed) {
	this->pkMatchPassed = pkMatchPassed;
}

bool DataTuple::isPKMatchPassed() {
	return pkMatchPassed;
}

void DataTuple::setDuplicateCheckPassed(bool duplicateCheckPassed) {
	this->duplicateCheckPassed = duplicateCheckPassed;
}

bool DataTuple::isDuplicateCheckPassed() {
	return this->duplicateCheckPassed;
}

std::map<QString, QVariant> DataTuple::getAllAttributes() {
	std::map<QString, QVariant> attributeDBNameToValue = std::map<QString, QVariant>();
	  for (std::map<QString, AttributeData*>::const_iterator it = pathKeyToAttributeData.begin();
			  it != pathKeyToAttributeData.end(); it++) {
		  QString attributeDBName = it->first;
		  AttributeData* attributeData = it->second;
		  if (attributeData->isValue()) {
			  attributeDBNameToValue[attributeDBName] = attributeData->getValue();
		  } else if (attributeData->isForeignKeyToDB()) {
			  attributeDBNameToValue[attributeDBName] = attributeData->getForeignKeyToDB();
		  } else if (attributeData->isForeignKeyToDataTuple()) {
			  // do nothing, we don´t consider foreign keys to data tuples
			  // attributes in the sense of this function, as they are
			  // no persistent information.  During import, we can´t resolve
			  // them, before all instances have been added to the database.
		  }
	  }

	return attributeDBNameToValue;
}

void DataTuple::printData(DatabaseContext *context) {
	printData(context, 1);
}

void DataTuple::printData(DatabaseContext *context, int level) {
	QString thisAddress = QString::number((qlonglong)this, 16);
	context->log("ImportExport", "DataTuple.cpp", level, "DataTuple " + thisAddress + " for entity " + entityDBName + " containing");

	for (std::map<QString, AttributeData*>::const_iterator it = pathKeyToAttributeData.begin();
		it != pathKeyToAttributeData.end(); it++) {
		QString attributeDBName = it->first;
	    AttributeData *attributeData = it->second;
		if (attributeData->isForeignKeyToDB()) {
			context->log("ImportExport", "DataTuple.cpp", level, "  Foreign key with path " + attributeDBName +
													  + " to database-id " + attributeData->getForeignKeyToDB().toString());
		} else if (attributeData->isForeignKeyToDataTuple()) {
			QString ftAddress = QString::number((qlonglong)attributeData->getForeignKeyToDataTuple(), 16);
			context->log("ImportExport", "DataTuple.cpp", level, "  Foreign key with path " + attributeDBName +
													  + " to tuple " + ftAddress);
		} else if (attributeData->isValue()) {
			QString value = attributeData->getValue().toString();
			context->log("ImportExport", "DataTuple.cpp", level, "  Value with path " + attributeDBName + " and value " + value);
		} else if (attributeData->isNull()) {
			context->log("ImportExport", "DataTuple.cpp", level, "  NULL for path " + attributeDBName);
		}
	}
}
