#include "ForeignKeyNode.h"

#include "ForeignKeyEdge.h"
#include "ForeignKeyGraph.h"


ForeignKeyNode::ForeignKeyNode(const Entity *entity, QString pathKey) {
	this->entity = entity;
	this->pathKey = pathKey;
	this->attributeDBNameToAttribute = std::map<QString, const Attribute*>();
	this->attributeDBNameToChildEdge = std::map<QString, ForeignKeyEdge*>();

}

ForeignKeyNode::~ForeignKeyNode() {
	for (std::map<QString, ForeignKeyEdge*>::const_iterator it = attributeDBNameToChildEdge.begin();
		it != attributeDBNameToChildEdge.end(); it++) {
		delete it->second;
	}
}

bool ForeignKeyNode::hasAttribute(QString attributeDBName) {
	return attributeDBNameToAttribute.find(attributeDBName) != attributeDBNameToAttribute.end();
}

const Attribute* ForeignKeyNode::getAttribute(QString attributeDBName) {
	if (hasAttribute(attributeDBName)) {
		return attributeDBNameToAttribute[attributeDBName];
	} else {
		return NULL;
	}
}

void ForeignKeyNode::addAttribute(const Attribute *attribute) {
	attributeDBNameToAttribute[attribute->getDBName()] = attribute;
}

std::vector<const Attribute*> ForeignKeyNode::getAttributes() {
	std::vector<const Attribute*> attributes = std::vector<const Attribute*>();
	for (std::map<QString, const Attribute*>::const_iterator it = attributeDBNameToAttribute.begin(); 
		it != attributeDBNameToAttribute.end(); it++) {
		attributes.push_back(it->second);
	}
	return attributes;
}

bool ForeignKeyNode::hasChildNode(QString attributeDBName) {
	return attributeDBNameToChildEdge.find(attributeDBName) != attributeDBNameToChildEdge.end();
}

ForeignKeyNode *ForeignKeyNode::getChildNode(QString attributeDBName) {
	if (hasChildNode(attributeDBName)) {
		ForeignKeyEdge *edge = attributeDBNameToChildEdge[attributeDBName];
		return edge->getOutNode();
	} else {
		return NULL;
	}
}

bool ForeignKeyNode::hasDescendant(QString entityDBName) {
	bool found = false;
	for (std::map<QString, ForeignKeyEdge*>::const_iterator it = attributeDBNameToChildEdge.begin();
			  it != attributeDBNameToChildEdge.end(); it++) {
		ForeignKeyEdge *edge = it->second;
		ForeignKeyNode *outNode = edge->getOutNode();
		QString outNodeEntityDBName = outNode->getEntityDBName();
		if (outNode != NULL) {
			if (outNodeEntityDBName == entityDBName) {
				return true;
			} else {
				found |= outNode->hasDescendant(entityDBName);
			}
		}
	}
	return found;
}

void ForeignKeyNode::addChildEdge(ForeignKeyGraph *graph, ForeignKeyEdge *edge) {
	this->attributeDBNameToChildEdge[edge->getAttributeDBName()] = edge;
	graph->addNodeToGraph(edge->getOutNode());
}

std::vector<ForeignKeyEdge*> ForeignKeyNode::getChildEdges() {
	std::vector<ForeignKeyEdge*> childEdges = std::vector<ForeignKeyEdge*>();
	for (std::map<QString, ForeignKeyEdge*>::const_iterator it = attributeDBNameToChildEdge.begin();
		it != attributeDBNameToChildEdge.end(); it++) {
		ForeignKeyEdge *edge = it->second;
		childEdges.push_back(edge);
	}
	return childEdges;
}

const Entity *ForeignKeyNode::getEntity() {
	return entity;
}

QString ForeignKeyNode::getEntityDBName() {
	return entity->getDBName();
}

QString ForeignKeyNode::getPathKey() {
	return pathKey;
}

void ForeignKeyNode::addAttributePaths(std::vector<QStringList> &paths, QStringList prefix) {
	const std::vector<Attribute*> attributes = entity->getAllAttributesIncludingBaseClass();

	for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		Attribute *attribute = *it;
		QString attributeDBName = attribute->getDBName();

		if (attributeDBNameToChildEdge.find(attributeDBName) != attributeDBNameToChildEdge.end()) {
			QStringList newPrefix = prefix;

			// there is an edge for the attribute at hand - follow it, perform a recursive call
			ForeignKeyEdge *edge = attributeDBNameToChildEdge[attributeDBName];
			newPrefix << attributeDBName;
			ForeignKeyNode *outNode = edge->getOutNode();
			// call by value for prefix, thus we pass the same copy to all edges
			outNode->addAttributePaths(paths, newPrefix);
		} else if (attributeDBNameToAttribute.find(attributeDBName) != attributeDBNameToAttribute.end()) {
			QStringList newPrefix = prefix;
			// this is not an edge, but an attribute - finish the path and return
			newPrefix << attributeDBName;

			// path completed
			paths.push_back(newPrefix);
		}
	}
}

void ForeignKeyNode::printStructure(DatabaseContext *context) {
	context->log("ImportExport", "ForeignKeyNode.cpp", 4, "    Node for entity " + entity->getDBName() 
													+ " with pathKey " + pathKey + " contains:");
	for (std::map<QString, ForeignKeyEdge*>::const_iterator it = attributeDBNameToChildEdge.begin();
		it != attributeDBNameToChildEdge.end(); it++) {
		QString attributeDBName = it->first;
		ForeignKeyEdge *edge = it->second;
		QString outEntityDBName = edge->getOutNodeEntityDBName();
		QString edgePathKey = edge->getPathKey();
		context->log("ImportExport", "ForeignKeyNode.cpp", 4,
					  "      Edge with pathKey " + edgePathKey + " for attribute "
					  + attributeDBName + " to entity " + outEntityDBName);
	}

	for (std::map<QString, const Attribute*>::const_iterator it = attributeDBNameToAttribute.begin();
		  it != attributeDBNameToAttribute.end(); it++) {
		QString attributeDBName = it->first;
		context->log("ImportExport", "ForeignKeyNode.cpp", 4, "      Attribute with name " + attributeDBName);
	}
}
