#ifndef FOREIGNKEYNODE_H
#define FOREIGNKEYNODE_H

#include <QString>
#include <QStringList>
#include <QVariant>

#include <map>
#include <set>
#include <vector>

#include "Attribute.h"
#include "DatabaseContext.h"
#include "Entity.h"

class ForeignKeyEdge;
class ForeignKeyGraph;

/** Instances of this class represent nodes in the ForeignKeyGraph.  Nodes
 *  correspond to entities in the data model.  Usually, nodes exist for the
 *  entity to import/export, plus all entities that are needed because of
 *  Import/Export keys containing foreign references.
 *
 *  Generally, nodes contain information for all attributes of the
 *  import/export key of the entity.  Specifically, nodes contain two kinds
 *  of elements:
 *  - edges to nodes/entities reachable from a node through a foreign key
 *    reference
 *  - attributes that are part of the import/export key
 *  (TODO: correct?)
 */
class ForeignKeyNode {

public:
	/** Constructs a new ForeignKeyNode for the given entity with the given
     *  path key.  The path key is a unique identifier for the node in a
     *  ForeignKeyGraph.  TODO: correct?
     *  @param entity entity of the node
     *  @param pathKey the path key as described
	 */
	ForeignKeyNode(const Entity *entity, QString pathKey);

	/** Destructs the ForeignKeyNode.  Deletes all outgoing ForeignKeyEdges,
     *  but not their destination nodes.
     */
	~ForeignKeyNode();

	/** Returns wether this ForeignKeyNode stores an attribute with the given
     *  db name.
     *  @param the attribute db name
     *  @return wether this ForeignKeyNode stores an attribute with the given
     */
	bool hasAttribute(QString attributeDBName);

	/** Returns the attribute of this ForeignKeyNode with the given db name,
     *  or NULL if there is no such attribute.
	 *  @return the attribute of this ForeignKeyNode with the given db name
     *  as described
	 */
	const Attribute* getAttribute(QString attributeDBName);

	/** Adds the given Attribute to this ForeignKeyNode.
     *  @param attribute the attribute
	 */
	void addAttribute(const Attribute *attribute);

	/** Returns all attributes of this ForeignKeyNode
     *  @return all attributes of this ForeignKeyNode
	 */
	std::vector<const Attribute*> getAttributes();

	/** Returns wether this ForeignKeyNode has an outgoing edge for a foreign
	 *  key attribute with the given name to some other ForeignKeyNode.
	 *  @return wether this ForeignKeyNode has an outgoing edge for a foreign
	 *  key attribute with the given name to some other ForeignKeyNode.
	 */
	bool hasChildNode(QString attributeDBName);

	/** Returns the destination node of the outgoing edge for the attribute
     *  with the given db name
     *  @param attributeDBName db name of the attribute
     *  @return the destination node of the outgoing edge for the attribute
     *  with the given db name
	 */
	ForeignKeyNode *getChildNode(QString attributeDBName);

	// TODO: Design: Shouldn't the graph be referenced in nodes / edges, and
	// not given explicitely here?
	/** Adds the given edge to this node.  The destination node of that edge
     *  will be registered in the given graph.
	 *  @param graph the graph where the destination node will be registered
     *  @param edge the new edge
	 */
	void addChildEdge(ForeignKeyGraph *graph, ForeignKeyEdge *edge);

	/** Returns all outgoing edges of this node.
     *  @return all outgoing edges of this node.
	 */
	std::vector<ForeignKeyEdge*> getChildEdges();

	/** Returns wether this node has a node for the given entity as descendant.
	 *  May only be called on cycle-free graphs (as foreign key graphs are
	 *  supposed to be), otherwise there may be a infinite loop!
	 */
	bool hasDescendant(QString entityDBName);

	/** Returns the entity of this ForeignKeyNode
	 *  @return the entity of this ForeignKeyNode
	 */
	const Entity *getEntity();

	/** Returns the db name of the entity of this ForeignKeyNode
	 *  @return the db name of the entity of this ForeignKeyNode
	 */
	QString getEntityDBName();

	/** Returns the path key of this ForeignKeyNode.  It is a unique identifier
	 *  of this node in a ForeignKeyGraph (TODO: correct?)
	 *  @return the path key of this ForeignKeyNode
	 */
	QString getPathKey();

	/** Adds paths for all attributes in the graph starting at this
	 *  ForeignKeyNode to the vector of paths given as argument call
     *  by reference.  All paths starting at this node will show up.
	 *  Paths consist of the attribute db names corresponding to edges
	 *  in the graph starting at this node.  Those edges may not form
	 *  any cycle!
	 *
	 *  Paths end if an attribute is registered in a node, but not
	 *  with an outgoing edge (i.e. a foreign key being part of the
	 *  import/export key).  Thus, the last part of any path is the db
	 *  name of such an attribute.  The separator of different parts
	 *  of the path is a dot.
     *  @param paths vector of paths to be constructed
	 *  @param prefix the prefix of a path formed by the
	 *                depth-first-traversal so far
	 */
	void addAttributePaths(std::vector<QStringList> &paths, QStringList prefix);

	/** Prints information about all outgoing edges and attributes of
	 *  this ForeignKeyNode to standard output.
	 */
	void printStructure(DatabaseContext *context);


private:
	const Entity *entity;
	QString pathKey;
	std::map<QString, ForeignKeyEdge*> attributeDBNameToChildEdge;
	std::map<QString, const Attribute*> attributeDBNameToAttribute;

};
#endif
