#include "ImpExpSchema.h"

#include "Attribute.h"
#include "DataModel.h"
#include "DataSet.h"
#include "Entity.h"
#include "Exception.h"

ImpExpSchema::ImpExpSchema(const Entity *entity) {
	this->entity = entity;
	this->fkDBNameToFKGraph = std::map<QString, ForeignKeyGraph*>();
}


ImpExpSchema::~ImpExpSchema() {
	for (std::map<QString, ForeignKeyGraph*>::iterator it = fkDBNameToFKGraph.begin();
		  it != fkDBNameToFKGraph.end(); it++) {
		delete it->second;
	}
}

QString ImpExpSchema::getEntityDBName() {
	return entity->getDBName();
}

bool ImpExpSchema::isAbstractEntity() {
	return entity->isAbstract();
}

bool ImpExpSchema::hasForeignKeyGraph(QString attributeDBName) {
	return (fkDBNameToFKGraph.find(attributeDBName) != fkDBNameToFKGraph.end());
}

ForeignKeyGraph *ImpExpSchema::getForeignKeyGraph(QString attributeDBName) {
	return fkDBNameToFKGraph[attributeDBName];
}

std::map<QString, ForeignKeyGraph*> ImpExpSchema::getAttributeToGraphMap() {
	return fkDBNameToFKGraph;
}

void ImpExpSchema::addForeignKeyGraph(QString fkDBName, ForeignKeyGraph *graph) {
	fkDBNameToFKGraph[fkDBName] = graph;
}

const std::vector<Attribute*> ImpExpSchema::getAllAttributes() {
	return entity->getAllAttributesIncludingBaseClass();
}

std::vector<QStringList> ImpExpSchema::getAttributePaths(QString attributeDBName) {
	const Attribute* attribute = entity->getAttributeByDBName(attributeDBName);

	if (attribute->isForeignKey()) {
		if (fkDBNameToFKGraph.find(attributeDBName) != fkDBNameToFKGraph.end()) {
			ForeignKeyGraph *graph = fkDBNameToFKGraph[attributeDBName];
			return graph->getAttributePaths();
		} else {
			return std::vector<QStringList>();
			// TODO maybe error handling, in fact we expect the presence of 
			// a graph for any foreign key
		}
	} else {
		QStringList attributePath = QStringList();
		attributePath << attributeDBName;

		std::vector<QStringList> paths = std::vector<QStringList>();
		paths.push_back(attributePath);
		return paths;
	}
}

std::multimap<QString, QStringList> ImpExpSchema::getAttributePaths() {
	std::multimap<QString, QStringList> attributeDBNameToPath = std::multimap<QString, QStringList>();

	const std::vector<Attribute*> attributes = entity->getAllAttributesIncludingBaseClass();
	for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute* attribute = *it;
		QString attributeDBName = attribute->getDBName();
		if (attribute->isForeignKey()) {
			if (fkDBNameToFKGraph.find(attributeDBName) != fkDBNameToFKGraph.end()) {
				ForeignKeyGraph *graph = fkDBNameToFKGraph[attributeDBName];
				std::vector<QStringList> graphPaths = graph->getAttributePaths();
				for (std::vector<QStringList>::const_iterator it = graphPaths.begin(); it != graphPaths.end(); it++) {
					attributeDBNameToPath.insert(std::multimap<QString, QStringList>::value_type(attributeDBName, *it));
				}
			} else {
				// TODO maybe error handling, in fact we expect the presence of
				// a graph for any foreign key
			}
		} else {
			QStringList attributePath = QStringList();
			attributePath << attributeDBName;
			attributeDBNameToPath.insert(std::multimap<QString, QStringList>::value_type(attributeDBName, attributePath));
		}
	}

	return attributeDBNameToPath;
}

ImpExpSchema* ImpExpSchema::constructFromAttributePaths(DatabaseContext *context, QString rootEntityDBName,
														std::vector<QStringList> attributePaths) {
// TODO: Add code for ignoring unknown attributes we don't care about (e.g. after a schema change)
	const DataModel *dataModel = context->getDataModel();
	const Entity *rootEntity = dataModel->getEntityByDBName(rootEntityDBName);

	context->log("ImportExport", "ImpExpSchema.cpp", 2, "constructFromAttributePaths: Started for entity " + rootEntityDBName);

	if (rootEntity == NULL) {
		QString shortMessage = trUtf8("Die Entitaet ist nicht bekannt.");
		QString longMessage = trUtf8("Eine Entitaet (Tabelle) namens %1 gibt es im Datenmodell nicht.")
								.arg(rootEntityDBName);
		throw Exception("ImportExport", "ImpExpSchema", shortMessage, longMessage);
	} else {
		ImpExpSchema *schema = new ImpExpSchema(rootEntity);
		for (std::vector<QStringList>::const_iterator it = attributePaths.begin(); it != attributePaths.end(); it++) {
			const Entity *entity = rootEntity;
			ForeignKeyGraph *graph;
			ForeignKeyNode *node;
			// handle one attribute path
			QStringList attributePath = *it;
			QString pathKey = "";

			context->log("ImportExport", "ImpExpSchema.cpp", 1, "  Processing path ");
			for (int n = 0; n < attributePath.size(); n++) {
				// iterate through attribute path
				QString attributeDBName = attributePath[n];
				context->log("ImportExport", "ImpExpSchema.cpp", 1, "    Processing path attribute " + attributeDBName);
				QString entityDBName = entity->getDBName();
				const Attribute* attribute = entity->getAttributeByDBName(attributeDBName);
				if (attribute == NULL) {
					delete schema;
					// check wether current attribute is known, if not throw an Exception
					QString shortMessage = trUtf8("Attribut nicht vorhanden");
					QString longMessage = trUtf8("In der Tabelle %1 gibt es kein Attribut namens %2."
												 "Bitte pruefen Sie Ihre Spaltenueberschriften in der Schemazeile.")
												  .arg(entityDBName).arg(attributeDBName);
					throw Exception("ImportExport", "ImpExpSchema", shortMessage, longMessage);
				} else {
					if (n < attributePath.size() - 1 && !attribute->isForeignKey()) {
					// no non-foreign keys except at the end of the path
						delete schema;
						QString shortMessage = trUtf8("Fehlerhafte Verweiskette");
						QString longMessage = trUtf8("Eine Spaltenueberschrift der Form a.b.c kennzeichnet "
							"eine Kette von Verweisen auf andere Tabelle, um das Ziel "
							"eines Verweises eindeutig bestimmen zu koennen. "
							"Nur das letzte Glied einer solchen Kette darf ein Attribut "
							"bezeichnen, das keinen Verweis darstellt.  Diese Regel wurde hier verletzt."
							"Bitte ueberpruefen Sie das Glied %1 in der Schemazeile zur Tabelle %2.")
							.arg(attributeDBName).arg(entityDBName);
						throw Exception("ImportExport", "ImpExpSchema", shortMessage, longMessage);
					}
					if (n == attributePath.size() - 1 && attribute->isForeignKey()) {
					// no foreign key at the end of the path
						delete schema;
						QString shortMessage = trUtf8("Fehlerhafte Verweiskette");
						QString longMessage = trUtf8("Eine Spaltenueberschrift der Form a.b.c kennzeichnet "
							"eine Kette von Verweisen auf andere Tabelle, um das Ziel "
							"eines Verweises eindeutig bestimmen zu koennen. "
							"Das letzte Glied einer solchen Kette muss ein Attribut "
							"bezeichnen, das keinen Verweis darstellt.  Diese Regel wurde hier verletzt."
							"Bitte ueberpruefen Sie das Glied %1 in der Schemazeile zur Tabelle %2.")
							.arg(attributeDBName).arg(entityDBName);
						throw Exception("ImportExport", "ImpExpSchema", shortMessage, longMessage);
					}
					if (n == 0) {
						// the very first attribute in the path, we first have to fetch/create the corresponding graph
						if (schema->hasForeignKeyGraph(attributeDBName)) {
							graph = schema->getForeignKeyGraph(attributeDBName);
						} else {
							graph = new ForeignKeyGraph(entity);
							schema->addForeignKeyGraph(attributeDBName, graph);
						}
						node = graph->getRootNode();
						entity = node->getEntity();
					}
					if (n == 0) {
						pathKey += attributeDBName;
					} else {
						pathKey += getPathSeparator() + attributeDBName;
					}
					// Situation: Given node and attribute, we have to fetch/add an 
					// (foreign key or not) attribute.
					if (attribute->isForeignKey()) {
						// foreign key attribute: We have to go one step down the graph
						// create new node if necessary (i.e. this path prefix wasn't visited yet)
						if (false)  {
//						if (node->hasChildNode(attributeDBName)) {
							node = node->getChildNode(attributeDBName);
							entity = node->getEntity();
						} else {
							const Entity* outEntity = attribute->getForeignKeyEntity();
							if (outEntity == NULL) {
								throw Exception("ImportExport", "ImpExpSchema", trUtf8("Interner Fehler"), trUtf8("FK-Attribute %1 without Entity")
																				  .arg(attributeDBName));
							} else {
								ForeignKeyNode *outNode = new ForeignKeyNode(outEntity, pathKey);
								ForeignKeyEdge *edge = new ForeignKeyEdge(node, attribute, outNode, pathKey);
								node->addChildEdge(graph, edge);
								node = outNode;
								entity = node->getEntity();
							}
						}
					} else {
						if (node->hasAttribute(attributeDBName)) {
							delete schema;
							QString shortMessage = trUtf8("Attribut doppelt angegeben");
							QString longMessage = trUtf8("In der Schemazeile zur Tabelle %1 gibt es einen doppelten "
														 "Eintrag.  Bitte ueberpruefen Sie diese Zeile.").arg(entityDBName);
							throw Exception("ImportExport", "ImpExpSchema", shortMessage, longMessage);
						} else {
							node->addAttribute(attribute);
						}
					}
				}
			}

		}
		return schema;
	}
}

QString ImpExpSchema::getPathSeparator() {
	return ".";
}

void ImpExpSchema::printStructure(DatabaseContext *context) {
	context->log("ImportExport", "ImpExpSchema.cpp", 4, "ImpExpSchema for entity " + entity->getDBName() + " containing");
	for (std::map<QString, ForeignKeyGraph*>::iterator it = fkDBNameToFKGraph.begin();
		  it != fkDBNameToFKGraph.end(); it++) {
		ForeignKeyGraph* graph = it->second;
		graph->printStructure(context);
	}
}




std::map<QString, ImpExpSchema*> ImpExpSchema::constructImpExpSchemas(DatabaseContext *context, DataSet *dataSet) {
	std::map<QString, ImpExpSchema*> entityDBNameToSchema = std::map<QString, ImpExpSchema*>();

	// Register all names of entitiys, where either explicit tuples exist in the dataset, or the information,
	// take all tuples of that type, was recorded.
	std::set<QString> entityDBNames = dataSet->getAllEntities();
	std::set<QString> entityDBNamesWithAllTuples = dataSet->getEntityDBNamesWithAllTuples();
	for (std::set<QString>::const_iterator it = entityDBNamesWithAllTuples.begin(); it != entityDBNamesWithAllTuples.end(); it++) {
		entityDBNames.insert(*it);
	}

	for (std::set<QString>::const_iterator it = entityDBNames.begin(); it != entityDBNames.end(); it++) {
		QString entityDBName = *it;
		const DataModel* dataModel = context->getDataModel();
		const Entity* entity = dataModel->getEntityByDBName(entityDBName);
		ImpExpSchema *schema = constructImpExpSchema(context, entity);
		entityDBNameToSchema[entityDBName] = schema;
	}

	return entityDBNameToSchema;
}

ImpExpSchema* ImpExpSchema::constructImpExpSchema(DatabaseContext *context, const Entity *entity) {
	QString entityDBName = entity->getDBName();
	context->log("ImportExport", "ImpExpSchema.cpp", 5, "Starting to construct ImpExpSchema for entity " + entityDBName);

	ImpExpSchema *schema = new ImpExpSchema(entity);

	const std::vector<Attribute*> attributes = entity->getAllAttributesIncludingBaseClass();
	for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute* attribute = *it;
		if (attribute->isForeignKey()) {
			QString attributeDBName = attribute->getDBName();
			ForeignKeyGraph *foreignKeyGraph = constructForeignKeyGraph(context, attribute);
			schema->addForeignKeyGraph(attributeDBName, foreignKeyGraph);
		} // TODO: ordinary attributes (seems to work, but I have forgotten why...)
	}

	context->log("ImportExport", "ImpExpSchema.cpp", 5, "Finished constructing ImpExpSchema for entity " + entityDBName);
	return schema;
}

ForeignKeyGraph* ImpExpSchema::constructForeignKeyGraph(DatabaseContext *context, const Attribute *attribute) {
	QString dbName = attribute->getDBName();
	context->log("ImportExport", "ImpExpSchema.cpp", 5, "Starting to construct ForeignKeyGraph for attribute " + dbName);

	ForeignKeyGraph *graph = new ForeignKeyGraph(attribute);
	ForeignKeyNode *rootNode = graph->getRootNode();
	addForeignKeyLevel(context, graph, rootNode, attribute, attribute->getDBName());

	context->log("ImportExport", "ImpExpSchema.cpp", 5, "Finished constructing ForeignKeyGraph for attribute " + dbName);
	return graph;
}

void ImpExpSchema::addForeignKeyLevel(DatabaseContext *context, ForeignKeyGraph *graph, ForeignKeyNode *inNode,
									  const Attribute *attribute, QString pathKey) {

	context->log("ImportExport", "ImpExpSchema.cpp", 4, "addForeignKeyLevel called for pathKey "
				+ pathKey + " and attribute " + attribute->getDBName());

	const Entity *outEntity = attribute->getForeignKeyEntity();
	QString outEntityDBName = outEntity->getDBName();
	ForeignKeyNode *outNode;
/*
	if (graph->hasNodeForEntity(outEntityDBName)) {
		outNode = graph->getNodeForEntity(outEntityDBName);
		if (outNode->hasDescendant(inNode->getEntityDBName())) {
			// no cycles are allowed in the graph, as they might result in infinite loops

			QString shortText = trUtf8("Illegaler Import/Export-Schluessel");
			QString longText = trUtf8("Der beim Export aufgebaute Fremdschluesselgraph enthält einen "
									  "Zyklus, an dem die Entität %1 beteiligt ist. \n\n"
									  "Dies ist ein interner Fehler, der auf einen Fehler beim Entwurf "
									  "des Datenmodells hindeutet.  Bitte kontaktieren Sie Ihren "
									  "Administrator.").arg(outEntityDBName);
			throw Exception("ImportExport", "ImpExpSchema", shortText, longText);
		}
	} else {*/
		outNode = new ForeignKeyNode(outEntity, pathKey);
//	}

	ForeignKeyEdge *edge = new ForeignKeyEdge(inNode, attribute, outNode, pathKey);
	inNode->addChildEdge(graph, edge);

	context->log("ImportExport", "ImpExpSchema.cpp", 4, "addForeignKeyLevel adds edge for pathKey "
				+ pathKey + " to out node " + outEntity->getDBName());

	const std::vector<Attribute*> attributes = outEntity->getImpExpKey();
	for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute* outAttribute = *it;
		if (outAttribute->isForeignKey()) {
			QString newPathKey = pathKey + ImpExpSchema::getPathSeparator() + outAttribute->getDBName();
			addForeignKeyLevel(context, graph, outNode, outAttribute, newPathKey);
		} else {
			outNode->addAttribute(outAttribute);
		}
	}
}

QString ImpExpSchema::getDBType(DatabaseContext *context, QString entityDBName, QString pathKey) {
	const DataModel *dataModel = context->getDataModel();
	const Entity *entity = dataModel->getEntityByDBName(entityDBName);

	if (entity == NULL) {
		return QString();
	}

	QStringList parts = pathKey.split(".");
	for (int n = 0; n < parts.size() - 1; n++) {
		QString currPart = parts.at(n);
		const Attribute *fkAttribute = entity->getAttributeByDBName(currPart);
		if (fkAttribute == NULL || !fkAttribute->isForeignKey()) {
			return QString();
		} else {
			entity = fkAttribute->getForeignKeyEntity();
			if (entity == NULL) {
				return QString();
			}
	   }
	}

	if (parts.size() <= 0) {
		return QString();
	} else {
		QString lastPart = parts.at(parts.size() - 1);
		const Attribute *lastAttribute = entity->getAttributeByDBName(lastPart);
		if (lastAttribute == NULL || lastAttribute->isForeignKey()) {
			return QString();
		} else {
			return lastAttribute->getDBType();
		}
	}
}
