#ifndef IMPEXPSCHEMA_H
#define IMPEXPSCHEMA_H

#include <QObject>
#include <QString>
#include <QStringList>

#include <map>
#include <vector>

#include "DatabaseContext.h"
#include "Entity.h"
#include "ForeignKeyGraph.h"

class DataSet;

/** Instances of this class store schema information for import and export
 *  of entity instances.  They store one ForeignKeyGraph per foreign key
 *  attribute, or, if appropriate as during import, per attribute at all.
 *  See ... (TODO add) for a detailed explanation of how import and export
 *  works.
 */
class ImpExpSchema : public QObject {
	Q_OBJECT


public:
	/** Constructs the ImpExpSchema instance.
     *  @param entity the entity the import/export schema is for
	 */
	ImpExpSchema(const Entity *entity);

	/** Destructs the ImpExpSchema instance.  Deletes all ForeignKeyGraph
     *  instances stored in this schema.
     */
	~ImpExpSchema();

	/** Returns the db name of the entity this ImpExpSchema is for.
	 *  @return the db name of the entity this ImpExpSchema is for.
	 */
	QString getEntityDBName();

	/** Returns wether this ImpExpSchema instance is for an abstract entity.
	 *  @return wether this ImpExpSchema instance is for an abstract entity.
	 */
	bool isAbstractEntity();

	/** Returns wether this ImpExpSchema instance has a ForeignKeyGraph
	 *  for the attribute with the given db name.
	 *  @param attributeDBName db name of an attribute of the entity of
	 *                         this ImpExpSchema
	 *  @return wether this ImpExpSchema instance has a ForeignKeyGraph
	 *  for the attribute with the given db name.
	 */
	bool hasForeignKeyGraph(QString attributeDBName);

	/** Returns the ForeignKeyGraph of the given attribute.
	 *  @param attributeDBName db name of an attribute of the entity of
	 *                         this ImpExpSchema
	 *  @return the ForeignKeyGraph of the given attribute.
	 */
	ForeignKeyGraph *getForeignKeyGraph(QString attributeDBName);

	/** Returns a map from attribute db name to ForeignKeyGraph containing
	 *  all ForeignKeyGraph stored in this ImpExpSchema.
	 *  @returna map from attribute db name to ForeignKeyGraph as described
	 */
	std::map<QString, ForeignKeyGraph*> getAttributeToGraphMap();

	/** Adds a ForeignKeyGraph to the ImpExpSchema and maps it to the given
     *  attribute.
	 *  @param fkDBName attribute db name of an attribute of the entity of
	 *                  this ImpExpSchema
	 *  @param graph ForeignKeyGraph for the given attribute
	 */
	void addForeignKeyGraph(QString fkDBName, ForeignKeyGraph *graph);

	/** Returns all Attributes for which this ImpExpSchema contains a
	 *  ForeignKeyGraph.
	 *  @return all Attributes for which this ImpExpSchema contains a
	 *  ForeignKeyGraph.
	 */
	const std::vector<Attribute*> getAllAttributes();

	/** Returns the list of attribute paths for the given attribute.  If the
	 *  attribute is a foreign key attribute, then any path through the
	 *  corresponding ForeignKeyGraph will be returned (if there is no such
	 *  graph, an empty list will be returned).  Otherwise, one path consisting
	 *  of the attribute db name will be returned.
	 *  @param attributeDBName the attribute db name of an attribute of the
	 *                         entity of this ImpExpSchema
	 *  @return list of attribute paths as described
	 */
	std::vector<QStringList> getAttributePaths(QString attributeDBName);

	/** Returns all attribute paths within this ImpExpSchema.
	 *  One attribute path is a sequence of foreign key references,
	 *  followed by an attribute that is no foreign key attribute.
	 *  E.g. type.name means from CodeValue to CodeType using the 
	 *  foreign key type, followed by the attribute name of entity 
	 *  CodeType.
	 *  @return a multimap from attribute db name to at least one
	 *          attribute path.  The strings in the QStringList are
	 *          the attribute names along the path, e.g.
	 *          [type, [type, name]] is a possible mapping for the
	 *          attribute CodeValue.type
	 */
	std::multimap<QString, QStringList> getAttributePaths();

	/** Constructs an ImpExpSchema based on the db name of the
	 *  corresponding entity and a list of attribute paths.  Can be
	 *  used to construct an ImpExpSchema based on the schema
	 *  information in some import file.
	 *  @param context the context
	 *  @param entityDBName name of the corresponding entity
	 *  @param attributePaths vector of attribute paths; note
	 *                        getPathSeparator().
	 *  @return newly constructed ImpExpSchema as described
	 */
	static ImpExpSchema* constructFromAttributePaths(DatabaseContext *context, QString entityDBName, 
													 std::vector<QStringList> attributePaths);

	/** Returns the separator used for separating attribute db
	 *  names in an attribute paths.
	 *  @return the separator used for separating attribute db
	 *  names in an attribute paths.
	 */
    static QString getPathSeparator();

	/** Logs information about this ImpExpSchema and all its
	 *  ForeignKeyGraphs using the EventLogger.
	 *  @param context the context
	 */
	void printStructure(DatabaseContext *context);


	/** Constructs new import/export schemas for all entities which exist in the
	 *  given DataSet.
	 *  @param context the context
	 *  @param dataSet the DataSet
	 *  @return map from entity db name to ImpExpSchema, containing one entry
	 *          for each entity that exists in the given DataSet
	 */
	static std::map<QString, ImpExpSchema*> constructImpExpSchemas(DatabaseContext *context, DataSet *dataSet);

	/** Constructs a new import/export schema for the given Entity.
	 *  The schema contains one foreign key graph per foreign key attribute.
	 *  @param context the context
	 *  @param entity the entity
	 *  @return the newly generated ImpExpSchema as described
	 */
	static ImpExpSchema* constructImpExpSchema(DatabaseContext *context, const Entity *entity);

	/** Constructs a new foreign key graph for the given foreign key attribute.
	 *  The foreign key graph contains paths along attributes of import/export 
	 *  keys.  A path ends as soon as an attribute on it is no
	 *  foreign key.
	 *  @param context the context
	 *  @param attribute any foreign key attribute
     *  @return the newly constructed foreign key graph as described
	 */
	static ForeignKeyGraph* constructForeignKeyGraph(DatabaseContext *context, const Attribute *attribute);

	/** Adds one foreign key level to the given foreign key graph.
	 *  This means, starting at the given node, a new foreign key edge for the
	 *  given foreign key attribute will be created and added to the node.
	 *  The entity of the destination node of that edge will be
	 *  the entity of the given foreign key attribute.
	 *  For each attribute of the import export key of the destination node entity,
	 *  the following will be done:
	 *  - if it is a foreign key attribute, the corresponding subtree will be 
	 *    added by recursively calling this function (loops will be detected and
	 *    resolved by referencing a already existent node instead of recursively
	 *    calling this function)
	 *  - if not, the attribute will simply be added to the destination node.
	 *
	 *  @param context the context
	 *  @param graph the foreign key graph we currently construct
	 *  @param node a foreign key node inside the graph
	 *  @param attribute a foreign key attribute of the entity corresponding 
	 *         to the given foreign key node
	 *  @param pathKey the path key representing the current position in a
	 *         path through the ForeignKeyGraph
	 */
	static void addForeignKeyLevel(DatabaseContext *context, ForeignKeyGraph *graph, ForeignKeyNode *node, 
							const Attribute *attribute, QString pathKey);

	static QString getDBType(DatabaseContext *context, QString entityDBName, QString pathKey);

private:
	const Entity *entity;
	std::map<QString, ForeignKeyGraph*> fkDBNameToFKGraph;
};
#endif
