#include "CSVDataExporter.h"

#include <QDataStream>
#include <QDate>
#include <QDir>
#include <QIODevice>
#include <QMetaType>
#include <QStringList>
#include <QTextCodec>
#include <QTextStream>

#include <vector>

#include "ArchiveHelper.h"
#include "Attribute.h"
#include "AttributeData.h"
#include "AttributePathController.h"
#include "ConstantValueFillConverter.h"
#include "DataModel.h"
#include "Entity.h"

#include "CSVFileData.h"
#include "DataExportColumn.h"
#include "ExportColumn.h"
#include "ExportConverter.h"
#include "ExportStrategy.h"
#include "GlobaleWidgetEigenschaften.h"
#include "PersistentExportStrategy.h"

void CSVDataExporter::writeData(DatabaseContext *context, QFile* file, ImpExpMetaInformation metaInformation,
								std::map<QString, ImpExpSchema*> entityDBNameToSchema, DataSet *dataSet, ExportParameters *exportParameters) {
	this->csvFormat = new CSVFormat();

	QTextStream* stream = new QTextStream(file);
	stream->setCodec("UTF-8");

	context->log("ImportExport", "CSVDataExporter", 7, "Starting to write data...");

	try {
		if (exportParameters == NULL || exportParameters->doExportMetaInformation()) {
			writeMetaInformation(context, stream, metaInformation, exportParameters);
		}

		for (std::map<QString, ImpExpSchema*>::const_iterator it = entityDBNameToSchema.begin();
			  it != entityDBNameToSchema.end(); it++) {
			ImpExpSchema *schema = it->second;
			QString entityDBName = schema->getEntityDBName();
			context->log("ImportExport", "CSVDataExporter", 6, "Writing data for entity " + entityDBName);

			ExportBlock* exportBlock = generateExportBlock(context, schema, dataSet, exportParameters, file);
			writeExportBlock(context, exportBlock, stream, exportParameters);

//			std::vector<ExportPath*> paths = writeSchemaInformation(context, stream, schema, exportParameters);

//			writeTuples(stream, file, context, entityDBName, paths, tuplesForSchema, exportParameters);
		}
		context->log("ImportExport", "CSVDataExporter", 7, "Finished writing data.");
} catch (Exception e) {
		delete stream;
		throw;
	}

	delete csvFormat;
}

void CSVDataExporter::writeMetaInformation(DatabaseContext *context, QTextStream *stream,
										   ImpExpMetaInformation metaInformation, ExportParameters *exportParameters) {
	context->log("ImportExport", "CSVDataExporter", 6, "Writing meta information");

	std::vector<QString> dataModelVersionLine = std::vector<QString>();
	if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
		dataModelVersionLine.push_back(csvFormat->getDataModelVersionDescriptor());
	}
	dataModelVersionLine.push_back(metaInformation.getDataModelVersion());
	context->log("ImportExport", "CSVDataExporter", 6, "Data model version: " + metaInformation.getDataModelVersion());
	writeLine(stream, dataModelVersionLine);

	std::vector<QString> generatorVersionLine = std::vector<QString>();
	if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
		generatorVersionLine.push_back(csvFormat->getGeneratorVersionDescriptor());
	}
    generatorVersionLine.push_back(metaInformation.getGeneratorVersion());
	context->log("ImportExport", "CSVDataExporter", 6, "Generator version: " + metaInformation.getGeneratorVersion());
	writeLine(stream, generatorVersionLine);

	std::vector<QString> timeStampLine = std::vector<QString>();
	if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
		timeStampLine.push_back(csvFormat->getWrittenAtDescriptor());
	}
	QString timeStampString = metaInformation.getWriteTimeStamp().toString("dd.MM.yyyy hh:mm:ss");
	context->log("ImportExport", "CSVDataExporter", 6, "Time stamp:" + timeStampString);
	timeStampLine.push_back(timeStampString);
	writeLine(stream, timeStampLine);

	context->log("ImportExport", "CSVDataExporter", 6, "Finished writing meta information.");
}

ExportBlock* CSVDataExporter::generateExportBlock(DatabaseContext *context, ImpExpSchema *schema, DataSet *dataSet, ExportParameters *exportParameters, QFile *file) {
	ExportBlock *exportBlock = new ExportBlock();
	std::vector<ExportPath*> paths = initializeExportColumns(exportBlock, context, schema, exportParameters);
	fetchData(exportBlock, context, schema, dataSet, paths, exportParameters, file);
	convertColumns(exportBlock, context, schema, exportParameters);
	return exportBlock;
}

std::vector<ExportPath*> CSVDataExporter::initializeExportColumns(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, ExportParameters *exportParameters) {
	QString entityDBName = schema->getEntityDBName();
	context->log("ImportExport", "CSVDataExporter", 5, "Starting to initialize export columns for entity " + entityDBName);

	bool hasExportPaths = false;
	std::vector<ExportPath*> exportPathsForEntity;
	// Check which export paths to export.
	// If such a vector is given: Use its paths instead of the paths given in the schema.  The resulting file will in general not be importable
	// If not: Standard export case - use paths given in the schema.  The resulting file will be importable.
	// The former case is for csv output for information purposes, the latter for data exchange.
	if (exportParameters != NULL) {
		std::map<QString, std::vector<ExportPath*> > exportPaths = exportParameters->getExportPaths();
		if (exportPaths.find(entityDBName) != exportPaths.end()) {
			exportPathsForEntity = exportPaths[entityDBName];
			hasExportPaths = true;
		}
	}

	std::vector<ExportPath*> paths = std::vector<ExportPath*>();
	if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
		context->log("ImportExport", "CSVDataExporter", 1, "Adding line specifiers.");
		ExportPath *path = new ExportPath(QStringList(), csvFormat->getSchemaDescriptor(), new ConstantValueFillConverter(csvFormat->getDataDescriptor()));
		ExportColumn *column = new DataExportColumn(csvFormat->getSchemaDescriptor(), path);
		exportBlock->addColumn(csvFormat->getSchemaDescriptor(), column);
		paths.push_back(path);
	}
	if (exportParameters == NULL || exportParameters->doExportEntityNames()) {
		context->log("ImportExport", "CSVDataExporter", 1, "Adding entity db names.");
		ExportPath *path = new ExportPath(QStringList(), entityDBName, new ConstantValueFillConverter(entityDBName));
		ExportColumn *column = new DataExportColumn(entityDBName, path);
		exportBlock->addColumn(entityDBName, column);
		paths.push_back(path);
	}

	ExportStrategy *exportStrategy = getExportStrategy(entityDBName);
	if (hasExportPaths) {
		for (std::vector<ExportPath*>::const_iterator it = exportPathsForEntity.begin(); it != exportPathsForEntity.end(); it++) {
			paths.push_back(*it);
		}

		for (std::vector<ExportPath*>::const_iterator it = paths.begin(); it != paths.end(); it++) {
			ExportPath* path = *it;
			QString caption = path->getCaption();
			QString pathString = path->getPathString();
			ExportColumn *column = new DataExportColumn(caption, path);
			exportBlock->addColumn(pathString, column);
		}
	} else {
		const std::vector<Attribute*> attributes = schema->getAllAttributes();
		for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
			Attribute *attribute = *it;
			QString attributeDBName = attribute->getDBName();

			if (exportStrategy->exportAttribute(entityDBName, attributeDBName)) {
				std::vector<QStringList> pathLists = schema->getAttributePaths(attributeDBName);
				for (std::vector<QStringList>::const_iterator it = pathLists.begin(); it != pathLists.end(); it++) {
					QStringList pathList = *it;
					QString pathString = "";
					for (int n = 0; n < pathList.size(); n++) {
						pathString += pathList.at(n) + (n < pathList.size() - 1 ? "." : "");
					}
					ExportPath *path = new ExportPath(pathList, pathString, NULL);
					ExportColumn *column = new DataExportColumn(pathString, path);
					exportBlock->addColumn(pathString, column);
					paths.push_back(path);
				}
			}
		}
	}

	context->log("ImportExport", "CSVDataExporter", 5, "Finished initializing export columns for entity " + entityDBName);
	return paths;
}

void CSVDataExporter::fetchData(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, DataSet *dataSet,
								std::vector<ExportPath*> paths, ExportParameters *exportParameters, QFile *file) {
	QString entityDBName = schema->getEntityDBName();
	context->log("ImportExport", "CSVDataExporter", 5, "Starting to fetch data for entity " + entityDBName);

	AttributePathController controller = AttributePathController();
	const DataModel *dataModel = context->getDataModel();
	const Entity *entity = dataModel->getEntityByDBName(entityDBName);

	std::vector<DataTuple*> tuples = dataSet->getTuples(entityDBName);
	std::vector<qlonglong> primaryKeys = std::vector<qlonglong>();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;
		primaryKeys.push_back(tuple->getPrimaryKey().toLongLong());
	}

	QDir dir = QFileInfo(*file).dir();
	for (std::vector<ExportPath*>::const_iterator it = paths.begin(); it != paths.end(); it++) {
		ExportPath* path = *it;

		QStringList pathList = path->getPath();
		if (pathList.isEmpty()) {
			// Empty path means that we don´t want to load data using AttributePathController for this column
			// Example: The line specifier column.
			continue;
		}

		QString pathString = path->getPathString();
		if (!exportBlock->hasColumn(pathString)) {
			throw Exception("ImportExport", "CSVDataExporter", trUtf8("Interner Fehler"), trUtf8("Erwartete Spalte nicht vorhanden: %1").arg(pathString));
		}

		ExportColumn *column = exportBlock->getColumn(pathString);
		DataExportColumn *dataColumn = (DataExportColumn*)column;
		if (pathList.size() == 1 && entity->getAttributeByDBName(pathList[0])->getDBType() == "longblob") {
			std::vector<qlonglong> blobPrimaryKeys = std::vector<qlonglong>();
			/* This code does not work, but might be adjusted to do so later.
			 * The problem is that if the addAll feature is used, no tuples will be explicitely constructed,
			 * but on the other hand side, writeBLOBs function below needs the tuples. 
			// If we need all tuples, overwrite the pk list generated above here.
			// (in this case it will only be used for getting the BLOBs)
			if (dataSet->isEntityWithAllTuples(entityDBName)) {
				const Attribute *pkAttribute = entity->getPrimaryKeyAttribute();
				QString pkDBName = pkAttribute->getDBName();
				QStringList pkPathList = QStringList();
				pkPathList << pkDBName;
				// Passing an empty pk vector means fetching all pks
				std::vector<QVariant> blobPrimaryKeyVariants = controller.getByPath(context, entityDBName, pkPathList, std::vector<qlonglong>());
				for (std::vector<QVariant>::const_iterator it2 = blobPrimaryKeyVariants.begin(); it2 != blobPrimaryKeyVariants.end(); it2++) {
					QVariant variant = *it2;
					blobPrimaryKeys.push_back(variant.toLongLong());
				}
			} else */ {
				for (std::vector<qlonglong>::const_iterator it2 = primaryKeys.begin(); it2 != primaryKeys.end(); it2++) {
					blobPrimaryKeys.push_back(*it2);
				}
			}

			QString attributeDBName = pathList[0];

			// For BLOBs, we do one query per data item, to avoid storing too much data in main memory at a time
			for (unsigned int z = 0; z < blobPrimaryKeys.size(); z++) {
				qlonglong currPK = blobPrimaryKeys[z];
				std::vector<qlonglong> currPKVector = std::vector<qlonglong>();
				currPKVector.push_back(currPK);

				std::vector<QVariant> values = controller.getByPath(context, entityDBName, pathList, currPKVector);
				QVariant v = values[0];
				QByteArray b = v.toByteArray();
				context->log("ImportExport", "CSVDataExporter", 1, "Loaded a vector of length " + QString::number(currPKVector.size())
																  + " with a BLOB value of size " + QString::number(b.size())
																  + " for attribute " + attributeDBName + " and primary key " + QString::number(currPK));

				DataTuple *currTuple = tuples[z];
				QVariant blobData = writeBLOB(dir, context, currTuple, attributeDBName, values[0]);
				QString valueAsString = convertForOutput(blobData);
				dataColumn->setValue(z, valueAsString);
			}
		} else {
			std::vector<qlonglong> searchPrimaryKeys = std::vector<qlonglong>();
			if (dataSet->isEntityWithAllTuples(entityDBName)) {
				searchPrimaryKeys = std::vector<qlonglong>();
			} else {
				for (std::vector<qlonglong>::const_iterator it2 = primaryKeys.begin(); it2 != primaryKeys.end(); it2++) {
					searchPrimaryKeys.push_back(*it2);
				}
			}

			std::vector<QVariant> values = controller.getByPath(context, entityDBName, pathList, searchPrimaryKeys);
			context->log("ImportExport", "CSVDataExporter", 1, "Found " + QString::number(values.size()) + " values");

			for (unsigned int n = 0; n < values.size(); n++) {
				if (values[n].isValid()) {
					QString valueAsString = convertForOutput(values[n]);
					dataColumn->setValue(n, valueAsString);
				} else {
					dataColumn->setValue(n, "");
				}
			}
		}
	}

	context->log("ImportExport", "CSVDataExporter", 5, "Finished fetching data for entity " + entityDBName);
}

void CSVDataExporter::convertColumns(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, ExportParameters *exportParameters) {
	QString entityDBName = schema->getEntityDBName();
	context->log("ImportExport", "CSVDataExporter", 5, "Starting to convert columns for entity " + entityDBName);

	exportBlock->processConverters(context);

	context->log("ImportExport", "CSVDataExporter", 5, "Finished column conversions for entity " + entityDBName);
}

void CSVDataExporter::writeExportBlock(DatabaseContext *context, ExportBlock *exportBlock, QTextStream *stream, ExportParameters *exportParameters) {
	exportBlock->setCaptionCoordinates(context, 0, 0);

	CSVFileData *csvFileData = new CSVFileData();
	exportBlock->calculateCaptions(context, csvFileData);
	int maxCaptionRow = exportBlock->getMaxCaptionRow();
	exportBlock->calculateContents(context, csvFileData, maxCaptionRow + 1);
	writeToFile(context, csvFileData, stream);
}

void CSVDataExporter::writeToFile(DatabaseContext *context, CSVFileData *csvFileData, QTextStream *stream) {
	context->log("ImportExport", "CSVDataExporter", 2, "Starting to write csv file data to file");

	std::map<int, std::map<int, QString>* > contentMap = csvFileData->getContentMap();
	for (std::map<int, std::map<int, QString>* >::const_iterator it = contentMap.begin(); it != contentMap.end(); it++) {
		std::vector<QString> line = std::vector<QString>();

		std::map<int, QString>* columnToData = it->second;

		context->log("ImportExport", "CSVDataExporter", 1, "Writing line " + QString::number(it->first) + " containing " + QString::number(columnToData->size()) + " tokens");

		int lastColumn = -1;
		for (std::map<int, QString>::const_iterator it2 = columnToData->begin(); it2 != columnToData->end(); it2++) {

			// Fill gaps (e.g. we have a string for column 1, and one for column 3, then we need to generate an empty string to fill column 2 here.)
			int currColumn = it2->first;
			for (int n = lastColumn + 1; n < currColumn; n++) {
				line.push_back("");
			}
			lastColumn = currColumn;

			line.push_back(it2->second);
		}

		writeLine(stream, line);
	}

	context->log("ImportExport", "CSVDataExporter", 2, "Finished writing csv file data to file");
}

std::vector<ExportPath*> CSVDataExporter::writeSchemaInformation(DatabaseContext *context, QTextStream *stream, ImpExpSchema *schema, ExportParameters *exportParameters) {
	std::vector<ExportPath*> retPaths = std::vector<ExportPath*>();
	QString entityDBName = schema->getEntityDBName();


	bool hasExportPaths = false;
	std::map<QString, std::vector<ExportPath*> > exportPaths;
	std::vector<ExportPath*> exportPathsForEntity;

	// Check which export paths to export.
	// If such a vector is given: Use its paths instead of the paths given in the schema.  The resulting file will in general not be importable
	// If not: Standard export case - use paths given in the schema.  The resulting file will be importable.
	// The former case is for csv output for information purposes, the latter for data exchange.
	if (exportParameters != NULL) {
		exportPaths = exportParameters->getExportPaths();
		if (exportPaths.find(entityDBName) != exportPaths.end()) {
			exportPathsForEntity = exportPaths[entityDBName];
			hasExportPaths = true;
		}
	}

	std::vector<QString> line = std::vector<QString>();

	QString schemaType = csvFormat->getSchemaDescriptor();
	if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
		line.push_back(schemaType);
	}
	if (exportParameters == NULL || exportParameters->doExportEntityNames()) {
		line.push_back(entityDBName);
	}

	ExportStrategy *exportStrategy = getExportStrategy(entityDBName);
	std::vector<ExportPath*> paths = std::vector<ExportPath*>();
	if (hasExportPaths) {
		paths = exportPathsForEntity;
		for (std::vector<ExportPath*>::const_iterator it = paths.begin(); it != paths.end(); it++) {
			ExportPath* path = *it;
			QStringList pathList = path->getPath();
			QString caption = path->getCaption();
			QString pathString = "";
			if (caption.isEmpty()) {
				for (int n = 0; n < pathList.size(); n++) {
					pathString += pathList.at(n) + (n < pathList.size() - 1 ? "." : "");
				}
			} else {
				pathString = caption;
			}
			line.push_back(pathString);
			retPaths.push_back(path);
		}
	} else {
		const std::vector<Attribute*> attributes = schema->getAllAttributes();
		for (std::vector<Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
			Attribute *attribute = *it;
			QString attributeDBName = attribute->getDBName();

			if (exportStrategy->exportAttribute(entityDBName, attributeDBName)) {
				std::vector<QStringList> paths = schema->getAttributePaths(attributeDBName);
				for (std::vector<QStringList>::const_iterator it = paths.begin(); it != paths.end(); it++) {
					QStringList path = *it;
					QString pathString = "";
					for (int n = 0; n < path.size(); n++) {
						pathString += path.at(n) + (n < path.size() - 1 ? "." : "");
					}
					line.push_back(pathString);
					retPaths.push_back(new ExportPath(path, pathString, NULL));
				}
			}
		}
	}

	if (exportParameters == NULL || exportParameters->doExportSchema()) {
		writeLine(stream, line);
	}
	context->log("ImportExport", "CSVDataExporter", 5, "Finished writing schema information for entity " + entityDBName);
	return retPaths;
}

void CSVDataExporter::writeTuples(QTextStream *stream, QFile *file, DatabaseContext *context, QString entityDBName,
								  std::vector<ExportPath*> paths, std::vector<DataTuple*> tuples, ExportParameters *exportParameters) {

	stream->setCodec("UTF-8");
	context->log("ImportExport", "CSVDataExporter.cpp", 7, "Starting to write tuples to file using codec "
													  + stream->codec()->name());

	AttributePathController controller = AttributePathController();
	const DataModel *dataModel = context->getDataModel();
	const Entity *entity = dataModel->getEntityByDBName(entityDBName);
	QDir dir = QFileInfo(*file).dir();

	std::vector<QStringList> lines = std::vector<QStringList>();
	std::vector<qlonglong> primaryKeys = std::vector<qlonglong>();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;
		lines.push_back(QStringList());
		primaryKeys.push_back(tuple->getPrimaryKey().toLongLong());
	}

	for (std::vector<ExportPath*>::const_iterator it = paths.begin(); it != paths.end(); it++) {
		ExportPath* path = *it;
		QStringList pathList = path->getPath();

		std::vector<QVariant> unconvertedValues = controller.getByPath(context, entityDBName, pathList, primaryKeys);
		std::vector<std::vector<QVariant> > values = std::vector<std::vector<QVariant> >();
		ExportConverter *converter = path->getConverter();
		for (std::vector<QVariant>::const_iterator it = unconvertedValues.begin(); it != unconvertedValues.end(); it++) {
			QVariant test = *it;
			if (converter != NULL) {
				std::vector<QVariant> convertedData = converter->convertData(context, *it); 
				values.push_back(convertedData);
			} else {
				std::vector<QVariant> unconvertedData = std::vector<QVariant>();
				unconvertedData.push_back(*it);
				values.push_back(unconvertedData);
			}
		}

		context->log("ImportExport", "CSVDataExporter", 1, "Found " + QString::number(values.size()) + " values");

		// exchange BLOBs by file names, and save files
		if (pathList.size() == 1) {
			QString attributeDBName = pathList[0];
			const Attribute *attribute = entity->getAttributeByDBName(attributeDBName);
			if (attribute->getDBType() == "longblob") {
				context->log("ImportExport", "CSVDataExporter", 1, "==> Processing a longblob value");
				int n = 0;
				for (std::vector<DataTuple*>::const_iterator it2 = tuples.begin(); it2 != tuples.end(); it2++) {
					DataTuple *tuple = *it2;
					QVariant blobData = writeBLOB(dir, context, tuple, attributeDBName);
					std::vector<QVariant> blobVector = std::vector<QVariant>();
					blobVector.push_back(blobData);
					values[n] = blobVector;
					n++;
				}
			}
		}

		for (unsigned int n = 0; n < values.size(); n++) {
			std::vector<QVariant> data = values[n];
			for (unsigned int z = 0; z < data.size(); z++) {
				lines[n].append(convertForOutput(data[z]));
			}
		}
	}

	for (std::vector<QStringList>::const_iterator it = lines.begin(); it != lines.end(); it++) {
		QStringList values = *it;
		std::vector<QString> line = std::vector<QString>();
		if (exportParameters == NULL || exportParameters->doExportLineSpecifiers()) {
			line.push_back(csvFormat->getDataDescriptor());
		}
		if (exportParameters == NULL || exportParameters->doExportEntityNames()) {
			line.push_back(entityDBName);
		}
		for (int n = 0; n < values.size(); n++) {
			line.push_back(values.at(n));
		}
		writeLine(stream, line);
	}

	context->log("ImportExport", "CSVDataExporter.cpp", 7, "Finished writing tuples to file.");
}

QString CSVDataExporter::convertForOutput(QVariant data) {
	if (data.type() == QVariant::Date) {
		QDate date = data.toDate();
		QString retValue = (date.isValid() ? date.toString(GlobaleWidgetEigenschaften::getDateFormat()) : "");
		return retValue;
	} else if (data.type() == QVariant::DateTime) {
		QDateTime dateTime = data.toDateTime();
		return (dateTime.isValid() ? dateTime.toString(GlobaleWidgetEigenschaften::getDateTimeFormat()) : "");
	} else if (data.type() == QVariant::Bool) {
		bool b = data.toBool();
		return (b ? "true" : "false");
	} else {
		return data.toString();
	}
}

QVariant CSVDataExporter::writeBLOB(QDir dir, DatabaseContext *context, DataTuple *tuple, QString attributeDBName) {
	QString entityDBName = tuple->getEntityDBName();
	AttributeData *attributeData = tuple->getAttributeData(attributeDBName);
	QVariant value = attributeData->getValue();
	return writeBLOB(dir, context, tuple, attributeDBName, value);
}

QVariant CSVDataExporter::writeBLOB(QDir dir, DatabaseContext *context, DataTuple *tuple, QString attributeDBName, QVariant blobData) {
	if (blobData.isValid()) {
		QByteArray byteArray = blobData.toByteArray();

		QString entityDBName = tuple->getEntityDBName();
		ExportStrategy *exportStrategy = getExportStrategy(entityDBName);
		QString blobFileName = exportStrategy->getFileNameForBLOB(tuple, attributeDBName);
		blobFileName = ArchiveHelper::convertFileNameForOutput(blobFileName);
		QString blobFilePath = dir.absoluteFilePath(blobFileName);

		context->log("ImportExport", "CSVDataExporter.cpp", 1, "Writing BLOB to file " + blobFilePath);

		QFile *blobFile = new QFile(blobFilePath);
		blobFile->open(QIODevice::WriteOnly);
		QDataStream* blobFileStream = new QDataStream(blobFile);
		blobFileStream->writeRawData(byteArray.data(), byteArray.size());
		delete blobFileStream;
		delete blobFile;

		return QVariant(blobFileName);
	} else {
		return QVariant();
	}
}

void CSVDataExporter::writeLine(QTextStream *stream, std::vector<QString> line) {
	QString separator = csvFormat->getDelimiter();
	int n = 0;
	for (std::vector<QString>::const_iterator it = line.begin(); it != line.end(); it++) {
		QString linePart = *it;
		// Replace " by \" in output file.  Otherwise, using the "token" notation for
		// tokens containing the separator ; would not work reliable.
		linePart.replace("\"", "\\\"");
		linePart.replace("\n", "");
		if (linePart.contains(separator)) {
			linePart = "\"" + linePart + "\"";
		}
		if (n > 0) {
			*stream << separator;
		}
		*stream << linePart;
		n++;
	}
	*stream << endl;
}
