#ifndef CSVDATAEXPORTER_H
#define CSVDATAEXPORTER_H

#include "DataExporter.h"

#include <QDir>
#include <QString>
#include <QStringList>
#include <QTextStream>

#include <set>
#include <vector>

#include "BlockExportColumn.h"
#include "CSVFileData.h"
#include "CSVFormat.h"
#include "DatabaseContext.h"
#include "DataExportColumn.h"
#include "DataSet.h"
#include "DataTuple.h"
#include "ExportBlock.h"
#include "ExportParameters.h"
#include "ExportPath.h"
#include "ImpExpMetaInformation.h"
#include "ImpExpSchema.h"

/** DataExporter implementation for exporting to csv files, according to
 *  the csv file format given here ...
 *  TODO add that file, include reference, also below
 */
class CSVDataExporter : public DataExporter {
	Q_OBJECT

protected:
	/** Writes the data in the given DataSet, plus the given schema to a
	 *  csv file.  Specifically, it iterates over all given schemas, and 
	 *  for each, writes schema information plus all tuples of the
	 *  corresponding type to the csv file.
	 *
	 *  Note that no check wether the given map actually contains schema
	 *  information for all tuples in the DataSet is performed.
	 *  Building these data structures correctly is the responsibility of
	 *  the superclass or some other code.  This class is merely
	 *  responsible for printing the given information to a csv file.
	 *
	 *  For details about the csv file format, see ...
	 */
	virtual void writeData(DatabaseContext *context, QFile* file, ImpExpMetaInformation metaInformation,
						   std::map<QString, ImpExpSchema*> entityDBNameToSchema, DataSet *dataSet, ExportParameters *exportParameters);
private:

	void writeMetaInformation(DatabaseContext *context, QTextStream *stream, ImpExpMetaInformation metaInformation, ExportParameters *exportParameters);

	ExportBlock* generateExportBlock(DatabaseContext *context, ImpExpSchema *schema, DataSet *dataSet, ExportParameters *exportParameters, QFile *file);
	std::vector<ExportPath*> initializeExportColumns(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, ExportParameters *exportParameters);
	void fetchData(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, DataSet *dataSet,
				   std::vector<ExportPath*> paths, ExportParameters *exportParameters, QFile *file);
	void convertColumns(ExportBlock *exportBlock, DatabaseContext *context, ImpExpSchema *schema, ExportParameters *exportParameters);

	void writeExportBlock(DatabaseContext *context, ExportBlock *exportBlock, QTextStream *stream, ExportParameters *exportParameters);
	void writeToFile(DatabaseContext *context, CSVFileData *csvFileData, QTextStream *stream);

	/** Writes the schema information for the given ImpExpSchema.  Returns the list of attribute
	 *  paths, i.e. the i-th index of the vector contains the attribute path for the i-th 
	 *  column according to the schema.
	 *  Example: If a CodeValue referencing some CodeType is to be exported, the list may
	 *  look like this: [typeId.name, name, kurzText, globalText]  typeId.name in this
	 *  example is an attribute path through edge/attribute typeId to CodeType.
	 *  @param stream stream for writing the csv file
	 *  @param schema any ImpExpSchema
	 *  @return list of attribute paths as described
	 */
	std::vector<ExportPath*> writeSchemaInformation(DatabaseContext *context, QTextStream *stream, ImpExpSchema *schema, ExportParameters *exportParameters);

	/** Writes all DataTuples in the given vector of DataTuples in 
	 *  csv format (see ...) to the given stream.  All tuples must
	 *  have the same type/entity.  The only information that is
	 *  used from the given DataTuples is the primary key.  All
	 *  other values to be exported are queried from the database
	 *  according to the given vector of attribute paths, regardless
	 *  of the contents of the DataType.  
	 *
	 *  The given vector of
	 *  attribute paths is used for querying the contents, e.g. for
	 *  [typeId.name], the field name of
	 *  all CodeTypes linked via typeId to the CodeValues with the
	 *  primary keys stored in the given DataTuples will be queried.
	 *  Note that one query per attribute path, not one per tuple
	 *  will be executed.
	 *
	 *  @param stream stream for writing the csv file
	 *  @param file the csv file
	 *  @param context the context
	 *  @param entityDBName all DataTuples to export in this call of
     *                      the method must be of this entity, given
	 *                      by its db name
	 *  @param tuples the DataTuples to export
	 */
	void writeTuples(QTextStream *stream, QFile *file, DatabaseContext *context, QString entityDBName,
					 std::vector<ExportPath*> paths, std::vector<DataTuple*> tuples, ExportParameters *exportParameters);

	QVariant writeBLOB(QDir dir, DatabaseContext *context, DataTuple *tuple, QString attributeDBName);
	QVariant writeBLOB(QDir dir, DatabaseContext *context, DataTuple *tuple, QString attributeDBName, QVariant blobData);

// TODO REMOVE
//	void writeTuple(QTextStream *stream, const Entity *entity, DataTuple *tuple);

	/** Writes the given line to the csv file, according to the file format specified 
	 *  here (TODO: Add).
	 *  @param stream the stream to write to
	 *  @param line the tokens of the csv line
	 */
	void writeLine(QTextStream *stream, std::vector<QString> line);

	QString convertForOutput(QVariant data);

	CSVFormat *csvFormat;
};
#endif
