#include "DataExporter.h"

#include <map>
#include <set>
#include <vector>

#include <QString>
#include <QVariant>

#include "DataHelper.h"
#include "DataModel.h"
#include "DataModelInfo.h"
#include "DataModelInfoController.h"
#include "ExportStrategy.h"
#include "PersistentExportStrategy.h"

DataExporter::DataExporter() {
	standardExportStrategy = new PersistentExportStrategy();
	entityDBNameToExportStrategy = std::map<QString, ExportStrategy*>();
	ExportStrategy *bildDatenExportStrategy = new BildDatenExportStrategy();
	entityDBNameToExportStrategy["BildDaten"] = bildDatenExportStrategy;
}

DataExporter::~DataExporter() {
	delete standardExportStrategy;
	for (std::map<QString, ExportStrategy*>::const_iterator it = entityDBNameToExportStrategy.begin();
		  it != entityDBNameToExportStrategy.end(); it++) {
		delete it->second;
	}
}

void DataExporter::exportData(DatabaseContext *context, QString fileName, DataSet *dataSet) {
	exportData(context, fileName, dataSet, NULL);
}

void DataExporter::exportData(DatabaseContext *context, QString fileName, DataSet *dataSet, ExportParameters *exportParameters) {
	QFile *file = new QFile(fileName);
	exportData(context, file, dataSet, exportParameters);
	delete file;
}

void DataExporter::exportData(DatabaseContext *context, QFile* file, DataSet *dataSet) {
	exportData(context, file, dataSet, NULL);
}

void DataExporter::exportData(DatabaseContext *context, QFile* file, DataSet *dataSet, ExportParameters *exportParameters) {
	if (!file->open(QIODevice::WriteOnly)) {
		QString longMessage = trUtf8("Die Datei %1 kann leider nicht geoeffnet bzw. angelegt werden.").arg(file->fileName());
		throw Exception("ImportExport", "DataExporter", trUtf8("Kann Datei nicht oeffnen"), longMessage);
	}

	try {
		context->log("ImportExport", "DataExporter", 7, "Starting to export data");
		std::map<QString, ImpExpSchema*> entityDBNameToSchema = ImpExpSchema::constructImpExpSchemas(context, dataSet);
		ImpExpMetaInformation metaInformation = constructMetaInformation(context);
		writeData(context, file, metaInformation, entityDBNameToSchema, dataSet, exportParameters);
		context->log("ImportExport", "DataExporter", 7, "Finished exporting data");
	} catch(Exception e) {
		delete file;
		throw;
	}
}

DataSet* DataExporter::constructDataSetForExport() {
	return new DataSet();
}

void DataExporter::addToDataSetForExport(DatabaseContext *context, DataSet *dataSet, Persistent *persistent) {
	// Determine the entityDBName of the given persistent
	const DataModel *dataModel = context->getDataModel();
	QString entityCodeName = persistent->returnName();
	const Entity *entity = dataModel->getEntityByCodename(entityCodeName);
	QString entityDBName = entity->getDBName();
	ExportStrategy *exportStrategy = getExportStrategy(entityDBName);

	QString idString = (persistent->hasId() ? QString::number(persistent->getId()) : "(none)");
	context->log("ImportExport", "DataExporter", 4, "Adding instance of entity " + entityDBName + " with id " + idString + " to DataSet");

	// then fetch the attribute map
	std::map<QString, QVariant> attributeDBNameToValue = persistent->exportToDBMap();

	// filter according to exportStrategy, e.g. don't export interal ids
	for (std::map<QString, QVariant>::iterator it = attributeDBNameToValue.begin();
		it != attributeDBNameToValue.end(); it++) {
		QString attributeDBName = it->first;
		if (!exportStrategy->exportAttribute(entityDBName, attributeDBName)) {
			attributeDBNameToValue.erase(attributeDBName);
		}
	}

	DataTuple *dataTuple = new DataTuple(entityDBName, attributeDBNameToValue, QVariant(persistent->getId()));
	dataSet->addDataTuple(entityDBName, dataTuple);
}

void DataExporter::addToDataSetForExport(DatabaseContext *context, DataSet *dataSet, 
										 std::vector<Persistent*> persistents) {
	for (std::vector<Persistent*>::const_iterator it = persistents.begin(); it != persistents.end(); it++) {
		addToDataSetForExport(context, dataSet, *it);
	}
}



ExportStrategy* DataExporter::getExportStrategy(QString entityDBName) {
	if (entityDBNameToExportStrategy.find(entityDBName) != entityDBNameToExportStrategy.end()) {
		return entityDBNameToExportStrategy[entityDBName];
	} else {
		return standardExportStrategy;
	}
}

ImpExpMetaInformation DataExporter::constructMetaInformation(DatabaseContext *context) {
	context->log("ImportExport", "DataExporter", 4, "Starting to construct meta information");

	DataModelInfoController controller = DataModelInfoController();
	std::vector<DataModelInfo*> dataModelInfos = controller.get(context, std::map<QString, QVariant>());
	QString shortText = trUtf8("Datenmodell-Information fehlerhaft");
	if (dataModelInfos.size() != 1) {
		QString longText = trUtf8("Die Datenmodell-Informationen (Version etc.) konnten nicht sauber geladen werden. \n\n"
								  "Bitte kontaktieren Sie den Administrator.");
		DataHelper::deleteVectorContents(dataModelInfos);
		throw Exception("ImportExport", "DataExporter", shortText, longText);
	} else {
		DataModelInfo* dataModelInfo = dataModelInfos[0];
		if (!dataModelInfo->hasVersion()) {
			throw Exception("ImportExport", "DataExporter", shortText, trUtf8("Die Datenmodellversion fehlt auf der Datenbank. "
											"Bitte kontaktieren Sie den Administrator."));
		} else if (!dataModelInfo->hasGeneratorVersion()) {
			throw Exception("ImportExport", "DataExporter", shortText, trUtf8("Die Generatorversion fehlt auf der Datenbank. "
											"Bitte kontaktieren Sie den Administrator."));
		}

		QString dataModelVersion = dataModelInfo->getVersion();
		QString generatorVersion = dataModelInfo->getGeneratorVersion();
		QDateTime exportTimeStamp = QDateTime::currentDateTime();

		DataHelper::deleteVectorContents(dataModelInfos);
		return ImpExpMetaInformation(dataModelVersion, generatorVersion, exportTimeStamp);
	}

	context->log("ImportExport", "DataExporter", 4, "Finished constructing meta information");
}
