#ifndef DATAEXPORTER_H
#define DATAEXPORTER_H

#include <map>
#include <vector>

#include <QFile>
#include <QObject>
#include <QString>

#include "Attribute.h"
#include "DatabaseContext.h"
#include "DataSet.h"
#include "Entity.h"
#include "Exception.h"
#include "ExportParameters.h"
#include "ForeignKeyEdge.h"
#include "ForeignKeyGraph.h"
#include "ForeignKeyNode.h"
#include "ImpExpMetaInformation.h"
#include "ImpExpSchema.h"
#include "Persistent.h"
#include "ExportStrategy.h"
#include "BildDatenExportStrategy.h"
#include "PersistentExportStrategy.h"

class DataExporter : public QObject {
	Q_OBJECT

public:

	DataExporter();
	~DataExporter();

	/** Exports all data given in the given DataSet to a file with the given
	 *  name.  If the file already exists, it will be overwritten.
	 *  (TODO: check this)
	 *  @param context the context
	 *  @param fileName name of the file to write to
	 *  @param dataSet any DataSet
	 */
	virtual void exportData(DatabaseContext *context, QString fileName, DataSet *dataSet);
	virtual void exportData(DatabaseContext *context, QString fileName, DataSet *dataSet, ExportParameters *exportParameters);

	/** Exports all data given in the given DataSet to the given file.
	 *  @param context the context
	 *  @param file the file
	 *  @param dataSet any DataSet
	 */
	virtual void exportData(DatabaseContext *context, QFile* file, DataSet *dataSet);
	virtual void exportData(DatabaseContext *context, QFile* file, DataSet *dataSet, ExportParameters *exportParameters);

	/** Constructs a new, empty DataSet for exporting data.  Data can be 
	 *  added afterwards e.g. using addToDataSetForExport.
	 *  @return the new DataSet
	 */
	DataSet* constructDataSetForExport();

	/** Adds a DataTuple with the data from the given persistent instance to
	 *  the DataSet.
	 *  @param context the context
	 *  @param dataSet any DataSet
	 *  @param persistent any Persistent
	 */
	void addToDataSetForExport(DatabaseContext *context, DataSet *dataSet, Persistent *persistent);

	/** Adds all given Persistents to the given DataSet.
	 *  @param context the context
	 *  @param dataSet any DataSet
	 *  @param peristents vector of Persistents
	 */
	void addToDataSetForExport(DatabaseContext *context, DataSet *dataSet, std::vector<Persistent*> persistents);
protected:

	// TODO: ImpExpSchemas inside DataSet - better?
	/** Writes all data from the given DataSet with the given schemas to the given file.
	 *  The details, e.g. which file format, are the responsibility of the implementing
	 *  subclass.
	 *  @param context the context
	 *  @param file the output file
	 *  @param entityDBNameToSchema map from entity db names to ImpExpSchemas
	 *  @param dataSet the DataSet to export to
     */
	virtual void writeData(DatabaseContext *context, QFile* file, ImpExpMetaInformation metaInformation,
						   std::map<QString, ImpExpSchema*> entityDBNameToSchema, DataSet *dataSet, ExportParameters *exportParameters) = 0;

    virtual ExportStrategy* getExportStrategy(QString entityDBName);

private:


	ExportStrategy *standardExportStrategy;
	std::map<QString, ExportStrategy*> entityDBNameToExportStrategy;

	ImpExpMetaInformation constructMetaInformation(DatabaseContext *context);
};
#endif
