#ifndef EXPORTHELPER_H
#define EXPORTHELPER_H

#include <vector>

#include <QString>

#include "DatabaseContext.h"
#include "DataExporter.h"
#include "DataSet.h"
#include "Mitglied.h"
#include "Verein.h"
#include "Wettbewerb.h"

/** Contains static methods for exporting certain subsets of the data,
 *  e.g. all VereinsMitglieder, using the more technical classes like
 *  DataExporter.  Thus, here the choice which instances to export,
 *  not the technical details are implemented.
 */
class ExportHelper {

public:
	static void exportiereVerein(DatabaseContext *databaseContext, Verein *verein, QString fileName);
	static void exportiereVereinsHierarchie(DatabaseContext *databaseContext, Verein *verein, QString fileName);

	/** Exports all Mitglieder of the given Verein to a csv file with
	 *  the given name.
	 *  @param databaseContext the context
	 *  @param verein any Verein
	 *  @param fileName name of the export file
	 */
	static void exportiereVereinsMitglieder(DatabaseContext *databaseContext, Verein *verein, QString fileName);
	static void exportiereVereinsHierarchieMitglieder(DatabaseContext *databaseContext, Verein *verein, QString fileName);

	/** Exports a complete Wettbewerb with all data attached to it.
	 *  In detail, beside the given Wettbewerb itself, all instances of the
	 *  following types that are attached to the given Wettbewerb will be
	 *  exported:
	 *  - JuriertBei
	 *  - WettbewerbsTeilnahme
	 *  - Sparte
	 *  - Bild
	 *  - BildDaten
	 *  - Jurierung
	 *  - Durchgang
	 *  - Wertung
	 *  - EinzelWertung
	 */
	static void exportWettbewerb(DatabaseContext *databaseContext, Wettbewerb *wettbewerb, QString fileName);

	/** Exportiert einen kompletten Wettbewerb plus die Daten aller Mitglieder
	 *  und Vereine, die daran teilnehmen.  Das bedeutet, neben den bei
	 *  exportWettbewerb genannten wettbewerbsspezifischen Daten werden alle
	 *  Instanzen der Typen
	 *  - Mitglied
	 *  - Verein
	 *  - ExternerJuror
	 *  die für den gegebenen Wettbewerb relevant sind exportiert.  Bei
	 *  Mitgliedern werden aus Datenschutzgründen nur die Felder
	 *  - verein
	 *  - mitgliedsArt
	 *  - mitgliedsNummerImVerein
	 *  - mitgliedsNummerKomplett
	 *  - jugend
	 *  - datumEintritt
	 *  - datumAustritt
	 *  - anrede
	 *  - titel
	 *  - vorname
	 *  - nachname
	 *  exportiert, alle anderen Felder werden auf Null gesetzt.
	 */
	static void exportWettbewerbMitMitgliedsDaten(DatabaseContext *databaseContext, 
												  Wettbewerb *wettbewerb, QString fileName);

	static std::vector<Persistent*> getPersistentsZuWettbewerb(DatabaseContext *databaseContext, Wettbewerb *wettbewerb);

	/** Filtert die Mitgliedsattribute.  Gibt eine neue Mitglieds-Instanz
	 *  zurück, die nur die gemäß exportWettbewerbMitMitgliedsDaten zulässigen
	 *  Attribute enthält.
	 *  @param mitglied beliebiges Mitglied
	 *  @return neue Mitglieds-Instanz mit nur genau den zulässigen Attributen
	 */
	static Mitglied* filtereMitgliedsAttribute(DatabaseContext *context, Mitglied *mitglied);

	static void exportiereAlleFachlichenDaten(DatabaseContext *context, QString fileName);
	static void exportiereAlleTechnischenDaten(DatabaseContext *context, QString fileName);
	static void exportiereDatenberechtigungen(DatabaseContext *context, QString fileName);
	static void exportiereAlleDaten(DatabaseContext *context, QString fileName);

	static void addAlleFachlichenDaten(DatabaseContext *context, DataSet *dataSet);
	static void addFachlicheDaten(DatabaseContext *context, std::vector<Persistent*> &persistents);
	static void addAlleTechnischenDaten(DatabaseContext *context, DataSet *dataSet);
	static void addTechnischeDaten(DatabaseContext *context, std::vector<Persistent*> &persistents);

	static void exportiere(DatabaseContext *context, DataSet *dataSet, QString fileName);
	static void exportiere(DatabaseContext *context, std::vector<Persistent*> persistents, QString fileName);

private:
	static void fuegeVereinsDatenZuDataSetHinzu(DatabaseContext *databaseContext, DataExporter *exporter, DataSet *dataSet, Verein *verein);
	static void fuegeVereinsMitgliederZuDataSetHinzu(DatabaseContext *databaseContext, DataExporter *exporter, DataSet *dataSet, Verein *verein);
};
#endif
