#include "EinzelWertungenExportCallStrategy.h"

#include "CSVDataExporter.h"
#include "DataExporter.h"
#include "DataSet.h"
#include "ExportHelper.h"

#include <set>
#include <vector>

#include <QAbstractListModel>

#include "EinzelWertungsListeJoinController.h"
#include "EinzelWertungsListeJoinModel.h"
#include "EinzelWertungsListeJoinSearchModel.h"
#include "EinzelWertungenConverter.h"
#include "ExportParameters.h"
#include "ExportPath.h"
#include "ExternerJuror.h"
#include "ExternerJurorController.h"
#include "Mitglied.h"
#include "MitgliedController.h"
#include "Persistent.h"
#include "Person.h"
#include "RangConverter.h"
#include "SortHelper.h"
#include "SummenWertungConverter.h"

EinzelWertungenExportCallStrategy::EinzelWertungenExportCallStrategy(Sparte *sparte, qlonglong jugendCodeId) {
	this->sparte = sparte;
	this->jugendCodeId = jugendCodeId;
}

void EinzelWertungenExportCallStrategy::exportData(DatabaseContext *context, QString fileName) {

	std::vector<Persistent*> persistents = std::vector<Persistent*>();

	CodeManager *codeManager = context->getCodeManager();
	qlonglong jugendJaCodeId = codeManager->getCodeValueByTypeAndName("Jugend", "Ja")->getId();
	qlonglong jugendNeinCodeId = codeManager->getCodeValueByTypeAndName("Jugend", "Nein")->getId();
	qlonglong jugendAlsErwCodeId = codeManager->getCodeValueByTypeAndName("DurchgangsModus", "JUGEND_ALS_ERWACHSENE")->getId();

	EinzelWertungsListeJoinSearchModel *searchModel = new EinzelWertungsListeJoinSearchModel();
	searchModel->setBildSparte(sparte->getId());
	searchModel->setTeilnehmerWertungJugend(jugendCodeId);

	if (jugendCodeId == jugendJaCodeId || (sparte->hasDurchgangsModus() && sparte->getDurchgangsModus() != jugendAlsErwCodeId)) {
		// Falls wir gerade die Erwachsenen verarbeiten, und der Durchgangsmodus Jugend als Erwachsene ist, dürfen wir den Filter
		// nicht setzen.  Grund: Genau dann werden sowohl Jugendliche als auch Erwachsene gebraucht.
		searchModel->setMitgliedJugend(jugendCodeId);
	}


	EinzelWertungsListeJoinController bilderSucheJoinController = EinzelWertungsListeJoinController();
	std::vector<EinzelWertungsListeJoinModel*> joinModels = bilderSucheJoinController.get(context, searchModel);

	std::map<qlonglong, Person*> juriertBeiIdToPerson = std::map<qlonglong, Person*>();
	MitgliedController mitgliedController = MitgliedController();
	ExternerJurorController externerJurorController = ExternerJurorController();

	std::vector<QAbstractListModel*> sortModels = std::vector<QAbstractListModel*>();
	for (unsigned int n = 0; n < joinModels.size(); n++) {
		sortModels.push_back(joinModels[n]);

		// Die Person-Instanzen zu den Juroren extrahieren / laden.
		if (joinModels[n]->hasJuriertBeiId()) {
			qlonglong juriertBeiId = joinModels[n]->getJuriertBeiId();
			if (juriertBeiIdToPerson.find(juriertBeiId) == juriertBeiIdToPerson.end()) {
				Person *person = NULL;
				if (joinModels[n]->hasJuriertBeiMitglied()) {
					person = mitgliedController.getByPrimaryKey(context, joinModels[n]->getJuriertBeiMitglied());
				} else if (joinModels[n]->hasJuriertBeiExternerJuror()) {
					person = externerJurorController.getByPrimaryKey(context, joinModels[n]->getJuriertBeiExternerJuror());
				}

				if (person != NULL) {
					juriertBeiIdToPerson[juriertBeiId] = person;
				}
			}
		}
	}

	SortHelper::sort(sortModels, EinzelWertungsListeJoinModel::MITGLIEDMITGLIEDSNUMMERKOMPLETT, SortHelper::STANDARD_STRING_SORT, Qt::AscendingOrder);
	SortHelper::sort(sortModels, EinzelWertungsListeJoinModel::TEILNEHMERWERTUNGPUNKTE, SortHelper::INT_SORT, Qt::DescendingOrder);

	std::vector<EinzelWertungsListeJoinModel*> neueSortierung = std::vector<EinzelWertungsListeJoinModel*>();

	context->log("ImportExport", "EinzelWertungenExportCallStrategy", 7, "Starting preparation loop for exporting " + QString::number(sortModels.size()) + " join models.");

	// Bei Gleichstand der Gesamtpunkte zweier Mitglieder werden die Bilder angesehen.
	int punkte = -1;
	std::vector<EinzelWertungsListeJoinModel*> currSortModels = std::vector<EinzelWertungsListeJoinModel*>();
	for (unsigned int n = 0; n < sortModels.size(); n++) {
		EinzelWertungsListeJoinModel *currSortModel = (EinzelWertungsListeJoinModel*)sortModels[n];
		int currPunkte = (currSortModel->hasTeilnehmerWertungPunkte() ? currSortModel->getTeilnehmerWertungPunkte() : -1);

		if (punkte == -1 || currPunkte == punkte) {
			punkte = currPunkte;
			currSortModels.push_back(currSortModel);
		}

		if (punkte != currPunkte || n == sortModels.size() - 1) {
			context->log("ImportExport", "EinzelWertungenExportCallStrategy", 7, "Preparing block of " + QString::number(currSortModels.size()) + " join models for " + QString::number(punkte) + " points for export.");

			// Alle SortModels für eine Punkteanzahl gefunden (wg. der Sortierung sind sie alle nacheinander), jetzt muss innerhalb der Punkteanzahl
			// korrekt sortiert werden.

			// Absteigend nach Bildbewertung sortieren
			SortHelper::sort(currSortModels, EinzelWertungsListeJoinModel::BILDBEWERTUNG, SortHelper::INT_SORT, Qt::DescendingOrder);

			std::set<qlonglong> registrierteBilder = std::set<qlonglong>();
			// Jedem Mitglied einen absteigend sortierten Vektor mit seinen Bildbewertungen zuordnen
			std::map<qlonglong, std::vector<int>* > mitgliedsIdToBildPunkte = std::map<qlonglong, std::vector<int>* >();
			for (unsigned int z = 0; z < currSortModels.size(); z++) {
				// Jedes Bild nur einmal in den Vektor aufnehmen, nicht einmal pro Einzelwertung
				qlonglong bildId = currSortModels[z]->getBildId();
				if (currSortModels[z]->hasMitgliedId() && currSortModels[z]->hasBildBewertung() && registrierteBilder.find(bildId) == registrierteBilder.end()) {
					qlonglong mitgliedsId = currSortModels[z]->getMitgliedId();
					int bildPunkte = currSortModels[z]->getBildBewertung();
					if (mitgliedsIdToBildPunkte.find(mitgliedsId) == mitgliedsIdToBildPunkte.end()) {
						mitgliedsIdToBildPunkte[mitgliedsId] = new std::vector<int>();
					}
					std::vector<int>* punkteVector = mitgliedsIdToBildPunkte[mitgliedsId];
					punkteVector->push_back(bildPunkte);
					registrierteBilder.insert(bildId);
				}
			}

			for (std::map<qlonglong, std::vector<int>* >::const_iterator it = mitgliedsIdToBildPunkte.begin(); it != mitgliedsIdToBildPunkte.end(); it++) {
				qlonglong mitgliedsId = it->first;
				std::vector<int> punkteVector = *(it->second);
				QString punkteString = "";
				for (int w = 0; w < punkteVector.size(); w++) {
					punkteString += QString::number(punkteVector[w]) + " ";
				}
				context->log("ImportExport", "EinzelWertungenExportCallStrategy", 5, "For Mitglied with id " + QString::number(mitgliedsId) + ": Constructed punkte vector " + punkteString);
			}

			std::vector<qlonglong> rangListe = std::vector<qlonglong>();
			unsigned int anzahlMitglieder = mitgliedsIdToBildPunkte.size();

			// Schleife für jeden Platz in der Rangliste (lokal für diesen Wert von TeilnehmerWertungPunkte) einmal ausführen
			for (unsigned int z = 0; z < anzahlMitglieder; z++) {
				context->log("ImportExport", "EinzelWertungenExportCallStrategy", 5, "Starting to calculate position " + QString::number(z));

				// Erstmal alle Mitglieder die noch nicht einem Platz zugewiesen wurden als Kandidatenmitglieder registrieren.
				// In der folgenden Schleife werden dann alle entfernt, die nicht am besten sind.
				std::set<qlonglong> kandidatenMitglieder = std::set<qlonglong>();
				for (std::map<qlonglong, std::vector<int>* >::const_iterator it = mitgliedsIdToBildPunkte.begin(); it != mitgliedsIdToBildPunkte.end(); it++) {
					kandidatenMitglieder.insert(it->first);
					context->log("ImportExport", "EinzelWertungenExportCallStrategy", 5, "Registering Mitglied " + QString::number(it->first) + " as candidate.");
				}

				// Schleife über maximal Mitglieder viele Bildbewertungen
				for (unsigned int v = 0; v < mitgliedsIdToBildPunkte.size(); v++) {
					int maxPunkte = -1;
					// Schleife über alle Kandidatenmitglieder, um die beste Bewertung unter den v-t besten Bildern der Mitglieder zu bekommen
					for (std::map<qlonglong, std::vector<int>* >::const_iterator it = mitgliedsIdToBildPunkte.begin(); it != mitgliedsIdToBildPunkte.end(); it++) {
						qlonglong mitgliedsId = it->first;
						std::vector<int> punkteVector = *(it->second);
						if (kandidatenMitglieder.find(mitgliedsId) != kandidatenMitglieder.end() && punkteVector.size() > v) {
							// Falls das Mitglied w mindestens v viele Bilder hat, prüfe ob die Bewertung die beste ist
							maxPunkte = (maxPunkte < punkteVector[v] ? punkteVector[v] : maxPunkte);
						}
					}

					// Alle Mitglieder herauswerfen, die eine Bewertung echt kleiner als maxPunkte für ihr v-t bestes Bild haben
					for (std::map<qlonglong, std::vector<int>* >::const_iterator it = mitgliedsIdToBildPunkte.begin(); it != mitgliedsIdToBildPunkte.end(); it++) {
						qlonglong mitgliedsId = it->first;
						std::vector<int> punkteVector = *(it->second);
						int punkte = (punkteVector.size() > v ? punkteVector[v] : 0);
						if (punkte < maxPunkte) {
							context->log("ImportExport", "EinzelWertungenExportCallStrategy", 5, "Removing candidate Mitglied " + QString::number(mitgliedsId) + " as "
												+ QString::number(punkte) + " < " + QString::number(maxPunkte) + " maximal points.");
							kandidatenMitglieder.erase(mitgliedsId);
						}
					}
				}

				// Jetzt wissen wir welches Mitglied bezogen auf den aktuellen Ranglistenplatz am besten ist
				// In dem (seltenen) Fall dass Mitglieder bei allen Bildern identisch sind, gibt es hier mehr als
				// ein Kandidatenmitglied.

				// Alle dazugehörigen JoinModels isolieren
				std::vector<EinzelWertungsListeJoinModel*> kandidatenJoinModels = std::vector<EinzelWertungsListeJoinModel*>();
				for (std::vector<EinzelWertungsListeJoinModel*>::const_iterator it = currSortModels.begin(); it != currSortModels.end(); it++) {
					EinzelWertungsListeJoinModel *currKandidatenJoinModel = *it;
					if (currKandidatenJoinModel->hasMitgliedId() && kandidatenMitglieder.find(currKandidatenJoinModel->getMitgliedId()) != kandidatenMitglieder.end()) {
						kandidatenJoinModels.push_back(currKandidatenJoinModel);
					}
				}

				SortHelper::sort(kandidatenJoinModels, EinzelWertungsListeJoinModel::BILDBEWERTUNG, SortHelper::INT_SORT, Qt::DescendingOrder);
				SortHelper::sort(kandidatenJoinModels, EinzelWertungsListeJoinModel::MITGLIEDMITGLIEDSNUMMERKOMPLETT, SortHelper::STANDARD_STRING_SORT, Qt::AscendingOrder);
				for (std::vector<EinzelWertungsListeJoinModel*>::const_iterator it = kandidatenJoinModels.begin(); it != kandidatenJoinModels.end(); it++) {
					neueSortierung.push_back(*it);
				}

				for (std::set<qlonglong>::const_iterator it = kandidatenMitglieder.begin(); it != kandidatenMitglieder.end(); it++) {
					qlonglong mitgliedsId = *it;
					context->log("ImportExport", "EinzelWertungenExportCallStrategy", 5, "Erasing Mitglied " + QString::number(mitgliedsId) + " from data structures.");
					delete mitgliedsIdToBildPunkte[mitgliedsId];
					mitgliedsIdToBildPunkte.erase(mitgliedsId);
				}
			}

			currSortModels.clear();

			currSortModels.push_back(currSortModel);
			punkte = currPunkte;
		}
	}

	joinModels = std::vector<EinzelWertungsListeJoinModel*>();
	for (unsigned int n = 0; n < neueSortierung.size(); n++) {
		EinzelWertungsListeJoinModel *joinModel = (EinzelWertungsListeJoinModel*)neueSortierung[n];
		joinModels.push_back(joinModel);
	}


//	joinModels = std::vector<EinzelWertungsListeJoinModel*>();
//	for (int n = 0; n < sortModels.size(); n++) {
//		EinzelWertungsListeJoinModel *joinModel = (EinzelWertungsListeJoinModel*)sortModels[n];
//		joinModels.push_back(joinModel);
//	}

	std::vector<Bild*> bilder = std::vector<Bild*>();
	std::set<qlonglong> bildIds = std::set<qlonglong>();
	for (std::vector<EinzelWertungsListeJoinModel*>::const_iterator it = joinModels.begin(); it != joinModels.end(); it++) {
		Bild *bild = EinzelWertungsListeJoinController::extractBild(*it);
		if (bildIds.find(bild->getId()) == bildIds.end()) {
			bildIds.insert(bild->getId());
			bilder.push_back(bild);
			persistents.push_back(bild);
		}
	}

	context->log("ImportExport", "EinzelWertungenExportCallStrategy", 6, "Exportiere " + QString::number(bilder.size()) + " Bilder.");

	// construct Exporter and DataSet
	DataExporter *exporter = new CSVDataExporter();
	DataSet *dataSet = exporter->constructDataSetForExport();
	exporter->addToDataSetForExport(context, dataSet, persistents);

	std::map<QString, std::vector<ExportPath*> > exportPaths = std::map<QString, std::vector<ExportPath*> >();
	std::vector<ExportPath*> bildVector = std::vector<ExportPath*>();

	QStringList bildNummerLokalList = QStringList();
	bildNummerLokalList << "bildDaten" << "nummerProWettbewerb";
	bildVector.push_back(new ExportPath(bildNummerLokalList, "Bildnummer"));

	QStringList idList = QStringList();
	idList << "id";
	bildVector.push_back(new ExportPath(idList, "Einzelwertungen der Juroren", new EinzelWertungenConverter(joinModels, juriertBeiIdToPerson)));

	QStringList bildTitelList = QStringList();
	bildTitelList << "titel";
	bildVector.push_back(new ExportPath(bildTitelList, "Titel"));

	QStringList bildThemenbereichList = QStringList();
	bildThemenbereichList << "themenbereich";
	bildVector.push_back(new ExportPath(bildThemenbereichList, "Themenbereich"));

	QStringList sparteList = QStringList();
	sparteList << "sparte" << "name";
	bildVector.push_back(new ExportPath(sparteList, "Sparte"));

	QStringList vornameList = QStringList();
	vornameList << "wettbewerbsTeilnahme" << "mitglied" << "vorname";
	bildVector.push_back(new ExportPath(vornameList, "Vorname"));

	QStringList nachnameList = QStringList();
	nachnameList << "wettbewerbsTeilnahme" << "mitglied" << "nachname";
	bildVector.push_back(new ExportPath(nachnameList, "Nachname"));

	QStringList mitgliedsnrList = QStringList();
	mitgliedsnrList << "wettbewerbsTeilnahme" << "mitglied" << "mitgliedsNummerKomplett";
	bildVector.push_back(new ExportPath(mitgliedsnrList, "Mitgliedsnummer"));

	QStringList vereinList = QStringList();
	vereinList << "wettbewerbsTeilnahme" << "mitglied" << "verein" << "name";
	bildVector.push_back(new ExportPath(vereinList, "Verein"));

	QStringList wertungList = QStringList();
	wertungList << "bewertungString";
	bildVector.push_back(new ExportPath(wertungList, "Bewertung"));

	QStringList summenWertungsList = QStringList();
	summenWertungsList << "wettbewerbsTeilnahme" << "mitglied" << "id";
	bildVector.push_back(new ExportPath(summenWertungsList, "Summenwertung", new SummenWertungConverter(joinModels)));

	QStringList rangList = QStringList();
	rangList << "wettbewerbsTeilnahme" << "id";
	bildVector.push_back(new ExportPath(rangList, "Rang", new RangConverter(joinModels)));

	exportPaths["Bild"] = bildVector;

	ExportParameters *exportParameters = new ExportParameters(false, false, false, true, exportPaths);

	// export
	exporter->exportData(context, fileName, dataSet, exportParameters);

	delete exporter;
	delete dataSet;
	delete exportParameters;
}
