#include "CSVDataImporter.h"

#include <QStringList>
#include <QTextStream>
#include <QVariant>

#include <iostream>

#include "DatabaseContext.h"
#include "ImportException.h"

DataSet* CSVDataImporter::readData(QFile* file) {
	this->csvFormat = new CSVFormat();

	getContext()->log("ImportExport", "CSVDataImporter", 7, "Starting with to read file" + file->fileName());

	emit statusChanged(trUtf8("Beginne mit dem Lesen der Daten"));

	CSVFileContents contents = checkFileAndSplit(getContext(), file);

	emit statusChanged(trUtf8("Beginne mit dem Vorverarbeiten der Daten"));

	DataSet *dataSet = parseContents(contents);
	dataSet->printStructure(getContext());
	dataSet->printData(getContext());
	getContext()->log("ImportExport", "CSVDataImporter", 7, "Finished reading file");

	delete csvFormat;
	return dataSet;
}

ImpExpMetaInformation CSVDataImporter::extractMetaInformation(DatabaseContext *context, QFile *file) {
	context->log("ImportExport", "CSVDataImporter", 6, "Starting to extract meta information");

	QString dataModelVersion = QString();
	QString generatorVersion = QString();
	QDateTime writeTimeStamp = QDateTime();

	if (!file->open(QIODevice::ReadOnly)) {
		QString longMessage = trUtf8("Kann die Datei %1 nicht oeffnen.").arg(file->fileName());
		throw Exception("ImportExport", "CSVDataImporter", trUtf8("Datei nicht lesbar"), longMessage);
	}

	CSVFileContents contents = checkFileAndSplit(context, file);

	for (unsigned int n = 0; n < contents.size(); n++) {
		CSVLine csvLine = contents[n];
		if (csvLine.size() > 0) {
			QString lineType = csvLine[0];
			if (lineType == csvFormat->getDataDescriptor() || lineType == csvFormat->getSchemaDescriptor()) {
				// data and schema lines, we ignore them here
				continue;
			} else if (lineType == csvFormat->getDataModelVersionDescriptor() 
					|| lineType == csvFormat->getGeneratorVersionDescriptor() 
				    || lineType == csvFormat->getWrittenAtDescriptor()) {
				if (csvLine.size() != 2) {
					QString shortMessage = trUtf8("Falsche Zeilenlaenge");
					QString longMessage = trUtf8("Die Zeile, die mit dem Schluesselwort %1 beginnt muss inklusive des Schluesselworts "
												 "genau zwei Datenworte haben.  Sie hat aber %2 Datenworte.").arg(lineType).arg(csvLine.size());
					throw Exception("ImportExport", "CSVDataImporter", shortMessage, longMessage);
				} else {
					if (lineType == csvFormat->getDataModelVersionDescriptor()) {
						if (dataModelVersion.isNull()) {
							dataModelVersion = csvLine[1];
							context->log("ImportExport", "CSVDataImporter", 6, "Data model version is " + dataModelVersion);
						} else {
							throwDuplicateMetaInfoLineException(lineType);
						}
					} else if (lineType == csvFormat->getGeneratorVersionDescriptor()) {
						if (generatorVersion.isNull()) {
							generatorVersion = csvLine[1];
							context->log("ImportExport", "CSVDataImporter", 6, "Generator version is " + generatorVersion);
						} else {
							throwDuplicateMetaInfoLineException(lineType);
						}
					} else if (lineType == csvFormat->getWrittenAtDescriptor()) {
						if (!writeTimeStamp.isValid()) {
							writeTimeStamp = QDateTime::fromString(csvLine[1], "dd.MM.yyyy hh:mm:ss");
							context->log("ImportExport", "CSVDataImporter", 6, "Generator version is " + generatorVersion);
							if (!writeTimeStamp.isValid()) {
								QString shortMessage = trUtf8("Ungueltiger Zeitstempel");
								QString longMessage = trUtf8("Der Zeitstempel %1 ist nicht gueltig.  Zeitstempel in csv-Dateien muessen das Format "
															 "dd.MM.yyyy hh:mm:ss haben.");
								throw Exception("ImportExport", "CSVDataImporter", shortMessage, longMessage);
							}
						} else {
							throwDuplicateMetaInfoLineException(lineType);
						}
					}
				}
			} else {
				// We don´t care about anything unknown here, this function is just for figuring out
				// basic properties of the file, e.g. wether it has some old version we have to
				// treat special in some way
				continue;
			}
		}
	}

	if (dataModelVersion.isNull()) {
//		throwMissingMetaInfoLineException("DatenmodellVersion");
	} else if (generatorVersion.isNull()) {
//		throwMissingMetaInfoLineException("GeneratorVersion");
	} else if (!writeTimeStamp.isValid()) {
//		throwMissingMetaInfoLineException("GeschriebenAm");
	}

	return ImpExpMetaInformation(dataModelVersion, generatorVersion, writeTimeStamp);
}

void CSVDataImporter::throwDuplicateMetaInfoLineException(QString lineType) {
	QString shortMessage = trUtf8("Doppelte Zeile");
	QString longMessage = trUtf8("Es darf in der CSV-Datei nur genau eine Zeile geben, die mit dem Schluesselwort %1 beginnt.").arg(lineType);
	throw Exception("ImportExport", "CSVDataImporter", shortMessage, longMessage);
}

void CSVDataImporter::throwMissingMetaInfoLineException(QString lineType) {
	QString shortMessage = trUtf8("Fehlende Zeile");
	QString longMessage = trUtf8("Es muss in der CSV-Datei genau eine Zeile geben, die mit dem Schluesselwort %1 beginnt.  Diese fehlt.").arg(lineType);
	throw Exception("ImportExport", "CSVDataImporter", shortMessage, longMessage);
}

DataSet* CSVDataImporter::parseContents(CSVFileContents contents) {
	getContext()->log("ImportExport", "CSVDataImporter", 6, "Starting to parse csv file contents");

	std::map<QString, ImpExpSchema*> entityDBNameToSchema = std::map<QString, ImpExpSchema*>();

	// Map from entity names to the paths in the order they are given in the schema-lines
	std::map<QString, QStringList> entityDBNameToPathKeys = std::map<QString, QStringList>();

	getContext()->log("ImportExport", "CSVDataImporter", 4, "Starting to inspect schema lines");

	// first have a look on the Schema-Lines
	int lineNumber = 1;
	for (unsigned int n = 0; n < contents.size(); n++) {
		CSVLine csvLine = contents[n];
		if (csvLine.size() > 0) {
			QString lineType = csvLine[0];
			if (lineType == csvFormat->getSchemaDescriptor()) {
				QString entityDBName = csvLine[1];
				getContext()->log("ImportExport", "CSVDataImporter", 4, "Inspecting schema line for entity" + entityDBName);
				if (entityDBNameToSchema.find(entityDBName) != entityDBNameToSchema.end()) {
					QString longMessage = trUtf8("Die Schemainformation zur Entitaet %1 ist mehrfach vorhanden, "
											 "u.a. in Zeile %2.  Sie darf nur genau einmal vorkommen.")
											 .arg(entityDBName).arg(lineNumber);
					throw Exception("ImportExport", "CSVDataImporter", trUtf8("CSV-Datei enthaelt mehrfache Schemainformation zur Entitaet %1").arg(entityDBName),
									longMessage);
				} else {
					// ignore whitespace at the end of lines
					for (unsigned int z = csvLine.size() - 1; z >= 2; z--) {
						if (csvLine[z].trimmed().length() == 0) {
							csvLine.resize(z);
						} else {
							break;
						}
					}

					QStringList pathKeys = QStringList();

					std::vector<QStringList> paths = std::vector<QStringList>();
					for (unsigned int z = 2; z < csvLine.size(); z++) {
						if (csvLine[z].length() == 0) {
							QString shortMessage = trUtf8("Leeres Datenwort in Schemabeschreibung");
							QString longMessage = trUtf8("Die Schemabeschreibung zur Tabelle %1 enthaelt "
									  "an Position %2 ein leeres Datenwort.  Dies ist nicht erlaubt.");
							throw Exception("ImportExport", "CSVDataImporter", shortMessage, longMessage);
						}
						paths.push_back(csvLine[z].split("."));
						pathKeys.push_back(csvLine[z]);
					}
					ImpExpSchema *schema = ImpExpSchema::constructFromAttributePaths(getContext(), entityDBName, paths);
					entityDBNameToSchema[entityDBName] = schema;
					entityDBNameToPathKeys[entityDBName] = pathKeys;
				}
			}
		}
		lineNumber++;
	}

	getContext()->log("ImportExport", "CSVDataImporter", 4, "Starting to inspect data lines");

	DataSet *dataSet = new DataSet();

	lineNumber = 1;
	for (unsigned int n = 0; n < contents.size(); n++) {
		CSVLine csvLine = contents[n];
		if (csvLine.size() > 0) {
			QString lineType = csvLine[0];
			if (lineType == csvFormat->getDataDescriptor()) {
				QString entityDBName = csvLine[1];
				if (entityDBNameToSchema.find(entityDBName) == entityDBNameToSchema.end()) {
					QString longMessage = trUtf8("Zur Tabelle %1 ist keine Schemainformation vorhanden. "
											 "Bitte ergaenzen Sie eine Schema-Zeile fuer diese Tabelle.")
											 .arg(entityDBName).arg(lineNumber);
					delete dataSet;
					throw Exception("ImportExport", "CSVDataImporter", trUtf8("Fehlende Schemainformation"), longMessage);
				} else if (entityDBNameToSchema[entityDBName]->isAbstractEntity()) {
					QString longMessage = trUtf8("Die in den zu importierenden Daten angegebene Tabelle %1 "
											 "existiert nicht im Datenmodell.  Wohl aber existiert eine "
											 "abstrakte Entitaet gleichen Namens.  Diese dient aber nur "
											 "der programminternen Organisation und hat keine Entsprechung "
											 "auf der Datenbank.").arg(entityDBName);
					delete dataSet;
					throw ImportException("ImportExport", "CSVDataImporter", trUtf8("Import in abstrakte Entitaet nicht moeglich."), longMessage);
				} else {
					QStringList pathKeys = entityDBNameToPathKeys[entityDBName];

					if (int(pathKeys.size()) > int(csvLine.size()) - 2) {
						QString longMessage = trUtf8("Die Datenzeile %1 zur Tabelle %2 ist zu kurz.  "
												 "Sie muss Datenworte zu allen Eintraegen in der dazugehoerigen "
												 "Schemazeile enthalten.  Die Schemazeile enthaelt %3 Datenworte, "
												 "die Datenzeile aber %4 Datenworte. \n\n"
												 "Falls die Importdatei von Hand erstellt wurde, korrigieren Sie "
												 "bitte Schema- oder Datenzeile.  Falls sie automatisch erzeugt wurde, "
												 "kontaktieren Sie bitte den Administrator.")
											  .arg(lineNumber).arg(entityDBName).arg(pathKeys.size()).arg(csvLine.size() - 2);
						delete dataSet;
						throw ImportException("ImportExport", "CSVDataImporter", trUtf8("Datenzeile zu kurz"), longMessage);
					} else {
						for (unsigned int z = csvLine.size() - 1; z >= (int)pathKeys.size() + 2; z--) {
							if (csvLine[z].trimmed().length() == 0) {
								// Leere Zellen am Ende werden einfach ignoriert.  Excel erzeugt sie manchmal
								// recht unmotiviert.
								csvLine.resize(z);
							} else {
								// Volle Zellen ohne Entsprechung im Schema sind dagegen eindeutig ein Fehler in der Datei.
								QString shortMessage = trUtf8("Datenwort ohne Entsprechung im Schema");
								QString longMessage = trUtf8("In der Datenzeile %1 zur Tabelle %2 "
										  "steht an Position %3 ein Datenwort, das keine Entsprechung "
										  "in der Schemazeile zur Tabelle %2 hat. \n\n"
										  "Falls die Importdatei von Hand erstellt wurde, korrigieren Sie "
										  "bitte Schema- oder Datenzeile.  Falls sie automatisch erzeugt wurde, "
										  "kontaktieren Sie bitte den Administrator.")
									  .arg(lineNumber).arg(entityDBName).arg(z);
								delete dataSet;
								throw ImportException("ImportExport", "CSVDataImporter", shortMessage, longMessage);
							}
						}

						// in a csv file, no two tuples can have the same lineNumber, thus using it as id is appropriate
						DataTuple *tuple = new DataTuple(entityDBName, lineNumber, lineNumber);
						for (unsigned int z = 2; z < csvLine.size() && z < (unsigned int)pathKeys.size() + 2; z++) {
							// ignore tokens that have no counterpart in the schema line
							if (csvLine[z].length() == 0) {
								tuple->setNull(pathKeys[z - 2]);
							} else {
								tuple->setValue(pathKeys[z - 2], QVariant(csvLine[z]));
							}
						}
						dataSet->addDataTuple(entityDBName, tuple);
						getContext()->log("ImportExport", "CSVDataImporter", 5,
								  "Finished reading a DataTuple of entity db name " + entityDBName);
					}
				}
			}
		}
		lineNumber++;
	}

	// Above we loaded the schema as given in the import file.  But lateron, we
	// want to use the real schema.
	std::map<QString, ImpExpSchema*> realEntityDBNameToSchema
		= ImpExpSchema::constructImpExpSchemas(getContext(), dataSet);
	dataSet->setEntityDBNameToSchemaMap(realEntityDBNameToSchema);

	getContext()->log("Import", "CSVDataImporter", 6, "Finished parseContents(CSVFileContents)");
	return dataSet;
}

CSVFileContents CSVDataImporter::checkFileAndSplit(DatabaseContext *context, QFile* file) {
	QString separator = csvFormat->getDelimiter();
	QString commentStart = csvFormat->getCommentStart();
	QString schemaType = csvFormat->getSchemaDescriptor();
	QString dataType = csvFormat->getDataDescriptor();

	context->log("ImportExport", "CSVDataImporter", 6, "Starting to split csv file into tokens");

	CSVFileContents contents = CSVFileContents();

	int lineNumber = 1;
	QTextStream stream(file);
	stream.setCodec("UTF-8");
	while (!stream.atEnd()) {
		if (lineNumber % 50 == 0) {
			emit statusChanged(trUtf8("%1 Datensaetze gelesen").arg(lineNumber));
		}

		QString line = stream.readLine();

		if (line.length() == 0 || line.startsWith(commentStart)) {
			// ignore empty and comment lines
			lineNumber++;

			// but we have to register them nevertheless, because otherwise
			// line numbers recorded later would be wrong
			contents.push_back(std::vector<QString>());
			continue;
		}

		std::vector<QString> lineParts = splitLine(lineNumber, line, separator);

		if (isLineEmpty(lineParts)) {
			// Ignore lines that only consist of the separator character,
			// e.g. ;;;;;;;;  They can e.g. be generated if an export from
			// Excel exports an empty line.
			lineNumber++;
			contents.push_back(std::vector<QString>());
			continue;
		} else if (lineParts.size() < 2) {
			// Erstes Token Info Schema oder Daten, zweites Entitaets-DB-Name, Rest Attributnamen oder Daten
			QString longMessage = trUtf8("Das erste Wort jeder Zeile der CSV-Datei muss angeben, "
									 "ob es sich um Schema-Information (%1) oder um Daten (%2) handelt. "
									 "Das zweite muss den Entitaetsnamen angeben, die restlichen Woerter "
									 "enthalten Attributnamen bzw. Daten.  "
									 "Zeile %3 enthaelt weniger als zwei Woerter."
									 ).arg(schemaType).arg(dataType).arg(lineNumber);
			throw ImportException("ImportExport", "CSVDataImporter", trUtf8("Zeile %1 der CSV-Datei ist zu kurz").arg(lineNumber), longMessage);
		} else if (lineParts[0] != schemaType && lineParts[0] != dataType 
					&& lineParts[0] != "DatenmodellVersion" && lineParts[0] != "GeneratorVersion" && lineParts[0] != "GeschriebenAm") {
			QString longMessage = trUtf8("Das erste Wort jeder Zeile der CSV-Datei muss die Art der in der Zeile gespeicherten Information angeben. "
									 "Erlaubt sind Schema, Daten, DatenmodellVersion, GeneratorVersion und GeschriebenAm. "
									 "Das erste Wort von Zeile %1 ist aber keines dieser Schluesselwoerter!").arg(lineNumber);
			throw ImportException("ImportExport", "CSVDataImporter", trUtf8("Zeile %1 der CSV-Datei ist weder Schema- noch Datenzeile")
									 .arg(lineNumber), longMessage);
		} else {
			// line ok
			contents.push_back(lineParts);
		}
		lineNumber++;
	}
	context->log("ImportExport", "CSVDataImporter", 6, "Finished splitting file up into tokens");
	return contents;
}

std::vector<QString> CSVDataImporter::splitLine(int lineNumber, QString line, QString separator) {
//	std::cout << "Splitting line " << qPrintable(line) << std::endl;

	std::vector<QString> linePartsAfterQuotes = std::vector<QString>();

	QStringList lineParts = line.split(separator);

	QString quote = "\"";
	QString escapedQuote = "\\\"";
	bool withinQuotes = false;
	QString appendedString = "";
	for (int n = 0; n < lineParts.size(); n++) {
		QString linePart = lineParts[n];
		if (withinQuotes) {
			// we append all strings within one pair of quotes
			// we have to add the separator here, because we know that we have
			// removed it before (split doesn't know anything about our quotes)
			appendedString += separator + linePart;
		}
		// Ignore escaped quotes
		if (((linePart.count(quote) - linePart.count(escapedQuote)) % 2) == 1) {
			// either a pair of quotes starts or ends
			withinQuotes = !withinQuotes;
			if (!withinQuotes) {
				// if we are no longer within quotes afterwards, the current
				// string is completed.  Remove its quotes.
				appendedString = appendedString.trimmed();
				// Remove all occurrences of " which are not escaped
				appendedString = removeNonEscapedQuotes(appendedString);
				// Do not transform escaped occurrences of " into non-escaped occurrences of ", as
				// we need them in escaped form when inserting into the database (i.e. \" is
			    // transformed into " by mysql)
//				appendedString = appendedString.replace("\\\"", "\"");
				linePartsAfterQuotes.push_back(appendedString);
//				std::cout << "==> appending " << qPrintable(appendedString) << std::endl;
				appendedString = "";
			} else {
				// we just started a pair of quotes
				appendedString = linePart;
			}
		} else if (!withinQuotes) {
			// we simply pass the current line part through, removing any quotes
			// (e.g. if it contains exactly two, like "555")
			linePart = linePart.trimmed();
			// Remove all occurrences of " which are not escaped
			linePart = removeNonEscapedQuotes(linePart);
			// Do not transform escaped occurrences of " into non-escaped occurrences of ", as
			// we need them in escaped form when inserting into the database (i.e. \" is
			// transformed into " by mysql)
//			linePart = linePart.replace("\\\"", "\"");
			linePartsAfterQuotes.push_back(linePart);
//			std::cout << "==> appending " << qPrintable(linePart) << std::endl;
		}
	}

	if (withinQuotes) {
		QString longMessage = trUtf8("Beim Einlesen der Zeile %1 der CSV-Datei: "
								 "Es ist eine ungerade Zahl an Anfuehrungszeichen vorhanden."
								 "Alles innerhalb eines Paares von Anfuehrungszeichen wird als ein Wort angesehen, "
								 "daher ist diese Zeile fehlerhaft.").arg(lineNumber);
		throw ImportException(trUtf8("Fehlerhafte Anfuehrungszeichen in Zeile %1").arg(lineNumber), longMessage);
	}

	return linePartsAfterQuotes;
}

bool CSVDataImporter::isLineEmpty(std::vector<QString> lineParts) {
	bool isEmpty = true;
	for (std::vector<QString>::const_iterator it = lineParts.begin(); it != lineParts.end(); it++) {
		QString linePart = *it;
		if (linePart.length() > 0) {
			isEmpty = false;
		}
	}
	return isEmpty;
}

QString CSVDataImporter::removeNonEscapedQuotes(QString s) {
	for (int n = 0; n < s.length(); n++) {
		QString character = s.mid(n, 1);
		if (character == "\"") {
			QString characterBefore = (n == 0 ? "" : s.mid(n - 1, 1));
			if (characterBefore != "\\") {
				s = s.replace(n, 1, "");
				// After removing the n-th character, all characters located right
				// move one position to the left
				n--;
			}
		}
	}
	return s;
}
