#include "DataImporter.h"

#include <QDir>
#include <QFile>
#include <QStringList>

#include <iostream>
#include <stdlib.h>

#include <set>
#include <vector>

#include "ArchiveHelper.h"
#include "Attribute.h"
#include "AttributeData.h"
#include "BenutzerImportStrategy.h"
#include "BildImportStrategy.h"
#include "BildDatenImportStrategy.h"
#include "CodeTypeImportStrategy.h"
#include "CodeValueImportStrategy.h"
#include "ControllerHelper.h"
#include "DataHelper.h"
#include "DataModel.h"
#include "DurchgangImportStrategy.h"
#include "EinzelWertungImportStrategy.h"
#include "Entity.h"
#include "EventLogImportStrategy.h"
#include "EventLogCategoryImportStrategy.h"
#include "EventLogSettingsImportStrategy.h"
#include "ExternerJurorImportStrategy.h"
#include "FotographischerTitelImportStrategy.h"
#include "FunktionstraegerImportStrategy.h"
#include "GlobaleEinstellungImportStrategy.h"
#include "IgnoreImportStrategy.h"
#include "ImportException.h"
#include "ImportStrategy.h"
#include "JuriertBeiImportStrategy.h"
#include "JurierungImportStrategy.h"
#include "MitgliedImportStrategy.h"
#include "PersistentImportStrategy.h"
#include "QueryLogImportStrategy.h"
#include "QueryLogSettingsImportStrategy.h"
#include "SessionImportStrategy.h"
#include "SparteImportStrategy.h"
#include "SpartenJurierungImportStrategy.h"
#include "TeilnehmerKombinationsWertungImportStrategy.h"
#include "TeilnehmerWertungImportStrategy.h"
#include "VereinImportStrategy.h"
#include "VereinsDatenBerechtigungImportStrategy.h"
#include "VereinsKombinationsWertungImportStrategy.h"
#include "VereinsWertungImportStrategy.h"
#include "WertungImportStrategy.h"
#include "WettbewerbImportStrategy.h"
#include "WettbewerbsTeilnahmeImportStrategy.h"

DataImporter::DataImporter() {
	entityDBNameToImportStrategy = std::map<QString, ImportStrategy*>();
	entityDBNameToImportStrategy["Benutzer"] = new BenutzerImportStrategy();
	entityDBNameToImportStrategy["Bild"] = new BildImportStrategy();
	entityDBNameToImportStrategy["BildDaten"] = new BildDatenImportStrategy();
	entityDBNameToImportStrategy["CodeType"] = new CodeTypeImportStrategy();
	entityDBNameToImportStrategy["CodeValue"] = new CodeValueImportStrategy();
	entityDBNameToImportStrategy["DataModelInfo"] = new IgnoreImportStrategy();
	entityDBNameToImportStrategy["DatenbankVerbindung"] = new IgnoreImportStrategy();
	entityDBNameToImportStrategy["Durchgang"] = new DurchgangImportStrategy();
	entityDBNameToImportStrategy["EinzelWertung"] = new EinzelWertungImportStrategy();
	entityDBNameToImportStrategy["EventLog"] = new EventLogImportStrategy();
	entityDBNameToImportStrategy["EventLogCategory"] = new EventLogCategoryImportStrategy();
	entityDBNameToImportStrategy["EventLogSettings"] = new EventLogSettingsImportStrategy();
	entityDBNameToImportStrategy["ExternerJuror"] = new ExternerJurorImportStrategy();
	entityDBNameToImportStrategy["FotographischerTitel"] = new FotographischerTitelImportStrategy();
	entityDBNameToImportStrategy["Funktionstraeger"] = new FunktionstraegerImportStrategy();
	entityDBNameToImportStrategy["JuriertBei"] = new JuriertBeiImportStrategy();
	entityDBNameToImportStrategy["Jurierung"] = new JurierungImportStrategy();
	entityDBNameToImportStrategy["GlobaleEinstellung"] = new GlobaleEinstellungImportStrategy();
	entityDBNameToImportStrategy["Mitglied"] = new MitgliedImportStrategy();
	entityDBNameToImportStrategy["QueryLog"] = new QueryLogImportStrategy();
	entityDBNameToImportStrategy["QueryLogSettings"] = new QueryLogSettingsImportStrategy();
	entityDBNameToImportStrategy["Session"] = new SessionImportStrategy();
	entityDBNameToImportStrategy["Sparte"] = new SparteImportStrategy();
	entityDBNameToImportStrategy["SpartenJurierung"] = new SpartenJurierungImportStrategy();
	entityDBNameToImportStrategy["TeilnehmerKombinationsWertung"] = new TeilnehmerKombinationsWertungImportStrategy();
	entityDBNameToImportStrategy["TeilnehmerWertung"] = new TeilnehmerWertungImportStrategy();
	entityDBNameToImportStrategy["Verein"] = new VereinImportStrategy();
	entityDBNameToImportStrategy["VereinsDatenBerechtigung"] = new VereinsDatenBerechtigungImportStrategy();
	entityDBNameToImportStrategy["VereinsKombinationsWertung"] = new VereinsKombinationsWertungImportStrategy();
	entityDBNameToImportStrategy["VereinsWertung"] = new VereinsWertungImportStrategy();
	entityDBNameToImportStrategy["Wertung"] = new WertungImportStrategy();
	entityDBNameToImportStrategy["Wettbewerb"] = new WettbewerbImportStrategy();
	entityDBNameToImportStrategy["WettbewerbsTeilnahme"] = new WettbewerbsTeilnahmeImportStrategy();
	standardImportStrategy = new PersistentImportStrategy();
}

DataImporter::~DataImporter() {
	for (std::map<QString, ImportStrategy*>::const_iterator it = entityDBNameToImportStrategy.begin();
		  it != entityDBNameToImportStrategy.end(); it++) {
		delete it->second;
	}
}

ImportResult* DataImporter::import(DatabaseContext *context, QString fileName) {
	return import(context, new QFile(fileName));
}

ImportResult* DataImporter::import(DatabaseContext *context, QFile* file) {
	context->log("ImportExport", "DataImporter.cpp", 8, "Started import for file " + file->fileName());

	importResult = new ImportResult();
	this->context = context;

	if (!file->open(QIODevice::ReadOnly)) {
		QString longMessage = trUtf8("Kann die Datei %1 nicht oeffnen.").arg(file->fileName());
		throw Exception("ImportExport", "DataImporter", trUtf8("Datei nicht lesbar"), longMessage);
	}

	context->log("ImportExport", "DataImporter", 3, "Successful opened the import file.");

	DataSet *dataSet = NULL;
	try {
		emit statusChanged(trUtf8("Import begonnen"));
		dataSet = readData(file);
		dataSet->saveAttributeData();
		emit statusChanged(trUtf8("Daten aus Datei geladen"));
		preprocessData(dataSet, file);
		dataSet->printData(context);

		mysqlpp::Transaction transaction(*context->getConnection());

		writeToDatabase(dataSet, file, importResult);
		updateForeignReferences(dataSet);
		emit statusChanged(trUtf8("Daten geschrieben, beginne mit Nachbearbeitungen"));
		postProcessAfterWrite(dataSet);

		transaction.commit();

		emit statusChanged(trUtf8("Import erfolgreich beendet"));
	} catch (Exception &e) {
		delete file;
		delete dataSet;
		throw;
	}

	context->log("ImportExport", "DataImporter", 8, "Finished import for file " + file->fileName());

	delete file;
	delete dataSet;
	return importResult;
}

void DataImporter::addNonFatalError(Exception e) {
	importResult->addException(e);
}

DatabaseContext *DataImporter::getContext() {
	return context;
}

void DataImporter::preprocessData(DataSet *dataSet, QFile *importFile) {
	context->log("ImportExport", "DataImporter", 7, "Starting to preprocess data");
	checkAndConvertRawTypes(dataSet);
	emit statusChanged(trUtf8("Pruefung auf korrekte Syntax der Basistypen erfolgreich abgeschlossen."));
	checkBLOBs(dataSet, importFile);
	emit statusChanged(trUtf8("Alle aus der Eingabedatei referenzierten Dateien sind vorhanden."));
	checkAndConvertBeforeResolve(dataSet);
	emit statusChanged(trUtf8("Pruefungen und Konvertierungen erfolgreich abgeschlossen"));
	resolveReferences(dataSet);
	emit statusChanged(trUtf8("Fremdschluesselbeziehungen erfolgreich aufgeloest"));
	mapToDatabaseTuples(dataSet);
	emit statusChanged(trUtf8("Zuordnung zu Tupeln auf der Datenbank erfolgreich hergestellt"));
	checkForDuplicatesInImportData(dataSet);
	emit statusChanged(trUtf8("Duplikaterkennung erfolgreich abgeschlossen"));
	context->log("ImportExport", "DataImporter", 7, "Finished preprocessing data");
}

void DataImporter::resolveReferences(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting to resolve references");

	emit statusChanged(trUtf8("Beginne mit Fremdschluesselaufloesung..."));

	std::vector<DataTuple*> tupleVector = dataSet->getAllTuples();
	// maps from tupleId to DataTuple
	std::map<int, DataTuple*> unfinishedTuples = std::map<int, DataTuple*>();
	std::map<int, DataTuple*> finishedTuples = std::map<int, DataTuple*>();

	// initially, all tuples are unfinished
	for (std::vector<DataTuple*>::const_iterator it = tupleVector.begin(); it != tupleVector.end(); it++) {
		DataTuple *tuple = *it;
		unfinishedTuples[tuple->getTupleId()] = tuple;
	}

	context->log("ImportExport", "DataImporter", 2, "Set all tuples to state unfinished");

	int loopNumber = 1;
	bool progress;
	// perform a fix point iteration.  We proceed until all tuples are
	// preprocessed or we can't make any further progress.
	do {
		QString msg = "Starting iteration " + QString::number(loopNumber)
					+ ", " + QString::number(unfinishedTuples.size())
					+ " unfinished tuples are left.";
		context->log("ImportExport", "DataImporter", 5, msg);

		progress = false;
		std::map<int, DataTuple*> finishedTuplesThisTime = std::map<int, DataTuple*>();
		int innerLoopNumber = 0;
		for (std::map<int, DataTuple*>::const_iterator it = unfinishedTuples.begin();
			  it != unfinishedTuples.end(); it++) {
			innerLoopNumber++;

			if (innerLoopNumber % 50 == 0) {
				emit statusChanged(trUtf8("Loese Fremdschluesselbeziehungen auf: Durchlauf %1,  %2 von %3 Datensaetzen des Durchlaufs verarbeitet")
					.arg(loopNumber).arg(innerLoopNumber).arg(unfinishedTuples.size()));
			}
			DataTuple *tuple = it->second;

			// try to resolve tuple
			bool tupleProgress = resolveTuple(dataSet, tuple);
			bool finished = isTupleFinished(dataSet, tuple);
			tupleProgress |= finished;

			// debug statements
			QString tupleAddress = QString::number((qlonglong)this, 16);
			if (tupleProgress) {
				context->log("ImportExport", "DataImporter", 4, "There was progress for tuple " + tupleAddress + " of type " + tuple->getEntityDBName());
			} else {
				context->log("ImportExport", "DataImporter", 3, "There was no progress for tuple " + tupleAddress + " of type " + tuple->getEntityDBName());
			}

			progress |= tupleProgress;

			if (finished) {
				finishedTuplesThisTime[tuple->getTupleId()] = tuple;
				context->log("ImportExport", "DataImporter", 4, "Tuple " + tupleAddress + " was finished.");
			}
		}

		std::map<int, DataTuple*> newUnfinishedTuples = std::map<int, DataTuple*>();
		for (std::map<int, DataTuple*>::const_iterator it = unfinishedTuples.begin();
				  it != unfinishedTuples.end(); it++) {
			DataTuple *tuple = it->second;
			int tupleId = tuple->getTupleId();
			if (finishedTuplesThisTime.find(tupleId) != finishedTuplesThisTime.end()) {
				finishedTuples[tupleId] = tuple;
			} else {
				newUnfinishedTuples[tupleId] = tuple;
			}
		}
		unfinishedTuples = newUnfinishedTuples;
		context->log("ImportExport", "DataImporter", 5,
								  "There was " + (progress ? QString("") : QString("no ")) + "progress. "
								+ "Unfinished: " + QString::number(unfinishedTuples.size())
								+ ". Finished: " + QString::number(finishedTuples.size()));
		dataSet->printData(context);

		emit statusChanged(trUtf8("Loese Fremdschluesselbeziehungen auf: Durchlauf %1, noch %2 Datensaetze offen").arg(loopNumber).arg(unfinishedTuples.size()));
		loopNumber++;
	} while (progress && unfinishedTuples.size() > 0);

	if (unfinishedTuples.size() > 0) {
		context->log("ImportExport", "DataImporter", 7, "There are unfinished tuples left, starting to generate error information...");

		std::vector<DataTuple*> unfinishedTuplesVector = std::vector<DataTuple*>();
		for (std::map<int, DataTuple*>::const_iterator it = unfinishedTuples.begin(); it != unfinishedTuples.end(); it++) {
			unfinishedTuplesVector.push_back(it->second);
		}

		std::vector<QString> messages = getUnfinishedTuplesErrorStrings(dataSet, unfinishedTuplesVector);
		QString gesamtMessage = "";
		for (uint n = 0; n < messages.size() && n < 5; n++) {
			gesamtMessage += messages[n] + "\n";
		}

		QString shortMessage = trUtf8("Import nicht moeglich");
		QString longMessage = trUtf8("Der Import ist nicht moeglich, da nicht alle "
						 "Querverweise zwischen (a) Datensaetzen der Importdatei untereinander und (b) "
						 "Datensaetzen der Importdatei und der Datenbank aufgeloest werden konnten.\n\n"
						 "Im Detail gab es folgende Probleme: \n%1").arg(gesamtMessage);
		throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
	}

	// Now we know that all tuples are in a state ready to import, i.e. all
	// references are resolved.

	context->log("ImportExport", "DataImporter", 7, "Finished resolving references");
}

std::vector<QString> DataImporter::getUnfinishedTuplesErrorStrings(DataSet *dataSet, std::vector<DataTuple*> unfinishedTuples) {
	std::vector<QString> messages = std::vector<QString>();

	for (std::vector<DataTuple*>::const_iterator it = unfinishedTuples.begin(); it != unfinishedTuples.end(); it++) {
		DataTuple *dataTuple = *it;
		int lineNumber = dataTuple->getLineNumber();

		QString entityDBName = dataTuple->getEntityDBName();
		ImpExpSchema* schema = dataSet->getSchema(entityDBName);

		std::map<QString, ForeignKeyGraph*> attributeDBNameToGraph = schema->getAttributeToGraphMap();
		for (std::map<QString, ForeignKeyGraph*>::const_iterator it = attributeDBNameToGraph.begin();
			it != attributeDBNameToGraph.end(); it++) {
			QString attributeDBName = it->first;
			ForeignKeyGraph *graph = it->second;

			ForeignKeyNode *rootNode = graph->getRootNode();

			std::vector<ForeignKeyEdge*> childEdges = rootNode->getChildEdges();
			for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
				ForeignKeyEdge *edge = *it;
				QString destEntityDBName = edge->getOutNodeEntityDBName();

				QString pathKey = edge->getPathKey();
				if (!dataTuple->hasAttributeData(pathKey)) {
					std::map<QString, QString> originalPathKeyToData = dataTuple->getOriginalPathKeyToData();
					QString detailData = "";
					for (std::map<QString, QString>::const_iterator it = originalPathKeyToData.begin(); it != originalPathKeyToData.end(); it++) {
						QString originalPathKey = it->first;
						if (originalPathKey.startsWith(pathKey)) {
							QString originalValue = it->second;
							detailData += originalPathKey + "=" + originalValue + " ";
						}
					}

					QString message = trUtf8("Zeile %1: Der Verweis %2 auf die Tabelle %3 konnte nicht aufgeloest werden.  "
											 "In der Importdatei steht folgendes: %4")
										  .arg(QString::number(lineNumber)).arg(pathKey).arg(destEntityDBName).arg(detailData);
					messages.push_back(message);
				}
			}
		}
	}

	return messages;
}

bool DataImporter::resolveTuple(DataSet *dataSet, DataTuple *dataTuple) {
	QString tupleAddress = QString::number((qlonglong)this, 16);
	context->log("ImportExport", "DataImporter", 3, "  resolveTuple called for tuple " + tupleAddress + " of type " + dataTuple->getEntityDBName());

	bool progress = false;
	QString entityDBName = dataTuple->getEntityDBName();
	ImpExpSchema* schema = dataSet->getSchema(entityDBName);

	std::map<QString, ForeignKeyGraph*> attributeDBNameToGraph = schema->getAttributeToGraphMap();
	for (std::map<QString, ForeignKeyGraph*>::const_iterator it = attributeDBNameToGraph.begin();
		it != attributeDBNameToGraph.end(); it++) {
		QString attributeDBName = it->first;
		ForeignKeyGraph *graph = it->second;

		ForeignKeyNode *rootNode = graph->getRootNode();
		context->log("ImportExport", "DataImporter", 3, "    resolveTuple will call processNode for the root node "
													  "of attribute " + attributeDBName);
		std::vector<ForeignKeyNode*> processedNodes = processNode(dataSet, dataTuple, rootNode, true);
		if (processedNodes.size() > 0) {
			context->log("ImportExport", "DataImporter", 3,
						 "      There was progress regarding the graph for attribute " + attributeDBName);
			progress = true;
		} else {
			context->log("ImportExport", "DataImporter", 3,
						 "      There was no progress regarding the graph for attribute " + attributeDBName);
		}
	}
	context->log("ImportExport", "DataImporter", 3, "  resolveTuple finished for tuple " + tupleAddress + " of type " + dataTuple->getEntityDBName());
	return progress;
}

bool DataImporter::isTupleFinished(DataSet *dataSet, DataTuple *dataTuple) {
	QString entityDBName = dataTuple->getEntityDBName();
	ImpExpSchema* schema = dataSet->getSchema(entityDBName);

	std::map<QString, ForeignKeyGraph*> attributeDBNameToGraph = schema->getAttributeToGraphMap();
	for (std::map<QString, ForeignKeyGraph*>::const_iterator it = attributeDBNameToGraph.begin();
		it != attributeDBNameToGraph.end(); it++) {
		QString attributeDBName = it->first;
		ForeignKeyGraph *graph = it->second;

		ForeignKeyNode *rootNode = graph->getRootNode();
		if (!isNodeFinished(rootNode, dataTuple)) {
			return false;
		}
	}
	return true;
}

bool DataImporter::isNodeFinished(ForeignKeyNode *node, DataTuple *dataTuple) {
	context->log("ImportExport", "DataImporter", 3, "        Calling isNodeFinished");
	std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();
		if (!dataTuple->hasAttributeData(pathKey)) {
			context->log("ImportExport", "DataImporter", 3, "          Inspecting pathKey " + pathKey + " ... no AttributeData available, thus node is not finished");
			return false;
		} else {
			context->log("ImportExport", "DataImporter", 3, "          Inspecting pathKey " + pathKey + " ... there is AttributeData available");
		}
	}

	// todo test attributes at all here?
	context->log("ImportExport", "DataImporter", 3, "        Node is finished.");
	return true;
}

void DataImporter::deleteChildren(DataTuple *dataTuple, ForeignKeyNode *node) {
	std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();
		dataTuple->deleteAttributeData(pathKey);
	}

	std::map<QString, QVariant> pathKeyToValue = std::map<QString, QVariant>();
	QString currNodePathKey = node->getPathKey();
	std::vector<const Attribute*> attributes = node->getAttributes();
	for (std::vector<const Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute *attribute = *it;
		QString attributeDBName = attribute->getDBName();
		QString pathKey = currNodePathKey + ImpExpSchema::getPathSeparator() + attributeDBName;
		dataTuple->deleteAttributeData(pathKey);
	}
}

std::vector<ForeignKeyNode*> DataImporter::resolveNode(DataSet *dataSet, DataTuple *dataTuple,
													   ForeignKeyNode *node) {
	QString pathKey = node->getPathKey();

	std::vector<ForeignKeyNode*> processedNodes = std::vector<ForeignKeyNode*>();

	if (resolveNullChildren(dataTuple, node)) {
		context->log("ImportExport", "DataImporter", 3,
					 "resolveNode considers node " + node->getPathKey() + " a NULL node");
		processedNodes.push_back(node);
		return processedNodes;
	}

	std::vector<QVariant> resultsOnDatabase = resolveOnDatabase(dataTuple, node);
	std::vector<DataTuple*> resultsInImportData = resolveInImportData(dataSet, dataTuple, node);

	if (resultsOnDatabase.size() >= 2 || resultsInImportData.size() >= 2) {

		QString infoString = "";
		std::set<QString> pathKeys = dataTuple->getAllPathKeys();
		for (std::set<QString>::const_iterator it = pathKeys.begin(); it != pathKeys.end(); it++) {
			QString pathKey = *it;
			if (dataTuple->hasAttributeData(pathKey)) {
				AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
				if (attributeData->isValue()) {
					QVariant value = attributeData->getValue();
					infoString += pathKey + " = " + value.toString() + "\n";
				}
			}
		}


		QString shortMessage = trUtf8("Verweis nicht eindeutig aufloesbar.");
		QString longMessage = trUtf8("Der Verweis %1 ist fuer einen Datensatz der Tabelle %4 nicht eindeutig aufloesbar. "
									 "Es gibt %2 Treffer auf der Datenbank und %3 Treffer in den "
									 "gerade importierten Daten. \n\n"
									 "Ist der Import/Export-Schluessel der entsprechenden Tabelle wirklich "
									 "in allen Faellen eindeutig?  Enthalten die Import-Daten eventuell "
									 "Duplikat-Datensaetze (z.B. mehrere gleiche Zeilen)? \n\n"
									 "Bitte pruefen Sie ihre Import-Daten oder kontaktieren Sie Ihren Administrator. \n\n"
									 "Der Inhalt einiger Felder des Datensatzes: \n %5")
									.arg(node->getPathKey()).arg(resultsOnDatabase.size())
									.arg(resultsInImportData.size()).arg(dataTuple->getEntityDBName()).arg(infoString);
		throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
	} else if (resultsOnDatabase.size() == 1 && resultsInImportData.size() == 1) {
		context->log("ImportExport", "DataImporter", 3,
					 "resolveNode resolves node " + pathKey + " both on db and in the input file");
		// Both in the input file and in the database is a match.  So the match
		// in the input file will overwrite the one in the database.  The database
		// id will stay constant, so we can take the id from the database here.
		dataTuple->setForeignKeyToDB(pathKey, resultsOnDatabase[0]);
		deleteChildren(dataTuple, node);
		processedNodes.push_back(node);
	} else if (resultsOnDatabase.size() == 1) {
		context->log("ImportExport", "DataImporter", 3, "resolveNode resolves node " + pathKey + " on db");
		dataTuple->setForeignKeyToDB(pathKey, resultsOnDatabase[0]);
		deleteChildren(dataTuple, node);
		processedNodes.push_back(node);
	} else if (resultsInImportData.size() == 1) {
		context->log("ImportExport", "DataImporter", 3, "resolveNode resolves node " + pathKey + " in the input data");
		dataTuple->setForeignKeyToDataTuple(pathKey, resultsInImportData[0]);
		deleteChildren(dataTuple, node);
		processedNodes.push_back(node);
	} else {
		// Only the case that both vectors are empty is left.  Then, the node
		// can't be resolved right now, but might be resolvable in future
		// (a referenced tuple may itself not yet be processed)
		// Thus return the empty vector.
	}

	return processedNodes;
}

bool DataImporter::resolveNullChildren(DataTuple *dataTuple, ForeignKeyNode *node) {
	bool anyChildNull = false;
	bool allChildrenNull = true;

	context->log("ImportExport", "DataImporter", 3, "Called resolveNullChildren for node " + node->getPathKey());

	// Find out wether (1) any child, and/or (2) all children of the given node
	// are NULL in the given DataTuple
	// First have a look on the foreign key edges ...
	std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();

		context->log("ImportExport", "DataImporter", 3, "  Checking foreign key edge " + pathKey);

		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		anyChildNull |= attributeData->isNull();
		allChildrenNull &= attributeData->isNull();
	}

	// ... and then on the attributes
	QString currNodePathKey = node->getPathKey();
	std::vector<const Attribute*> attributes = node->getAttributes();
	for (std::vector<const Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute *attribute = *it;
		// todo introduce constant for .
		QString attributeDBName = attribute->getDBName();
		QString pathKey = currNodePathKey + "." + attributeDBName;

		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData != NULL) {
			anyChildNull |= attributeData->isNull();
			allChildrenNull &= attributeData->isNull();
			context->log("ImportExport", "DataImporter", 3, "  Checking attribute " + pathKey + "... is not NULL - anyChildNull = " 
														  + (anyChildNull ? "true" : "false") + " - allChildrenNull = " + (allChildrenNull ? "true" : "false"));
		} else {
			context->log("ImportExport", "DataImporter", 3, "  Checking attribute " + pathKey + "... is NULL.");
		}
	}

  if (allChildrenNull) {
		// if all children are null, we don´t need them, and replace them
		// by a null value for the current node
		dataTuple->setNull(currNodePathKey);
		deleteChildren(dataTuple, node);
		context->log("ImportExport", "DataImporter", 3, "Finished: Deleting children of " + currNodePathKey + " as they are all null.");
		return true;
	} else {
		// if at least one child is not null, this function doesn´t have to do anything
		context->log("ImportExport", "DataImporter", 3, "At least one children of " + currNodePathKey + " is not null.");
		return false;
	}
}

std::vector<QVariant> DataImporter::resolveOnDatabase(DataTuple *dataTuple, ForeignKeyNode *node) {
	QString entityDBName = node->getEntityDBName();
	context->log("ImportExport", "DataImporter", 2, "  Resolving tuples for node with entity db name " + entityDBName + " on database...");

	// Construct a map for querying using an appropriate filter
	// First add values for all child edges
	std::map<QString, QVariant> attributeMap = std::map<QString, QVariant>();
	std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();
		QString attributeDBName = edge->getAttributeDBName();
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isForeignKeyToDB()) {
			attributeMap[attributeDBName] = attributeData->getForeignKeyToDB();
			context->log("ImportExport", "DataImporter", 1, "    Adding foreign key value " + attributeMap[attributeDBName].toString() +
														  " for attribute " + attributeDBName + " to attributeMap.");
		} else if (attributeData->isForeignKeyToDataTuple()) {
			// a tuple on the database cannot reference a DataTuple in a
			// import file, thus we don't need to search any further here
			context->log("ImportExport", "DataImporter", 1, "    AttributeData for attribute " + attributeDBName + " references a data tuple, "
															"thus there can´t be any matching tuple on the database.");
			return std::vector<QVariant>();
		} else if (attributeData->isValue()) {
			attributeMap[attributeDBName] = attributeData->getValue();
			context->log("ImportExport", "DataImporter", 1, "    Adding value " + attributeMap[attributeDBName].toString() +
														  " for attribute " + attributeDBName + " to attributeMap.");
		} else if (attributeData->isNull()) {
			context->log("ImportExport", "DataImporter", 1, "    Null AttributeData for attribute " + attributeDBName + " doesn´t change the "
															"query condition.");
			// Nothing to do here.  Omitted criteria imply a weaker condition,
			// not an error.
		}
	}

	// then on the attributes
	QString currNodePathKey = node->getPathKey();
	std::vector<const Attribute*> attributes = node->getAttributes();
	for (std::vector<const Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute *attribute = *it;
		// todo introduce constant for .
		QString attributeDBName = attribute->getDBName();
		QString pathKey = currNodePathKey + "." + attributeDBName;
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isValue()) {
			attributeMap[attributeDBName] = attributeData->getValue();
			context->log("ImportExport", "DataImporter", 1, "    Adding value " + attributeMap[attributeDBName].toString() + " for attribute " + attributeDBName);
		} else if (attributeData->isNull()) {
			// Nothing to do here.  Omitted criteria imply a weaker condition,
			// not an error.
		}
	}

	// Then find all primary keys of tuples matching the filter constructed
	// above.
	std::vector<QVariant> primaryKeysOnDatabase
			= ControllerHelper::getPrimaryKeys(getContext(), entityDBName, attributeMap);

	for (std::vector<QVariant>::const_iterator it = primaryKeysOnDatabase.begin(); it != primaryKeysOnDatabase.end(); it++) {
		QString pkString = it->toString();
		context->log("ImportExport", "DataImporter", 2, "  Found database tuple with primary key " + pkString);
	}

	return primaryKeysOnDatabase;
}

std::vector<QVariant> DataImporter::findEqualTuplesOnDatabase(DatabaseContext *context, DataTuple *tuple) {
	QString entityDBName = tuple->getEntityDBName();
	const DataModel *dataModel = context->getDataModel();
	const Entity* entity = dataModel->getEntityByDBName(entityDBName);
	if (entity->hasImpExpKey()) {
		const std::vector<Attribute*> impExpKeyAttributes = entity->getImpExpKey();
		std::map<QString, QVariant> attributeMap = std::map<QString, QVariant>();

		for (std::vector<Attribute*>::const_iterator it2 = impExpKeyAttributes.begin();
				it2 != impExpKeyAttributes.end(); it2++) {
			Attribute *attribute = *it2;
			QString attributeDBName = attribute->getDBName();
			if (tuple->hasAttributeData(attributeDBName)) {
				AttributeData* attributeData = tuple->getAttributeData(attributeDBName);
				if (attributeData->isForeignKeyToDB()) {
					attributeMap[attributeDBName] = attributeData->getForeignKeyToDB();
				} else if (attributeData->isValue()) {
					attributeMap[attributeDBName] = attributeData->getValue();
				} else if (attributeData->isForeignKeyToDataTuple()) {
					QVariant foreignPrimaryKey = attributeData->getForeignKeyToDataTuple()->getPrimaryKey();
					if (foreignPrimaryKey.isValid()) {
						attributeMap[attributeDBName] = foreignPrimaryKey;
					} else {
						QString title = QObject::trUtf8("Interner Fehler");
						QString text = QObject::trUtf8("Suchkriterien fuer Suche auf der Datenbank enthalten Verweis auf Tupel in den Importdaten. \n\n"
													   "Dies ist ein interner Fehler.  Bitte kontaktieren Sie den Administrator.");
						throw ImportException(title, text);
					}
				} else if (attributeData->isNull()) {
					// Nothing to do here.  Omitted criteria imply a weaker condition,
					// not an error.
				}
			} else {
				// Nothing to do here.  Omitted criteria imply a weaker condition,
				// not an error.
			}
		}

		return ControllerHelper::getPrimaryKeys(getContext(), entityDBName, attributeMap);
	} else {
		QString title = trUtf8("Kein Import/Export-Schluessel vorhanden");
		QString text = trUtf8("Fuer die Tabelle %1 ist kein Import/Export-Schluessel vorhanden.  \n\n"
							  "Dies ist hoechstwahrscheinlich ein Fehler im Datenmodell.  Bitte kontaktieren Sie den Administrator.").arg(entityDBName);
		throw ImportException("ImportExport", "DataImporter", title, text);
	}
}

std::vector<DataTuple*> DataImporter::resolveInImportData(DataSet *dataSet, DataTuple *dataTuple, ForeignKeyNode *node) {
	QString entityDBName = node->getEntityDBName();

	// TODO: Do we have to deal with NULL values in a special way?
	context->log("ImportExport", "DataImporter", 2, "  Resolving tuple of type " + entityDBName + " in import data...");

	std::map<QString, DataTuple*> pathKeyToTuple = std::map<QString, DataTuple*>();
	std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();
		QString attributeDBName = edge->getAttributeDBName();
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isForeignKeyToDataTuple()) {
			pathKeyToTuple[attributeDBName] = attributeData->getForeignKeyToDataTuple();
			context->log("ImportExport", "DataImporter", 1, "    Adding tuple for path key " + pathKey + " and attribute db name " + attributeDBName);
		} else {
			// nothing to do here
		}
	}

	std::map<QString, QVariant> pathKeyToForeignKey = std::map<QString, QVariant>();
	for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
		ForeignKeyEdge *edge = *it;

		QString pathKey = edge->getPathKey();
		QString attributeDBName = edge->getAttributeDBName();
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isForeignKeyToDB()) {
			pathKeyToForeignKey[attributeDBName] = attributeData->getForeignKeyToDB();
			context->log("ImportExport", "DataImporter", 1, "    Adding foreign key to db for path key " + pathKey + " and attribute db name " + attributeDBName);
		} else {
			// nothing to do here
		}
	}

	std::map<QString, QVariant> pathKeyToValue = std::map<QString, QVariant>();
	QString currNodePathKey = node->getPathKey();
	std::vector<const Attribute*> attributes = node->getAttributes();
	for (std::vector<const Attribute*>::const_iterator it = attributes.begin(); it != attributes.end(); it++) {
		const Attribute *attribute = *it;
		// todo introduce constant for .
		QString attributeDBName = attribute->getDBName();
		QString pathKey = currNodePathKey + "." + attributeDBName;
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isValue()) {
			pathKeyToValue[attributeDBName] = attributeData->getValue();
			context->log("ImportExport", "DataImporter", 1, "    Adding value " + pathKeyToValue[attributeDBName].toString() + " for path key " + pathKey
																				+ " for attribute db name " + attributeDBName);
		} else {
			// TODO stratey for dealing with missing values, etc.
			// see resolveOnDatabase for details
		}
	}

	std::vector<DataTuple*> foundTuples = dataSet->getTuplesByContent(context, entityDBName, pathKeyToValue,
																	  pathKeyToForeignKey, pathKeyToTuple);
	context->log("ImportExport", "DataImporter", 2, "  Found " + QString::number(foundTuples.size()) + " tuples in import data.");
	return foundTuples;
}

std::vector<DataTuple*> DataImporter::findEqualTuplesInImportData(DataSet *dataSet, DataTuple *tuple) {
	QString entityDBName = tuple->getEntityDBName();
	const DataModel *dataModel = context->getDataModel();
	const Entity* entity = dataModel->getEntityByDBName(entityDBName);

	std::vector<DataTuple*> foundTuples = std::vector<DataTuple*>();

	std::vector<DataTuple*> tupleVector = dataSet->getTuples(entityDBName);
	for (std::vector<DataTuple*>::const_iterator it = tupleVector.begin(); it != tupleVector.end(); it++) {
		DataTuple *testTuple = *it;

		bool tupleMatches = true;

		if (entity->hasImpExpKey()) {
			const std::vector<Attribute*> impExpKeyAttributes = entity->getImpExpKey();

			for (std::vector<Attribute*>::const_iterator it2 = impExpKeyAttributes.begin();
					it2 != impExpKeyAttributes.end(); it2++) {
				Attribute *attribute = *it2;
				QString attributeDBName = attribute->getDBName();
				if (tuple->hasAttributeData(attributeDBName) != testTuple->hasAttributeData(attributeDBName)) {
					tupleMatches = false;
					break;
				} else if (tuple->hasAttributeData(attributeDBName) && testTuple->hasAttributeData(attributeDBName)) {
					AttributeData *attributeData = tuple->getAttributeData(attributeDBName);
					AttributeData *testAttributeData = testTuple->getAttributeData(attributeDBName);
					if (attributeData->isValue() == true && testAttributeData->isValue() == true) {
						tupleMatches &= (attributeData->getValue() == testAttributeData->getValue());
					} else if (attributeData->isForeignKeyToDB() == true && testAttributeData->isForeignKeyToDB() == true) {
						tupleMatches &= (attributeData->getForeignKeyToDB() == testAttributeData->getForeignKeyToDB());
					} else if (attributeData->isForeignKeyToDataTuple() == true && testAttributeData->isForeignKeyToDataTuple() == true) {
						// equality means equality of pointers here
						tupleMatches &= (attributeData->getForeignKeyToDataTuple() == testAttributeData->getForeignKeyToDataTuple());
					} else {
						tupleMatches = false;
						break;
					}
				}
			}
		} else {
			QString title = trUtf8("Kein Import/Export-Schluessel vorhanden");
			QString text = trUtf8("Fuer die Tabelle %1 ist kein Import/Export-Schluessel vorhanden.  \n\n"
								  "Dies ist hoechstwahrscheinlich ein Fehler im Datenmodell.  Bitte kontaktieren Sie den Administrator.").arg(entityDBName);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}

		if (tupleMatches) {
			foundTuples.push_back(testTuple);
		}
	}
	return foundTuples;
}

	/** Processes the given node:
	 *  - If it is finished with respect to the given tuple, and is the root node,
	 *    simply return the root node for marking the given tuple finished.
	 *  - If it is finished, and is not the root node, first try to resolve a
	 *    foreign key reference or a foreign tuple using the information given
	 *    for the node in the DataTuple.  If
	 *    - two or more references/tuples could be resolved, resolving that
	 *      foreign key is not possible at all, thus throw an appropriate
	 *      ImportException
	 *    - exactly one reference/tuple could be resolved, exchange the data in
	 *      the given DataTuple accordingly (i.e. remove keys for children and
	 *      add one for the newly resolved foreign references) and return the
	 *      node.
	 *    - no reference/tuple could be resolved, return the empty vector.  Note
	 *      that this does not mean that the foreign key can't be resolved at some
	 *      time in the future, as there might be a tuple as foreign key destination
	 *      that will be resolved at some time afterwards.
	 * - If it is not yet finished, then there is at least some child node.  Inspect
	 *   all child nodes, calling this method recursively, and return the
	 *   concatenation of the results of the recursive calls.
	 *
	 * Thus, if the return value is
	 * - the empty vector, we know that there was no progress regarding the foreign
	 *   key graph at hand.  This includes the case that the function is called for
	 *   the root node, and the root node is already finished.
	 * - some other non-empty vector, we know that there was some progress regarding
	 *   resolving this foreign key reference, although we are not yet finished.
	 */
std::vector<ForeignKeyNode*> DataImporter::processNode(DataSet *dataSet, DataTuple *dataTuple,
													   ForeignKeyNode *node, bool isRootNode) {

	context->log("ImportExport", "DataImporter", 3, "        processNode started for node " + node->getPathKey());
	if (isNodeFinished(node, dataTuple)) {
		if (isRootNode) {
			context->log("ImportExport", "DataImporter", 3, "        processNode considers the root node for entity " 
						  + node->getEntityDBName() + " finished");
			// return the empty vector, as there is no real progress in this
			// case, we just note that we don't have to do anything here
			std::vector<ForeignKeyNode*> processedNodes = std::vector<ForeignKeyNode*>();
			return processedNodes;
		} else {
			context->log("ImportExport", "DataImporter", 3,
						 "        processNode considers node " + node->getPathKey() + " finished, calls resolveNode");
			return resolveNode(dataSet, dataTuple, node);
		}
	} else {
		std::vector<ForeignKeyNode*> processedNodes = std::vector<ForeignKeyNode*>();
		std::vector<ForeignKeyEdge*> childEdges = node->getChildEdges();
		for (std::vector<ForeignKeyEdge*>::const_iterator it = childEdges.begin(); it != childEdges.end(); it++) {
			ForeignKeyEdge *edge = *it;
			ForeignKeyNode *outNode = edge->getOutNode();
			QString edgePathKey = edge->getPathKey();
			context->log("ImportExport", "DataImporter", 3, "        processNode considers recursive call for edge " + edgePathKey + " to node " + outNode->getPathKey());

			if (!dataTuple->hasAttributeData(edgePathKey)) {
				// There are some edges finished, but not all. E.g. consider the following path keys 
				// stored in a DataTuple:
				// - durchgang.art
				// - durchgang.jurierung.wettbewerb
				// Because of the second one, the node durchgang is not finished.  However, node art is,
				// and thus we don´t have to do anything about it here, but have to wait until node
				// jurierung is finished.
				// The root node is an exception: It always has exactly one out edge, with the db name 
				// of the corresponding attribute.
				std::vector<ForeignKeyNode*> currNodes = processNode(dataSet, dataTuple, outNode, false);
				for (std::vector<ForeignKeyNode*>::const_iterator it2 = currNodes.begin(); 
					it2 != currNodes.end(); it2++) {
					processedNodes.push_back(*it2);
				}
			}
		}
		return processedNodes;
	}
}

void DataImporter::checkAndConvertRawTypes(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting checkAndConvertRawTypes");
	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;
		checkAndConvertRawTypes(dataSet, tuple);
	}
	context->log("ImportExport", "DataImporter", 7, "Finishing checkAndConvertRawTypes");
}

void DataImporter::checkAndConvertRawTypes(DataSet *dataSet, DataTuple *dataTuple) {
	QString entityDBName = dataTuple->getEntityDBName();
	std::set<QString> pathKeys = dataTuple->getAllPathKeys();
	for (std::set<QString>::const_iterator it = pathKeys.begin(); it != pathKeys.end(); it++) {
		QString pathKey = *it;
		QString dbType = ImpExpSchema::getDBType(context, entityDBName, pathKey);
		AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
		if (attributeData->isValue()) {
			QVariant value = attributeData->getValue();
			// Possibility: Set converted type here.  E.g. set a qlonglong as qlonglong instead of its QString representation here.
			// Currently, this is not done because the other variant is known to work.
			QVariant convertedVariant = checkAndConvertRawType(dataTuple, pathKey, dbType, value);
			dataTuple->setValue(pathKey, convertedVariant);
		}
	}
}

QVariant DataImporter::checkAndConvertRawType(DataTuple *dataTuple, QString pathKey, QString expectedType, QVariant variant) {
	if (!variant.isValid()) {
		return variant;
	}

	if (expectedType == "date") {
		QString value = variant.toString();
		QDate date = QDate::fromString(value, GlobaleWidgetEigenschaften::getDateFormat());
		if (date.isValid()) {
			return QVariant(date);
		} else {
			// Alternative date format supported, especially because in version 0.12, due to a bug
			// dates were exported in that format.
			QDate dateAlternative = QDate::fromString(value, "yyyy-M-d");
			if (dateAlternative.isValid()) {
				return QVariant(dateAlternative);
			} else {
				QString title = trUtf8("Datum falsch angegeben");
				QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird ein Datumswert erwartet.  "
						"Datumswerte werden im Format dd.mm.yyyy erwartet, wobei fuehrende Nullen weggelassen werden duerfen.\n\n"
						"Der angegebene Text %3 kann daher nicht als Datum erkannt werden.  Bitte korrigieren Sie ihn.")
						.arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
				throw ImportException("ImportExport", "DataImporter", title, text);
			}
		}
	} else if (expectedType == "timestamp") {
		QString value = variant.toString();
		QDateTime datetime = QDateTime::fromString(value, GlobaleWidgetEigenschaften::getDateTimeFormat());
		if (datetime.isValid()) {
			return QVariant(datetime);
		} else {
			QString title = trUtf8("Zeitstempel falsch angegeben");
			QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird ein Zeitstempel erwartet.  "
								  "Zeitstempel werden im Format Datumswerte werden im Format dd.mm.yyyy hh:mm:ss erwartet, wobei fuehrende Nullen beim Datum weggelassen werden duerfen.\n\n"
								  "Der angegebene Text \"%3\" kann daher nicht als Zeitstempel erkannt werden.  Bitte korrigieren Sie ihn.")
								  .arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}
	} else if (expectedType == "int") {
		QString value = variant.toString();
		bool ok;
		int number = value.toInt(&ok);
		if (ok) {
			return QVariant(number);
		} else {
			QString title = trUtf8("Zahl falsch angegeben");
			QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird eine Ganzzahl ohne Nachkommastellen erwartet. \n\n"
								  "Der angegebene Text \"%3\" kann nicht als solche erkannt werden.  Bitte korrigieren Sie ihn.")
								  .arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}
	} else if (expectedType == "bigint") {
		QString value = variant.toString();
		bool ok;
		qlonglong number = value.toLong(&ok);
		if (ok) {
			return QVariant(number);
		} else {
			QString title = trUtf8("Zahl falsch angegeben");
			QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird eine Ganzzahl ohne Nachkommastellen erwartet.\n\n"
								  "Der angegebene Text \"%3\" kann nicht als solche erkannt werden.  Bitte korrigieren Sie ihn.")
								  .arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}
	} else if (expectedType == "double") {
		QString value = variant.toString();
		bool ok;
		double number = value.toDouble(&ok);
		if (ok) {
			return QVariant(number);
		} else {
			QString title = trUtf8("Zahl falsch angegeben");
			QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird eine Fliesskommazahl (mit oder ohne Nachkommastellen) erwartet.\n\n"
								  "Der angegebene Text \"%3\" kann nicht als solche erkannt werden.  Bitte korrigieren Sie ihn.")
								  .arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}
	} else if (expectedType == "bool") {
		QString value = variant.toString();
		if (value == "true") {
			return QVariant(true);
		} else if (value == "false") {
			return QVariant(false);
		} else {
			QString title = trUtf8("Wahrheitswert falsch angegeben");
			QString text = trUtf8("Fuer den Pfad %1 der Tabelle %2 wird ein Wahrheitswert erwartet.  Der Text \"true\" steht dabei fuer "
								  "\"wahr\", der Text \"false\" fuer falsch.  Beispiel: Bei einem inaktiven Verein sollte unter \"inaktiv\""
								  "der Text \"true\" stehen, bei einem aktiven der Text \"false\".\n\n"
								  "Der angegebene Text \"%3\" kann nicht als Wahrheitswert erkannt werden.  Bitte korrigieren Sie ihn.")
								  .arg(pathKey).arg(dataTuple->getEntityDBName()).arg(value);
			throw ImportException("ImportExport", "DataImporter", title, text);
		}
	}

	return variant;
}

void DataImporter::checkAndConvertBeforeResolve(DataSet *dataSet) {
	dataSet->printData(context);

	context->log("ImportExport", "DataImporter", 7, "Starting checkAndConvertBeforeResolve");
	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *dataTuple = *it;
		QString entityDBName = dataTuple->getEntityDBName();
		ImportStrategy *importStrategy = getImportStrategy(entityDBName);
		importStrategy->checkAndConvertBeforeResolve(getContext(), dataSet, dataTuple, importResult);
	}

	// ImportStrategies may delete tuples, thus we have to use a second loop here
	tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *dataTuple = *it;
		QString entityDBName = dataTuple->getEntityDBName();

		ImpExpSchema *schema = dataSet->getSchema(entityDBName);
		if (schema == NULL) {
			throw ImportException("ImportExport", "DataImporter", trUtf8("Interner Fehler"), trUtf8("Kein Schema fuer Entitaet %1 im DataSet").arg(entityDBName));
		}
		std::multimap<QString, QStringList> attributeDBNameToPathKeys = schema->getAttributePaths();

		std::set<QString> knownPathKeys = std::set<QString>();
		for (std::multimap<QString, QStringList>::const_iterator it = attributeDBNameToPathKeys.begin(); it != attributeDBNameToPathKeys.end(); it++) {
			QStringList currPathKey = it->second;
			QString currPathKeyString = "";
			for (int n = 0; n < currPathKey.size(); n++) {
				currPathKeyString += currPathKey.at(n);
				currPathKeyString += (n < currPathKey.size() - 1 ? "." : "");
			}
			knownPathKeys.insert(currPathKeyString);
		}

		std::set<QString> givenPathKeys = dataTuple->getAllPathKeys();
		for (std::set<QString>::const_iterator it = givenPathKeys.begin(); it != givenPathKeys.end(); it++) {
			QString givenPathKey = *it;
			if (knownPathKeys.find(givenPathKey) == knownPathKeys.end()) {
				QString shortMessage = trUtf8("Unbekannter Pfad gefunden");
				QString longMessage = trUtf8("In der Importdatei ist ein Pfad %1 fuer Datensaetze der Tabelle %2 angegeben.  Dieser ist laut Datenmodell nicht erlaubt. "
											 "Bitte beheben Sie den Fehler, oder kontaktieren Sie den Administrator.").arg(givenPathKey).arg(entityDBName);
				throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
			}
		}
	}

	context->log("ImportExport", "DataImporter", 7, "Finishing checkAndConvertBeforeResolve");
}

void DataImporter::postProcessAfterWrite(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting postProcessAfterWrite(DataSet*)");
	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;
		postProcessAfterWrite(dataSet, tuple);
	}
	context->log("ImportExport", "DataImporter", 7, "Finishing postProcessAfterWrite(DataSet*)");
}

void DataImporter::postProcessAfterWrite(DataSet *dataSet, DataTuple *dataTuple) {
	context->log("ImportExport", "DataImporter", 3, "Starting postProcessAfterWrite(DataTuple*)");

	QString entityDBName = dataTuple->getEntityDBName();
	ImportStrategy *importStrategy = getImportStrategy(entityDBName);
	importStrategy->postProcessAfterWrite(getContext(), dataSet, dataTuple, importResult);

	context->log("ImportExport", "DataImporter", 3, "Finishing postProcessAfterWrite(DataTuple*)");
}

void DataImporter::mapToDatabaseTuples(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting mapToDatabaseTuples(DataSet*)");

	std::vector<DataTuple*> newUnfinishedTuples = dataSet->getAllTuples();

	// Erstmal alle Tupel auf nicht abgebildet setzen.  Zu Beginn es ersten Durchlaufs der
	// Schleife unten wird unfinishedTuples.size() == 0 sein, un newUnfinishedTuples.size() >= 0.
	// Falls newUnfinishedTuples.size() == 0, gibt es nichts zu tun, und es ist völlig ok keinen
	// Schleifendurchlauf zu machen.
	for (std::vector<DataTuple*>::const_iterator it = newUnfinishedTuples.begin(); it != newUnfinishedTuples.end(); it++) {
		DataTuple *tuple = *it;
		tuple->setPrimaryKey(QVariant());
	}

	int stepNumber = 1;
	std::vector<DataTuple*> unfinishedTuples = std::vector<DataTuple*>();
	while (unfinishedTuples.size() != newUnfinishedTuples.size()) {
		context->log("ImportExport", "DataImporter", 5, "Step " + QString::number(stepNumber) + ", unfinished tuples last time: "
														+ QString::number(unfinishedTuples.size()) + ", this time: " + QString::number(newUnfinishedTuples.size()));

		// Falls in einem Durchlauf alle Tupel wieder in newUnfinishedTuples eingefuegt werden,
		// herrscht hier Gleichstand und wir sind fertig.

		unfinishedTuples = std::vector<DataTuple*>();
		for (std::vector<DataTuple*>::const_iterator it = newUnfinishedTuples.begin(); it != newUnfinishedTuples.end(); it++) {
			unfinishedTuples.push_back(*it);
		}

		int loopNumber = 1;

	    newUnfinishedTuples = std::vector<DataTuple*>();
		for (std::vector<DataTuple*>::const_iterator it = unfinishedTuples.begin(); it != unfinishedTuples.end(); it++) {
			// ueber alle noch nicht fertigen Tupel iterieren
			DataTuple *tuple = *it;

			QString entityDBName = tuple->getEntityDBName();
			ImportStrategy *importStrategy = getImportStrategy(entityDBName);
			bool insertAlways = importStrategy->insertAlways(context, dataSet, tuple, importResult);

			context->log("ImportExport", "DataImporter", 3, "Verarbeite Tupel " + QString::number(loopNumber) + " vom Typ " + entityDBName + ", insertAlways = " + QString::number(insertAlways));

			if (!insertAlways) {
				// Nicht automatisch immer einfuegen
				if (containsReferencesToUnMappedDataTuples(tuple)) {
					context->log("ImportExport", "DataImporter", 3, "Tupel enthaelt Referenzen zu noch nicht verarbeiteten Tupeln, stelle es zurueck.");
					// Wir referenzieren DataTuple, fuer die noch kein passendes Tupel auf der Datenbank gefunden wurde.
					// Entweder weil es keines gibt (dann bleibt das Tupel am Ende der Fixpunktiteration uebrig), oder
					// weil wir es spaeter finden werden. Beispiel: Mitglied, sowohl in den Import- wie in den DB-Daten enthalten,
					// referenziert CodeValue, ebenfalls sowohl in den Importdaten wie auch in der DB enthalten.  In dem
					// Beispiel muss das CodeValue zuerst abgebildet werden.
					newUnfinishedTuples.push_back(tuple);
				} else {
					std::vector<QVariant> primaryKeys = findEqualTuplesOnDatabase(getContext(), tuple);
					if (primaryKeys.size() == 0) {
						context->log("ImportExport", "DataImporter", 3, "Kein passendes Tupel auf der Datenbank gefunden, daher wird es neu eingefuegt.");

						// Kein passendes Tupel
						tuple->setPrimaryKey(QVariant());
					} else if (primaryKeys.size() == 1) {
						context->log("ImportExport", "DataImporter", 3, "Genau ein passendes Tupel auf der Datenbank gefunden, id " + primaryKeys[0].toString()  + ", ueberschreibe es.");

						tuple->setPrimaryKey(primaryKeys[0]);
					} else {
						QString ids = "";
						for (unsigned int n = 0; n < primaryKeys.size() && n < 10; n++) {
							ids += (n == 0 ? "" : ", ") + primaryKeys[n].toString();
						}
						context->log("ImportExport", "DataImporter", 7, "Mehr als ein passendes Tupel auf der Datenbank gefunden, ids " + ids);

						QString shortMessage = trUtf8("Datensatz nicht eindeutig");
						QString longMessage = trUtf8("Es gibt mehrere Datensaetze auf der Datenbank, "
							 "die in Bezug auf den Importschluessel gleich sind.  Im einzelnen sind die "
							 "Datensaetze mit den IDs %1 in der Tabelle %2 problematisch (maximal 10 werden ausgegeben). " 
							 "Daher ist keine "
							 "Entscheidung moeglich, welcher dieser Datensaetze ueberschrieben werden muss.\n\n"
							 "Es koennte sich hierbei um eine Dateninkonsistenz handeln.  Bitte "
							 "kontaktieren Sie den Administrator.").arg(ids).arg(entityDBName);
						throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
					}
				}
			} else {
				tuple->setPrimaryKey(QVariant());
			}

			if (loopNumber % 50 == 0) {
				emit statusChanged(trUtf8("Abbildung auf Datenbank bei %1 Datensaetzen hergestellt").arg(loopNumber));
			}
			loopNumber++;
		}
		stepNumber++;
	}

	context->log("ImportExport", "DataImporter", 7, "Finishing mapToDatabaseTuples(DataSet*)");
}

bool DataImporter::containsReferencesToUnMappedDataTuples(DataTuple *tuple) {
	QString entityDBName = tuple->getEntityDBName();
	const DataModel *dataModel = context->getDataModel();
	const Entity* entity = dataModel->getEntityByDBName(entityDBName);
	if (entity->hasImpExpKey()) {
		const std::vector<Attribute*> impExpKeyAttributes = entity->getImpExpKey();
		std::map<QString, QVariant> attributeMap = std::map<QString, QVariant>();

		for (std::vector<Attribute*>::const_iterator it2 = impExpKeyAttributes.begin();
				it2 != impExpKeyAttributes.end(); it2++) {
			Attribute *attribute = *it2;
			QString attributeDBName = attribute->getDBName();
			if (tuple->hasAttributeData(attributeDBName)) {
				AttributeData* attributeData = tuple->getAttributeData(attributeDBName);
				if (attributeData->isForeignKeyToDataTuple()) {
					DataTuple *foreignTuple = attributeData->getForeignKeyToDataTuple();
					QVariant foreignPrimaryKey = foreignTuple->getPrimaryKey();
					if (!foreignPrimaryKey.isValid()) {
						return true;
					}
				}
			}
		}

		return false;
	} else {
		QString title = trUtf8("Kein Import/Export-Schluessel vorhanden");
		QString text = trUtf8("Fuer die Tabelle %1 ist kein Import/Export-Schluessel vorhanden.  \n\n"
							  "Dies ist hoechstwahrscheinlich ein Fehler im Datenmodell.  Bitte kontaktieren Sie den Administrator.").arg(entityDBName);
		throw ImportException("ImportExport", "DataImporter", title, text);
	}
}

void DataImporter::checkForDuplicatesInImportData(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting checkForDuplicatesInImportData(DataSet*)");

	std::vector<DataTuple*> newUnfinishedTuples = dataSet->getAllTuples();

	// Erstmal alle Tupel auf nicht fertig setzen
	for (std::vector<DataTuple*>::const_iterator it = newUnfinishedTuples.begin(); it != newUnfinishedTuples.end(); it++) {
		DataTuple *tuple = *it;
		tuple->setDuplicateCheckPassed(false);
	}

	int stepNumber = 1;
	std::vector<DataTuple*> unfinishedTuples = std::vector<DataTuple*>();
	while (unfinishedTuples.size() != newUnfinishedTuples.size()) {
		context->log("ImportExport", "DataImporter", 5, "Step " + QString::number(stepNumber) + ", unfinished tuples last time: "
														+ QString::number(unfinishedTuples.size()) + ", this time: " + QString::number(newUnfinishedTuples.size()));

		// Falls in einem Durchlauf alle Tupel wieder in newUnfinishedTuples eingefuegt werden,
		// herrscht hier Gleichstand und wir sind fertig.

		unfinishedTuples = std::vector<DataTuple*>();
		for (std::vector<DataTuple*>::const_iterator it = newUnfinishedTuples.begin(); it != newUnfinishedTuples.end(); it++) {
			unfinishedTuples.push_back(*it);
		}
		int loopNumber = 1;

	    newUnfinishedTuples = std::vector<DataTuple*>();
		for (std::vector<DataTuple*>::const_iterator it = unfinishedTuples.begin(); it != unfinishedTuples.end(); it++) {
			// ueber alle noch nicht fertigen Tupel iterieren
			DataTuple *tuple = *it;

			QString entityDBName = tuple->getEntityDBName();
			ImportStrategy *importStrategy = getImportStrategy(entityDBName);
			bool insertAlways = importStrategy->insertAlways(context, dataSet, tuple, importResult);
			context->log("ImportExport", "DataImporter", 3, "Verarbeite Tupel " + QString::number(loopNumber) + " vom Typ " + entityDBName + ", insertAlways = " + QString::number(insertAlways));

			if (!insertAlways) {
				// Nicht automatisch immer einfuegen
				if (containsReferencesToUnfinishedDataTuples(tuple)) {
					context->log("ImportExport", "DataImporter", 3, "Tupel enthaelt Referenzen zu noch nicht geprueften Tupeln, stelle es zurueck.");

					// Wir referenzieren DataTuple, fuer die noch kein passendes Tupel auf der Datenbank gefunden wurde.
					// Entweder weil es keines gibt (dann bleibt das Tupel am Ende der Fixpunktiteration uebrig), oder
					// weil wir es spaeter finden werden. Beispiel: Mitglied, sowohl in den Import- wie in den DB-Daten enthalten,
					// referenziert CodeValue, ebenfalls sowohl in den Importdaten wie auch in der DB enthalten.  In dem
					// Beispiel muss das CodeValue zuerst abgebildet werden.
					newUnfinishedTuples.push_back(tuple);
				} else {
					std::vector<DataTuple*> equalTuples = findEqualTuplesInImportData(dataSet, tuple);
					if (equalTuples.size() >= 2) {
						context->log("ImportExport", "DataImporter", 7, "Es wurden " + QString::number(equalTuples.size()) + " Tupel gefunden.");

						// Das Tupel selber sollte auch gefunden werden

						std::map<QString, QString> originalData = equalTuples[0]->getOriginalPathKeyToData();
						QString originalEntityDBName = equalTuples[0]->getEntityDBName();
						QString originalContent = "";
						for (std::map<QString, QString>::const_iterator it2 = originalData.begin(); it2 != originalData.end(); it2++) {
							originalContent += it2->first + " = " + it2->second + "\n";
						}

						QString shortMessage = trUtf8("Datensatz nicht eindeutig");
						QString longMessage = trUtf8("Es gibt mehrere Datensaetze in dem Importdaten, "
							 "die in Bezug auf den Importschluessel gleich sind.  Einer davon gehoert zur Tabelle %1 und hat den folgenden Inhalt: \n\n"
							 "%2\n"
							 "Daher ist keine "
							 "Entscheidung moeglich, welcher dieser Datensaetze gelten soll. \n\n"
							 "Bitte korrigieren Sie die Importdaten oder kontaktieren Sie den Administrator.").arg(originalEntityDBName, originalContent);
						throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
					} else {
						tuple->setDuplicateCheckPassed(true);
					}
				}
			}

			if (loopNumber % 50 == 0) {
				emit statusChanged(trUtf8("Duplikaterkennung auf %1 Datensaetzen durchgefuehrt").arg(loopNumber));
			}
			loopNumber++;
		}
		stepNumber++;
	}

	context->log("ImportExport", "DataImporter", 7, "Finishing checkForDuplicatesInImportData(DataSet*)");
}

bool DataImporter::containsReferencesToUnfinishedDataTuples(DataTuple *tuple) {
	QString entityDBName = tuple->getEntityDBName();
	const DataModel *dataModel = context->getDataModel();
	const Entity* entity = dataModel->getEntityByDBName(entityDBName);
	if (entity->hasImpExpKey()) {
		const std::vector<Attribute*> impExpKeyAttributes = entity->getImpExpKey();
		std::map<QString, QVariant> attributeMap = std::map<QString, QVariant>();

		for (std::vector<Attribute*>::const_iterator it2 = impExpKeyAttributes.begin();
				it2 != impExpKeyAttributes.end(); it2++) {
			Attribute *attribute = *it2;
			QString attributeDBName = attribute->getDBName();
			if (tuple->hasAttributeData(attributeDBName)) {
				AttributeData* attributeData = tuple->getAttributeData(attributeDBName);
				if (attributeData->isForeignKeyToDataTuple()) {
					DataTuple *foreignTuple = attributeData->getForeignKeyToDataTuple();
					if (!foreignTuple->isDuplicateCheckPassed()) {
						return true;
					}
				}
			}
		}

		return false;
	} else {
		QString title = trUtf8("Kein Import/Export-Schluessel vorhanden");
		QString text = trUtf8("Fuer die Tabelle %1 ist kein Import/Export-Schluessel vorhanden.  \n\n"
							  "Dies ist hoechstwahrscheinlich ein Fehler im Datenmodell.  Bitte kontaktieren Sie den Administrator.").arg(entityDBName);
		throw ImportException("ImportExport", "DataImporter", title, text);
	}
}

void DataImporter::writeToDatabase(DataSet *dataSet, QFile *importFile, ImportResult *importResult) {
	context->log("ImportExport", "DataImporter", 7, "Starting writeToDatabase");

	DatabaseContext *context = getContext();
	const DataModel *dataModel = context->getDataModel();
	// If any CodeTypes or CodeValue instances are inserted or updated by the
	// import, we have to reload the CodeValues cached in the DatabaseContext
	bool codesChanged = false;

	int loopNumber = 1;

	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;

		QString tupleAddress = QString::number((qlonglong)tuple, 16);
		context->log("ImportExport", "DataImporter", 6, "Processing tuple " + tupleAddress);

		loadBLOBs(tuple, importFile);

		QString entityDBName = tuple->getEntityDBName();
		if (entityDBName == "CodeValue" || entityDBName == "CodeType") {
			codesChanged = true;
		}
		const Entity* entity = dataModel->getEntityByDBName(entityDBName);
		ImportStrategy *importStrategy = getImportStrategy(entityDBName);

		QString pkDBName;
		if (!entity->hasPrimaryKeyAttribute()) {
			QString shortMessage = trUtf8("Kein Primaerschluessel definiert");
			QString longMessage = trUtf8("Es ist kein Primaerschluessel vorhanden. "
										 "Bitte kontaktieren Sie den Administrator.");
			throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
		} else {
			const Attribute* primaryKeyAttribute = entity->getPrimaryKeyAttribute();
			pkDBName = primaryKeyAttribute->getDBName();
		}

		// Decide wether we overwrite or insert.  If the import strategy says importAlways,
		// the primary key stored in the tuple will be an invalid QVariant(), thus we always
		// run into the setNull case here.
		QVariant pk = tuple->getPrimaryKey();
		if (pk.isValid()) {
			context->log("ImportExport", "DataImporter", 5,
						  "  Overwriting tuple " + pk.toString());
			tuple->setValue(pkDBName, tuple->getPrimaryKey());
		} else {
			context->log("ImportExport", "DataImporter", 5, "  Inserting new tuple");
			tuple->setNull(pkDBName);
		}

		// Insert/update tuple in the database.
		// This means, what now is missing is transforming
		// the foreign key to DataTuple references into foreign key
		// to db references and updating them in the database.
		ControllerHelper::insertOrUpdate(getContext(), tuple);
		importResult->instanceImported(entityDBName);

		removeBLOBDataFromTuple(tuple);

		if (loopNumber % 50 == 0) {
			emit statusChanged(trUtf8("%1 Datensaetze geschrieben").arg(loopNumber));
		}
		loopNumber++;
	}

	if (codesChanged) {
		context->log("ImportExport", "DataImporter", 7, "Codes changed, so reloading cached CodeValues in DatabaseContext "
													  "NOTE: Old CodeValues get deleted here.  So if they where shared "
													  "somewhere (currently no such case known), "
													  "this might lead to segfaults");
		context->initCodeManager();
	}
	context->log("ImportExport", "DataImporter", 7, "Finishing writeToDatabase");
}

void DataImporter::checkBLOBs(DataSet *dataSet, QFile *importFile) {
	context->log("ImportExport", "DataImporter", 7, "Starting checkBLOBs");

	QDir importDir = QFileInfo(*importFile).dir();

	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *dataTuple = *it;

		DatabaseContext *context = getContext();
		const DataModel *dataModel = context->getDataModel();
		QString entityDBName = dataTuple->getEntityDBName();
		const Entity *entity = dataModel->getEntityByDBName(entityDBName);

		std::set<QString> attributeDBNames = dataTuple->getAllPathKeys();
		for (std::set<QString>::const_iterator it = attributeDBNames.begin(); it != attributeDBNames.end(); it++) {
			QString attributeDBName = *it;
			const Attribute *attribute = entity->getAttributeByDBName(attributeDBName);
			// BLOB attributes never have a path key like a.b  For attributes with path keys that are not yet resolved
			// here, the corresponding Attribute will be NULL here.
			if (attribute != NULL && attribute->getDBType() == "longblob") {
				context->log("ImportExport", "DataImporter", 3, "checkBLOBs: Inspecting BLOB attribute " + attributeDBName);

				AttributeData *attributeData = dataTuple->getAttributeData(attributeDBName);
				if (attributeData->isNull()) {
					// do nothing
				} else if (!attributeData->isValue()) {
					QString shortMessage = trUtf8("Technischer Fehler");
					QString longMessage = trUtf8("Binaerdaten werden nicht aus der Importdatei selbst, "
							  "sondern aus separaten Dateien importiert.  In der Importdatei befinden sich "
							  "nur Verweise auf die zu importierenden separaten Binaerdateien. \n\n"
							  "Das Attribut %1 der Tabelle %2 scheint einen hierfuer nicht erlaubten Pfad zu haben. \n\n"
							  "Bitte kontaktieren Sie den Administrator.").arg(attributeDBName).arg(entityDBName);
					throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
				} else {
					QVariant value = attributeData->getValue();
					QString fileName = value.toString();
					fileName = ArchiveHelper::convertFileNameForOutput(fileName);
					QString blobFilePath = importDir.absoluteFilePath(fileName);

					QFile *file = new QFile(blobFilePath);
					if (!file->open(QIODevice::ReadOnly)) {
						QString shortMessage = trUtf8("Kann Datei nicht oeffnen");
						QString longMessage = trUtf8("Binaerdaten werden nicht aus der Importdatei selbst, "
							  "sondern aus separaten Dateien importiert.  In der Importdatei befinden sich "
							  "nur Verweise auf die zu importierenden separaten Binaerdateien. \n\n"
							  "Die Binaerdatei %1 ist entweder nicht vorhanden, oder konnte nicht geoeffnet werden.").arg(blobFilePath);
						delete file;
						throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
					} else {
						delete file;
					}
				}
			}
		}
	}

	context->log("ImportExport", "DataImporter", 7, "Finishing checkBLOBs");
}

void DataImporter::loadBLOBs(DataTuple *dataTuple, QFile *importFile) {
	context->log("ImportExport", "DataImporter", 3, "Starting loadBLOBs");

	QDir importDir = QFileInfo(*importFile).dir();

	DatabaseContext *context = getContext();
	const DataModel *dataModel = context->getDataModel();
	QString entityDBName = dataTuple->getEntityDBName();
	const Entity *entity = dataModel->getEntityByDBName(entityDBName);

	context->log("ImportExport", "DataImporter", 3, "  Inspecting tuple of type " + entityDBName);

	// In this step, all foreign references were resolved before, i.e. all path
	// keys have length 1.  This means, all path keys in fact are attribute db
	// names.
	std::set<QString> attributeDBNames = dataTuple->getAllPathKeys();
	for (std::set<QString>::const_iterator it = attributeDBNames.begin(); it != attributeDBNames.end(); it++) {
		QString attributeDBName = *it;
		context->log("ImportExport", "DataImporter", 3, "  Inspecting attribute " + attributeDBName);
		const Attribute *attribute = entity->getAttributeByDBName(attributeDBName);
		if (attribute->getDBType() == "longblob") {
//		if (attribute->getDBType().toLowerCase() == "longblob") {
			// We have a BLOB value we must load from an extra file
			AttributeData *attributeData = dataTuple->getAttributeData(attributeDBName);
			if (attributeData->isNull()) {
				// do nothing
			} else if (!attributeData->isValue()) {
				QString shortMessage = trUtf8("Technischer Fehler");
				QString longMessage = trUtf8("Das AttributeData zum Attribut %1 von "
				  "Tabelle %2 ist kein Value.  \n\n Dies ist ein technischer Fehler. "
				  "Bitte kontaktieren Sie den Administrator.").arg(attributeDBName).arg(entityDBName);
				throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
			} else {
				QVariant value = attributeData->getValue();
				QString fileName = value.toString();
				fileName = ArchiveHelper::convertFileNameForOutput(fileName);
				QString blobFilePath = importDir.absoluteFilePath(fileName);

				QFile *file = new QFile(blobFilePath);
				if (!file->open(QIODevice::ReadOnly)) {
					QString shortMessage = trUtf8("Kann Datei nicht oeffnen");
					QString longMessage = trUtf8("Binaerdaten werden nicht aus der Importdatei selbst, "
						  "sondern aus separaten Dateien importiert.  In der Importdatei befinden sich "
						  "nur Verweise auf die zu importierenden separaten Binaerdateien. \n\n"
						  "Die Binaerdatei %1 konnte leider nicht geoeffnet werden.").arg(blobFilePath);
					delete file;
					throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
				} else {
					QByteArray blobData = file->readAll();
					if (blobData.size() == 0) {
						QString shortMessage = trUtf8("Fehler beim Lesen der Datei");
						QString longMessage = trUtf8("Binaerdaten werden nicht aus der Importdatei selbst, "
							  "sondern aus separaten Dateien importiert.  In der Importdatei befinden sich "
							  "nur Verweise auf die zu importierenden separaten Binaerdateien. \n\n"
							  "Die Binaerdatei %1 ist entweder leer, oder konnte aus einem anderen Grund nicht "
							  "gelesen werden.").arg(blobFilePath);
						throw ImportException("ImportExport", "DataImporter", shortMessage, longMessage);
					}
					attributeData->setValue(QVariant(blobData));
					delete file;
//					context->log("Import", "DataImporter.cpp", 4, "Loading BLOB data of size " + blobData.size());
				}
			}
		}
	}
	context->log("ImportExport", "DataImporter", 3, "Finishing loadBLOBs");
}

void DataImporter::removeBLOBDataFromTuple(DataTuple *dataTuple) {
	DatabaseContext *context = getContext();
	const DataModel *dataModel = context->getDataModel();
	QString entityDBName = dataTuple->getEntityDBName();
	const Entity *entity = dataModel->getEntityByDBName(entityDBName);

	// In this step, all foreign references were resolved before, i.e. all path
	// keys have length 1.  This means, all path keys in fact are attribute db
	// names.
	std::set<QString> attributeDBNames = dataTuple->getAllPathKeys();
	for (std::set<QString>::const_iterator it = attributeDBNames.begin(); it != attributeDBNames.end(); it++) {
		QString attributeDBName = *it;
		context->log("ImportExport", "DataImporter", 3, "  Inspecting attribute " + attributeDBName);
		const Attribute *attribute = entity->getAttributeByDBName(attributeDBName);
		if (attribute->getDBType() == "longblob") {
			// We have a BLOB value we must load from an extra file
			AttributeData *attributeData = dataTuple->getAttributeData(attributeDBName);
			attributeData->setNull();
		}
	}
}

void DataImporter::updateForeignReferences(DataSet *dataSet) {
	context->log("ImportExport", "DataImporter", 7, "Starting updateForeignReferences");

	DatabaseContext *context = getContext();

	int loopNumber = 1;
	std::vector<DataTuple*> tuples = dataSet->getAllTuples();
	for (std::vector<DataTuple*>::const_iterator it = tuples.begin(); it != tuples.end(); it++) {
		DataTuple *tuple = *it;
		bool tupleChanged = false;

		QString tupleAddress = QString::number((qlonglong)tuple, 16);
		context->log("ImportExport", "DataImporter", 6, "  Processing tuple " + tupleAddress);

		// In this step, all tuples are expected to be finished, i.e. all
		// path keys should be ordinary attribute db names
		std::set<QString> attributeDBNames = tuple->getAllPathKeys();
		for (std::set<QString>::const_iterator it2 = attributeDBNames.begin(); it2 != attributeDBNames.end(); it2++) {
			QString attributeDBName = *it2;
			AttributeData *attributeData = tuple->getAttributeData(attributeDBName);
			if (attributeData->isForeignKeyToDataTuple()) {
				DataTuple *destinationTuple = attributeData->getForeignKeyToDataTuple();
				// TODO: Fetch pk in a general way
				QVariant destinationPrimaryKey = destinationTuple->getAttributeData("id")->getValue();
				tuple->setForeignKeyToDB(attributeDBName, destinationPrimaryKey);
				tupleChanged = true;
				context->log("ImportExport", "DataImporter", 5, "    Processing attribute " + attributeDBName
							 + " ... Is a foreign key to DataTuple, set it to fk " + destinationPrimaryKey.toString());
			} else {
				context->log("ImportExport", "DataImporter", 2, "    Processing attribute " + attributeDBName 
															 + " ... Is not a foreign key to DataTuple");
			}
		}
		if (tupleChanged) {
			context->log("ImportExport", "DataImporter", 6, "    Tuple has changed, calling insertOrUpdate"); 

			// If there was any foreign key reference to a DataTuple, we
			// have to update the database
			ControllerHelper::insertOrUpdate(context, tuple);
		}

		if (loopNumber % 50 == 0) {
				emit statusChanged(trUtf8("Passe Fremdschluessel an: %1 Datensaetze bearbeitet").arg(loopNumber));
		}
		loopNumber++;
	}
	context->log("ImportExport", "DataImporter", 7, "Finishing updateForeignReferences");
}

ImportStrategy* DataImporter::getImportStrategy(QString entityDBName) {
	if (entityDBNameToImportStrategy.find(entityDBName) != entityDBNameToImportStrategy.end()) {
		return entityDBNameToImportStrategy.find(entityDBName)->second;
	} else {
		return standardImportStrategy;
	}
}


void DataImporter::addTypeForCodeValueToTuple(DataTuple *tuple, QString codeValuePath, QString type) {
	QString namePath = codeValuePath + ".name";

	if (tuple->hasAttributeData(namePath)) {
		AttributeData *nameAttributeData = tuple->getAttributeData(namePath);
		QString typePath = codeValuePath + ".type.name";
		if (nameAttributeData->isValue()) {
			if (!tuple->hasAttributeData(typePath)
				|| tuple->getAttributeData(typePath)->isNull()) {
				tuple->setValue(typePath, type);
			}
		} else if (nameAttributeData->isNull()) {
			tuple->setNull(typePath);
		}
	}
}

void DataImporter::checkDataModelVersion(DatabaseContext *context, ImpExpMetaInformation metaInformation) {
	DataModelInfoController controller = DataModelInfoController();
	std::vector<DataModelInfo*> dataModelInfos = controller.get(context, std::map<QString, QVariant>());
	if (dataModelInfos.size() != 1) {
		QString shortText = trUtf8("Datenmodell-Information fehlerhaft");
		QString longText = trUtf8("Die Datenmodell-Informationen (Version etc.) konnten nicht sauber geladen werden. \n\n"
								  "Bitte kontaktieren Sie den Administrator.");
		DataHelper::deleteVectorContents(dataModelInfos);
		throw Exception("ImportExport", "DataImporter", shortText, longText);
	} else {
		DataModelInfo* dataModelInfo = dataModelInfos[0];
		QString dataModelVersion = dataModelInfo->getVersion();

		DataHelper::deleteVectorContents(dataModelInfos);

		if (metaInformation.getDataModelVersion().isNull()) {
			QString title = trUtf8("Keine Datenmodellversion angegeben");
			QString text = trUtf8("Es wurde keine Datenmodellversion in der Importdatei angegeben. "
								  "Diese kann in einer Zeile der Format \"DatenmodellVersion;version\" "
								  "angegeben werden, wobei statt \"version\" die Datenmodellversion angegeben wird. ");
			throw Exception("ImportExport", "DataImporter", title, text);
		} else if (metaInformation.getDataModelVersion() != dataModelVersion) {
			QString title = trUtf8("Abweichende Datenmodellversion erkannt");
			QString text = trUtf8("Die Datenmodellversion des Programms ist %1, die in der Importdatei aber %2.")
								  .arg(dataModelVersion).arg(metaInformation.getDataModelVersion());
			throw Exception("ImportExport", "DataImporter", title, text);
		}
	}
}
