#ifndef DATAIMPORTER_H
#define DATAIMPORTER_H

#include <QFile>
#include <QObject>
#include <QString>

#include "DatabaseContext.h"
#include "DataModelInfo.h"
#include "DataModelInfoController.h"
#include "DataSet.h"
#include "Exception.h"
#include "ForeignKeyEdge.h"
#include "ForeignKeyGraph.h"
#include "ForeignKeyNode.h"
#include "ImpExpMetaInformation.h"
#include "ImportResult.h"
#include "ImportStrategy.h"

/** DataImporters are responsible for importing data from a file into the
 *  database.  The file format is subclass-specific.  For more information
 *  see ... (TODO)
 */
class DataImporter : public QObject {
	Q_OBJECT

public:
	/** Constructs a new DataImporter instance.
     */
	DataImporter();

	/** Destructs the DataImporter instance.
	 */
	~DataImporter();

	/** See import(context, file).
	 */
	virtual ImportResult* import(DatabaseContext *context, QString fileName);

	/** Imports the data from the given file into the database.  See also
     *  ... (TODO).  Returns an ImportResult instance with information
	 *  about non-fatal errors and warnings.  Throws an Exception in
     *  case of fatal errors, e.g. syntax errors making the import file
	 *  unreadable.
	 *  @param context the context
	 *  @param file the input file
	 *  @return information about non-fatal errors and warnings
	 *  @throws Exception in case of fatal errors
	 */
	virtual ImportResult* import(DatabaseContext *context, QFile* file);

	/** Adds the type of a CodeValue to the given DataTuple.  Example: Assume
	 *  the DataTuple contains a path key jugend.name.  Then this function may
	 *  be called with argument codeValuePath = "jugend" and type = "Jugend".
	 *  It will then add a new path key jugend.type.name = "Jugend".
	 *  Note that this will only happen if the tuple actually contains the name
	 *  of the CodeValue.  If it does not, we assume that the CodeValue is
	 *  supposed to be NULL - and for NULL path keys, all children have to be
	 *  NULL, thus we may not set jugend.type.name in this case.
	 *  @param tuple any DataTuple
	 *  @param codeValuePath path key for some CodeValue
	 *  @param type intended type of the CodeValue
	 */
	static void addTypeForCodeValueToTuple(DataTuple *tuple, QString codeValuePath, QString type);

	virtual ImpExpMetaInformation extractMetaInformation(DatabaseContext *context, QFile *file) = 0;

	static void checkDataModelVersion(DatabaseContext *context, ImpExpMetaInformation metaInformation);

signals:
	void statusChanged(QString status);

protected:

	/** Returns the DatabaseContext instance used during import.  Useful for
	 *  accessing it from subclasses.
	 *  @return the DatabaseContext instance used during import
	 */
	DatabaseContext *getContext();

	/** Reads the data from the given file and constructs a DataSet based on it.
	 *  The DataSet must contain both schema information for all entities present
	 *  in the file, and the data itself.  The following problems must be checked
	 *  here:
	 *  - duplicate/inconsistent schema information
	 *  - missing schema information
	 *  - schema-information or data for abstract entities (there are no tables
	 *    for them, thus importing them doesn't make sense)
	 *  - inconsistencies between schema and data on a syntactic level, e.g.
	 *    wrong number of tokens on a csv line
	 *  Things like correct data types, correct attribute paths, duplicate data
	 *  tuples, unresolvable foreign key references are not checked here.
	 *  @param file the input file
	 *  @return the newly constructed DataSet as described
	 *  @throws Exception in case of fatal errors
	 */
	virtual DataSet* readData(QFile* file) = 0;

	/** Adds the given Exception representing a non-fatal error to the
	 *  ImportResult
	 *  @param e Exception for a non-fatal error
	 */
	void addNonFatalError(Exception e);

private:
	/** Transforms the given DataSet with its graph structure into a state
	 *  ready for import.  That is, after executing this method,
	 *  - all foreign key references must be resolved (i.e. replaced by pointers
	 *    to DataTuples or foreign key references)
	 *  - all transformations on the data, e.g. format conversions etc. have to
	 *    be performed
	 *  - all checks on the data, e.g. for illegal values have to be performed
	 *  @param dataSet any DataSet
	 */
	void preprocessData(DataSet *dataSet, QFile *importFile);

	/** Resolves all foreign key references.  E.g. replaces value
	 *  AttributeData instances a.b and a.c by an appropriatly chosen a.
	 *  @param dataSet any DataSet
	 */
	void resolveReferences(DataSet *dataSet);

	/** Tries to resolve foreign key references in the foreign key graph
	 *  corresponding to the given DataTuple.  E.g. replaces the value
	 *  AttributeData a.b and a.c by a foreign key AttributeData a.
	 *  Returns wether any such reference anywhere in the graphs for the
	 *  given DataTuple could be resolved.
	 *  @param dataSet the corresponding DataSet, must contain a ForeignKeyGraph
	 *                 for any attribute of any entity referenced in the
     *                 DataTuple
	 *  @param dataTuple any DataTuple present in the given DataSet
	 *  @return wether at least one foreign key reference could be resolved
	 */
	bool resolveTuple(DataSet *dataSet, DataTuple *dataTuple);

	/** Returns wether the given DataTuple is finished with respect to the given DataSet.
	 *  It is finished if for all attributes of the entity corresponding to the DataTuple,
	 *  an AttributeData is given (e.g. we need AttributeData for a, AttributeData for a.b
	 *  and a.c is not sufficient).
	 *  @param dataSet the DataSet
	 *  @param dataTuple any DataTuple present in the given DataSet
	 *  @return wether the DataTuple is finished in the sense described above
	 */
	bool isTupleFinished(DataSet *dataSet, DataTuple *dataTuple);

	/** Returns wether the given node is finished with respect to the given
	 *  DataTuple.  A node is finished, if and only if for all outgoing edges,
	 *  an AttributeData is present in the given DataTuple.
	 *
	 *  For example: Given some node with incoming edge a and outgoing
	 *  edges b, c, it is finished if the DataTuple contains AttributeData for
	 *  a.b and a.c.
	 */
	bool isNodeFinished(ForeignKeyNode *node, DataTuple *dataTuple);

	std::vector<QString> getUnfinishedTuplesErrorStrings(DataSet *dataSet, std::vector<DataTuple*> unfinishedTuples);

	/** Resolves the given node with respect to the given DataTuple.  This means,
	 *  if there are path keys for all outgoing edges stored in the DataTuple,
	 *  it matches them against existing tuples both on the database and in the
	 *  input file.  Cases:
	 *  - if more than one tuple exists either on the database or in the input
	 *    file, the reference is not resolvable and an Exception is thrown
	 *  - if both on the database and in the input file is exactly one
	 *    match, the input file overwrites the database, and we use the already
	 *    existing key on the database, as it will not change when overwriting
	 *    the tuple
	 *  - if either on the database or in the input file is exactly one match,
	 *    we use it for a corresponding AttributeData
	 *  - otherwise, we do nothing, as the reference might be resolvable later
	 *    (when other tuples have been resolved)
	 *  @param dataSet the DataSet
	 *  @param dataTuple any DataTuple present in the given DataSet
	 *  @param node any node in the ForeignKeyGraphs for the entity corresponding
	 *              to the DataTuple
	 *  @return vector with the given node if it could be resolved, otherwise
	 *          the empty vector
	 */
	std::vector<ForeignKeyNode*> resolveNode(DataSet *dataSet, DataTuple *dataTuple,
											 ForeignKeyNode *node);

	/** Children of a node may have NULL values in a DataTuple.  This represents
	 *  NULL foreign key reference.  E.g., if a.b and a.c are child edges set to
	 *  NULL in the DataTuple, then a is considered a NULL foreign key references.
	 *  This function checks wether all children have NULL values in the DataTuple.
	 *  If this is the case, the children will be deleted from the DataTuple and
	 *  replaced by a NULL value for the given node.
	 *  @param dataTuple some DataTuple
	 *  @param node node in a ForeignKeyGraph
	 *  @return wether there was any progress, i.e. the replacement as described
	 *          above has taken place
	 */
	bool resolveNullChildren(DataTuple *dataTuple, ForeignKeyNode *node);

	/** Tries to resolve the given node in the given tuple on the database.
	 *  That is, it performs a database search for a tuple that has
	 *  attribute values matching the values of the child edges and attributes
	 *  in the given DataTuple.
	 */
	std::vector<QVariant> resolveOnDatabase(DataTuple *dataTuple, ForeignKeyNode *node);

	std::vector<QVariant> findEqualTuplesOnDatabase(DatabaseContext *context, DataTuple *dataTuple);

	/** See resolveOnDatabase, here we resolve in the import data.
	 */
	std::vector<DataTuple*> resolveInImportData(DataSet *dataSet, DataTuple *dataTuple, ForeignKeyNode *node);

	std::vector<DataTuple*> findEqualTuplesInImportData(DataSet *dataSet, DataTuple *dataTuple);

	/** Deletes the AttributeData for all attributes and all out 
	 *  edges of the given node from the given DataTuple.
	 *  @param dataTuple any DataTuple
	 *  @param node some node in a ForeignKeyGraph
	 */
	void deleteChildren(DataTuple *dataTuple, ForeignKeyNode *node);


	/** Processes the given node:
	 *  - If it is finished with respect to the given tuple, and is the root node,
	 *    simply return the root node for marking the given tuple finished.
	 *  - If it is finished, and is not the root node, first try to resolve a
	 *    foreign key reference or a foreign tuple using the information given
	 *    for the node in the DataTuple.  If
	 *    - two or more references or two or more tuples could be resolved, 
	 *      resolving that foreign key is not possible at all, thus throw an 
	 *      appropriate ImportException
	 *    - exactly one reference xor tuple could be resolved, exchange the data in
	 *      the given DataTuple accordingly (i.e. remove keys for children and
	 *      add one for the newly resolved foreign references) and return the
	 *      node.
	 *    - one reference and one tuple could be resolved, take the reference and
	 *      proceed as specified by the previous point.  The referenced tuple will
     *      be updated with the contents of the resolved tuple afterwards, but
	 *      without changing its primary key.  Thus referencing the tuple in the
	 *      database right now is safe.
	 *    - no reference/tuple could be resolved, return the empty vector.  Note
	 *      that this does not mean that the foreign key can't be resolved at some
	 *      time in the future, as there might be a tuple as foreign key destination
	 *      that will be resolved at some time afterwards.
	 * - If it is not yet finished, then there is at least some child node.  Inspect
	 *   all child nodes, calling this method recursively, and return the
	 *   concatenation of the results of the recursive calls.
	 *
	 * Note that changing the Import/Export key of a tuple during a import is not
	 * allowed.
	 *
	 * Thus, if the return value is
	 * - the empty vector, we know that there was no progress regarding the foreign
	 *   key graph at hand.
	 * - a vector containing the root node, we know that the attribute or foreign
	 *   key reference is completely resolved.
	 * - some other non-empty vector, we know that there was some progress regarding
	 *   resolving this foreign key reference, although we are not yet finished.
	 */
    std::vector<ForeignKeyNode*> processNode(DataSet *dataSet, DataTuple *dataTuple,
											 ForeignKeyNode *node, bool isRootNode);


	void checkAndConvertRawTypes(DataSet *dataSet);
	void checkAndConvertRawTypes(DataSet *dataSet, DataTuple *dataTuple);
	QVariant checkAndConvertRawType(DataTuple *dataTuple, QString pathKey, QString expectedType, QVariant value);

	/** Performs checks and conversions that have to take place before
	 *  references are resolved.  Example: Generate parts of an import /
	 *  export key that is needed lateron for resolving references based
	 *  on other attributes of some tuple.
	 *  @param dataSet the DataSet
	 */
	void checkAndConvertBeforeResolve(DataSet *dataSet);

	/** For each DataTuple, this function searches for tuples on the database
	 *  which map this tuple in terms of the import/export key.  If
     *  - none is found, this is no harm since this indicates that a new tuple
	 *    is imported
	 *  - exactly one is found, its primary key is recorded in order to overwrite
	 *    it later
	 *  - more than one is found, this is regarded as data inconsistency and an
	 *    exception is thrown.
     */
	void mapToDatabaseTuples(DataSet *dataSet);

	/** Returns wether this DataTuple contains a foreign key reference to a
	 *  DataTuple that was not yet successfully mapped to a database tuple.
	 */
	bool containsReferencesToUnMappedDataTuples(DataTuple *dataTuple);

	/** Given a DataSet, where all foreign keys are resolved and DataTuples are mapped
	 *  to the corresponding tuples on the database wherever they exist, this checks
	 *  for duplicates in the import data.
	 *  If any are found, an exception is thrown.
	 *  @param dataSet the DataSet
	 */
	void checkForDuplicatesInImportData(DataSet *dataSet);

	bool containsReferencesToUnfinishedDataTuples(DataTuple *dataTuple);

	void postProcessAfterWrite(DataSet *dataSet);
	void postProcessAfterWrite(DataSet *dataSet, DataTuple *dataTuple);

	void writeToDatabase(DataSet *dataSet, QFile *importFile, ImportResult *importResult);

	/**  A (in fact optional) test wether all blob files can be found and opened.
	 *   @param dataSet the DataSet
	 */
	void checkBLOBs(DataSet *dataSet, QFile *importFile);

	/**  The contents of BLOBs is not saved directly within the import file,
	 *   but in separate files.  At the correspondig position in the input
	 *   file, just the name of the file where the BLOB data (and nothing else)
	 *   is given.
	 *
	 *   This function replaces these file names in a given DataTuple by
	 *   the correct data loaded from these files.  We do this for each
	 *   DataTuple separately, because otherwise we might run into
	 *   problems related to consuming too much main memory (e.g. when
	 *   importing 1000 pictures).
	 *   @param dataSet the DataSet
	 */
	void loadBLOBs(DataTuple *dataTuple, QFile *importFile);

	void removeBLOBDataFromTuple(DataTuple *dataTuple);

	void updateForeignReferences(DataSet *dataSet);

    ImportStrategy* getImportStrategy(QString entityDBName);

	ImportResult *importResult;
	DatabaseContext *context;
	ImportStrategy* standardImportStrategy;
	std::map<QString, ImportStrategy*> entityDBNameToImportStrategy;

};
#endif
