#include "ImportThread.h"

#include "ArchiveHelper.h"
#include "CSVDataImporter.h"
#include "DataImporter.h"
#include "DatenbankVerbindung.h"
#include "Exception.h"
#include "ImportException.h"
#include "ImportResult.h"
#include "NoSuchPrivilegeException.h"

ImportThread::ImportThread(DatabaseContext *context, bool createdTempDir, QString tempDir, QString fileName) {
	this->context = context;
	DatenbankVerbindung *andereVerbindung = context->getDatenbankVerbindung();
	this->datenbankVerbindung = andereVerbindung->copy();
	this->session = context->getSession()->copy();

	this->fileName = fileName;
	this->createdTempDir = createdTempDir;
	this->tempDir = tempDir;

	QObject::connect(this, SIGNAL(finished()), this, SLOT(deleteLater()));
}

void ImportThread::run() {
	context->log("ImportExport", "ImportThread", 8, "Starting import thread");

	this->context = new DatabaseContext(this->datenbankVerbindung, session);


	ImportResult *importResult;
	DataImporter* importer = new CSVDataImporter();

	QDateTime startDateTime = QDateTime::currentDateTime();
	try {
		QObject::connect(importer, SIGNAL(statusChanged(QString)),
						 this, SLOT(statusMessageFromImporter(QString)));
		importResult = importer->import(context, fileName);
		importResult->setStartDateTime(startDateTime);
	} catch (ImportException e) {
		importResult = new ImportResult();

		// Storing it in the ImportResult here is not quite nice, but I can´t
		// throw it to the other thread...
		QString locationString;
		if (e.hasLineNumber()) {
			locationString = trUtf8("Fehler beim Einlesen von Zeile %1 der Datei %2: \n\n")
							  .arg(e.getLineNumber()).arg(fileName);
		} else {
			locationString = trUtf8("Fehler beim Einlesen der Datei %1: \n\n").arg(fileName);
		}
		QString longText = locationString + e.getLongText()
						 + trUtf8("\n\nDaher konnten die Daten NICHT importiert werden."
							  "Bitte beheben Sie den Fehler und versuchen Sie es nochmals.");
		importResult->addException(e);
		importResult->setStartDateTime(startDateTime);
	} catch (Exception e) {
		importResult = new ImportResult();

		// Storing it in the ImportResult here is not quite nice, but I can´t
		// throw it to the other thread...
		importResult->addException(e);
		importResult->setStartDateTime(startDateTime);
	} catch (NoSuchPrivilegeException inException) {
		importResult = new ImportResult();
		Exception e = Exception(inException.getShortText(), inException.getLongText());
		importResult->addException(e);
		importResult->setStartDateTime(startDateTime);
	} catch (PersistenceException inException) {
		importResult = new ImportResult();
		Exception e = Exception(inException.getShortText(), inException.getLongText());
		importResult->addException(e);
		importResult->setStartDateTime(startDateTime);
	}
	delete importer;

	context->log("ImportExport", "ImportThread", 8, "Finishing import thread");
	emit importFinished(context, importResult, createdTempDir, tempDir);
}

void ImportThread::statusMessageFromImporter(QString status) {
	emit statusMessage(status);
}
