#include "ImportStrategy.h"

#include <QDate>
#include <QDateTime>

#include "AttributeData.h"
#include "Entity.h"

void ImportStrategy::checkAndConvertBeforeResolve(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple, ImportResult *importResult) {
	// Default: Do nothing
}

bool ImportStrategy::insertAlways(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple,
								  ImportResult *importResult) {
	// Default: Choose decision update (overwrite) vs. insert based on import/export key
	return false;
}

void ImportStrategy::postProcessAfterWrite(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple, ImportResult *importResult) {
	// Default: Do nothing
}

void ImportStrategy::testForMustValue(DataTuple *dataTuple, QString pathKey) {
	AttributeData *attributeData = dataTuple->getAttributeData(pathKey);
	if (attributeData == NULL || !attributeData->isValue()) {
		QString title = QObject::trUtf8("Attribut nicht angegeben");
		QString text = QObject::trUtf8("Im Datensatz in Zeile %1 ist fuer das Attribut %2  der Tabelle %3 kein Wert angegeben. Bitte geben Sie einen Wert an.").arg(dataTuple->getLineNumber()).arg(pathKey).arg(dataTuple->getEntityDBName());
		throw ImportException("ImportExport", "ImportStrategy", title, text);
	}
}

void ImportStrategy::testForMustFK(DataTuple *dataTuple, QString pathKeyPrefix) {
	bool wrong = false;

	std::set<QString> allPathKeys = dataTuple->getAllPathKeys();
	for (std::set<QString>::const_iterator it = allPathKeys.begin(); it != allPathKeys.end(); it++) {
		QString currPathKey = *it;
		if (currPathKey.startsWith(pathKeyPrefix + ".")) {
			AttributeData *attributeData = dataTuple->getAttributeData(currPathKey);
			if (attributeData->isNull()) {
				wrong = true;
			}
		}
	}

	if (wrong) {
		QString title = QObject::trUtf8("Verweis nicht angegeben");
		QString text = QObject::trUtf8("Im Datensatz in Zeile %1 der Tabelle %2 fehlt ein Wert fuer den Verweis mit Praefix %3.\n\n"
									   "Der Zweck dieses Verweises ist es, eine Verbindung zu einer Zieltabelle herzustellen.  Dazu werden bestimmte Werte der "
									   "Zieltabelle angegeben, anhand derer das Ziel eindeutig bestimmt wird.  Ein Verweis mit Namen a koennte beispielsweise "
									   "in Spalten des Namens a.b und a.c "
									   "ein Objekt der Zieltabelle referenzieren, das in den Attributen b bzw. c die angegebenen Werte besitzt.\n\n"
									   "Bitte fuellen Sie die entsprechende(n) Spalte(n) aus.")
										  .arg(dataTuple->getLineNumber()).arg(dataTuple->getEntityDBName()).arg(pathKeyPrefix);
		throw ImportException("ImportExport", "ImportStrategy", title, text);
	}
}

bool ImportStrategy::allowMissingAttribute(const Attribute* attribute) {
	return false;
}

QVariant ImportStrategy::getDefaultValue(const Attribute* attribute) {
	return QVariant();
}

bool ImportStrategy::allowUnknownAttribute(QString entityDBName, QString attributeDBName) {
	return false;
}

QVariant ImportStrategy::checkAndConvertValue(const Attribute* attribute, QVariant value, 
											  ImportResult *importResult) {
	if (attribute->isForeignKey()) {
		return QVariant();
	} else {
		QVariant convertedValue = convert(attribute, value, importResult);
		check(attribute, convertedValue, importResult);
		return convertedValue;
	}
}

QVariant ImportStrategy::convert(const Attribute* attribute, QVariant value, 
								 ImportResult *importResult) {
	if (attribute->isForeignKey()) {
		return QVariant();
	} else {
		QString codeType = attribute->getCodeType();
		QString valueString = value.toString();
		bool ok;
		if (codeType == "int") {
			int valueInt = valueString.toInt(&ok);
			return (ok ? QVariant(valueInt) : QVariant());
		} else if (codeType == "QString") {
			return value;
		} else if (codeType == "qlonglong") {
			qlonglong valueLongLong = valueString.toLongLong(&ok);
			return (ok ? QVariant(valueLongLong) : QVariant());
		} else if (codeType == "QDate") {
			QDate date = QDate::fromString(valueString, Qt::ISODate);
			return (date.isValid() ? QVariant(date) : QVariant());
		} else if (codeType == "QDateTime") {
			QDateTime dateTime = QDateTime::fromString(valueString, Qt::ISODate);
			return (dateTime.isValid() ? QVariant(dateTime) : QVariant());
		} else {
			QString entityDBName = attribute->getEntity()->getDBName();
			QString attributeDBName = attribute->getDBName();
			QString attributeDBType = attribute->getDBType();
			QString longMessage = tr("Das Attribut %1 der Entitt %2 hat im Datenmodell als "
				 "Datentyp %3.  Dieser Typ wird leider gegenwrtig vom Importer nicht "
				 "untersttzt.  Dies ist kein Benutzerfehler, sondern eine (noch) "
				 "nicht vorhandene Fhigkeit des Programms.")
				 .arg(attributeDBName).arg(entityDBName).arg(attributeDBType);
			throw Exception(tr("Nicht untersttzter Datentyp"), longMessage);
		}
	}
}
// 

void ImportStrategy::check(const Attribute* attribute, QVariant value, ImportResult *importResult) {
	// the default implementation does nothing
}
