#ifndef IMPORTSTRATEGY_H
#define IMPORTSTRATEGY_H

#include <QObject>
#include <QVariant>

#include "Attribute.h"
#include "DatabaseContext.h"
#include "DataSet.h"
#include "DataTuple.h"
#include "ImportResult.h"
#include "ImportException.h"

class ImportStrategy : public QObject {
	Q_OBJECT
public:

	virtual void checkAndConvertBeforeResolve(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple,
							  ImportResult *importResult);

	virtual bool insertAlways(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple,
							  ImportResult *importResult);

	virtual void postProcessAfterWrite(DatabaseContext *context, DataSet *dataSet, DataTuple *dataTuple,
							  ImportResult *importResult);

	static void testForMustValue(DataTuple *dataTuple, QString pathKey);
	static void testForMustFK(DataTuple *dataTuple, QString pathKeyPrefix);


	/** Checks and converts the given value of the given attribute.  Note that
	 *  via the Attribute instance, the entity can also be derived.  For
	 *  example, using this function BLOB values stored elsewhere 
	 */
	// DESIGN NOT YET FINISHED
//	virtual void checkAndConvertAfterResolve(const Attribute *attribute, QVariant value, ImportResult *importResult);

	/** This method returns, wether the given Attribute may be missing after
	 *  the import step, where missing Attributes are inspected.  For example,
	 *  one would return true here for an Attribute that will be set while
	 *  updating the database.  If you want to set a default value, you 
	 *  must return false here and use the appropriate method below.
	 *  @param attribute missing Attribute
	 */
	virtual bool allowMissingAttribute(const Attribute* attribute);

	/** If some attribute is not allowed to be missing, this method may specify
	 *  a default value.  If the returned QVariant is invalid, there is no 
	 *  default value, i.e. the importer will raise an error.
	 *  The default implementation here does not provide any default values.
	 *  @param attribute missing Attribute
	 */
	virtual QVariant getDefaultValue(const Attribute* attribute);

	/** This method returns, wether an unknown attribute with the given 
	 *  name and value should be allowed in the given entity. 
	 *  @param entityDBName db name of entity
	 *  @param attributeDBName db name of attribute
	 *  @param value value of attribute
	 *  @return wether the given unknown attribute should be allowed
	 */
	virtual bool allowUnknownAttribute(QString entityDBName, QString attributeDBName);

	/** Checks and converts the given value of the given Attribute according to the
	 *  attribute type and some possible entity and attribute specific extra
	 *  logic.  Expects a string QVariant.  May not be called for foreign
	 *  key attributes, as those are handled elsewhere.  Returns a QVariant
	 *  according to the code type of the given attribute.
	 *
	 *  If the value can´t be converted (e.g. if we get some malformed
	 *  string for an int attribute) or has an invalid value (e.g. 300 for an
	 *  int attribute that must be smaller than 100), an Exception will
	 *  be thrown.  Maybe at a later time, we also support adding the Exception
	 *  to the given ImportResult, allowing the import to proceed in the case
	 *  of non-fatal errors (e.g. a name had to be truncated because it was
	 *  too long).  But this is not yet supported!
	 *  @param attribute Attribute to check and convert
	 *  @param value value of the attribute, no invalid QVariant
	 *  @param importResult given to allow later extensions without having to
	 *                      adjust the interface
	 *  @return checked and converted data value, undefined for foreign key 
	 *          attributes
	 *  @throws Exception in case of errors
	 */
	virtual QVariant checkAndConvertValue(const Attribute* attribute, QVariant value, ImportResult *importResult);

protected:
	/** Converts the given string QVariant according to the type specification
	 *  in the given Attribute.  Returns an invalid QVariant if this is not
	 *  possible, e.g. if you try to load some string into an int attribute.
	 *  May not be called for foreign key attributes, then the return value 
	 *  will be undefined.  Errors must be reported using Exceptions, the
	 *  additional parameter ImportResult is only there to allow future
	 *  extensions without having to adjust the interface.
	 *  @param attribute Attribute to check
	 *  @param value value of the attribute
	 *  @param importResult given to allow later extensions without having to
	 *                      adjust the interface
	 *  @return converted data value according to the code type, 
	 *          undefined for foreign key attributes
	 */
	virtual QVariant convert(const Attribute* attribute, QVariant value, ImportResult *importResult);

	/** Checks the given QVariant value of the given Attribute.  Any problem
	 *  detected here must be reported by throwing an Exception. The
	 *  additional parameter ImportResult is only there to allow future
	 *  extensions without having to adjust the interface.
	 *  @param attribute Attribute to check
	 *  @param value value of the attribute
	 *  @param importResult given to allow later extensions without having to
	 *                      adjust the interface
	 */
	virtual void check(const Attribute* attribute, QVariant value, ImportResult *importResult);
};
#endif
