#include "MitgliedImportStrategy.h"

#include "AttributeData.h"
#include "DataImporter.h"
#include "ImportException.h"
#include "MitgliedsNummernManager.h"
#include "MitgliedsPruefer.h"

void MitgliedImportStrategy::checkAndConvertBeforeResolve(DatabaseContext *context, DataSet *dataSet,
														  DataTuple *dataTuple, ImportResult *importResult) {
	if (dataTuple->getEntityDBName() != "Mitglied") {
		throw ImportException("ImportExport", "MitgliedImportStrategy", trUtf8("Technischer Fehler"),
							trUtf8("MitgliedImportStrategy falsch aufgerufen"));
	} else {
		testForMustFK(dataTuple, "verein");
		testForMustFK(dataTuple, "mitgliedsArt");
		testForMustValue(dataTuple, "mitgliedsNummerImVerein");
		testForMustFK(dataTuple, "jugend");


		QString tupleAddress = QString::number((qlonglong)dataTuple, 16);
		context->log("ImportExport", "MitgliedImportStrategy.cpp", 5, "Calling MitgliedImportStrategy for tuple " + tupleAddress);

		// Die globale Mitgliedsnummer wird aus den restlichen Daten berechnet,
		// eine möglicherweise in den Importdaten vorhandene Nummer wird nicht
		// beachtet
		if (dataTuple->hasAttributeData("mitgliedsNummerImVerein")
			&& dataTuple->hasAttributeData("mitgliedsArt.name")
			&& dataTuple->hasAttributeData("verein.vereinsNummerKomplett")) {
			AttributeData *mitgliedsnrAD = dataTuple->getAttributeData("mitgliedsNummerImVerein");
			AttributeData *mitgliedsartAD = dataTuple->getAttributeData("mitgliedsArt.name");
			AttributeData *vereinsnrAD = dataTuple->getAttributeData("verein.vereinsNummerKomplett");
			if (mitgliedsnrAD->isValue() && vereinsnrAD->isValue()) {
				QString mitgliedsnrLokal = mitgliedsnrAD->getValue().toString();
				QString mitgliedsart = mitgliedsartAD->getValue().toString();
				QString vereinsnr = vereinsnrAD->getValue().toString();
				QString mitgliedsNrGlobal
					= MitgliedsNummernManager::berechneMitgliedsnummerKomplett(vereinsnr, mitgliedsnrLokal, mitgliedsart, false);
				context->log("ImportExport", "MitgliedImportStrategy", 2, "Mitgliedsnummer " + mitgliedsNrGlobal + " berechnet.");
				dataTuple->setValue("mitgliedsNummerKomplett", mitgliedsNrGlobal);
				QString mitgliedsNrSchluessel
					= MitgliedsNummernManager::berechneMitgliedsnummerKomplett(vereinsnr, mitgliedsnrLokal, mitgliedsart, true);
				context->log("ImportExport", "MitgliedImportStrategy", 2, "Mitgliedsnummer-Schluessel " + mitgliedsNrSchluessel + " berechnet.");
				dataTuple->setValue("mitgliedsNummerSchluessel", mitgliedsNrSchluessel);
			}
		}

		// Die Attribute mitgliedsArt.type.name bzw. jugend.type.name dürfen fehlen.
		// Sie werden dann auf die einzig möglichen Werte Mitgliedsart bzw. Jugend
		// gesetzt.  Das passiert aber nur dann wenn mitgliedsArt.name bzw. jugend.name
		// tatsächlich einen Wert haben, ansonsten würde fehlerhafterweise der NULL-Wert
		// eines Attributs ausgehebelt.
		DataImporter::addTypeForCodeValueToTuple(dataTuple, "mitgliedsArt", "Mitgliedsart");
		DataImporter::addTypeForCodeValueToTuple(dataTuple, "jugend", "Jugend");
		DataImporter::addTypeForCodeValueToTuple(dataTuple, "staat", "Staat");

		if (dataTuple->hasAttributeData("telefon")) {
			AttributeData *telefonAD = dataTuple->getAttributeData("telefon");
			if (telefonAD->isValue()) {
				QString telefon = telefonAD->getValue().toString();
				if (!MitgliedsPruefer::istTelefonnrSyntaxKorrekt(telefon)) {
					throw ImportException("ImportExport", "MitgliedImportStrategy", trUtf8("Telefonnummer falsch angegeben."),
										  trUtf8("Die Telefonnummer %1 wurde falsch angegeben. ").arg(telefon) + MitgliedsPruefer::getTelefonSyntaxInkorrektText());
				}
			}
		}

		if (dataTuple->hasAttributeData("mobil")) {
			AttributeData *mobilAD = dataTuple->getAttributeData("mobil");
			if (mobilAD->isValue()) {
				QString mobil = mobilAD->getValue().toString();
				if (!MitgliedsPruefer::istTelefonnrSyntaxKorrekt(mobil)) {
					throw ImportException("ImportExport", "MitgliedImportStrategy", trUtf8("Mobilfunknummer falsch angegeben."),
										  trUtf8("Die Mobilfunknummer %1 wurde falsch angegeben. ").arg(mobil) + MitgliedsPruefer::getMobilSyntaxInkorrektText());
				}
			}
		}

		context->log("ImportExport", "MitgliedImportStrategy.cpp", 5, "Finished MitgliedImportStrategy for tuple " + tupleAddress);
	}
}
