#ifdef JUROR_WINDOWS_VERSION

#include "WinKeyboardManager.h"

#include <QWidget>

#include <iostream>

// put this line into the header if you want compiler and linker errors ;-)
WinKeyboardManager *winKeyboardManagerInstancePointer;

WinKeyboardManager::WinKeyboardManager(DatabaseContext *context) {
	winKeyboardManagerInstancePointer = this;
	keyboardsAreRegistered = false;
	this->context = context;
}

LRESULT CALLBACK keyboardCallbackProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	return winKeyboardManagerInstancePointer->callbackCalled(hwnd, message, wParam, lParam);
}

LRESULT CALLBACK WinKeyboardManager::callbackCalled(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	if (message == WM_INPUT) {
		context->log("Keyboards", "WinKeyboardManager", 4, "Received some WM_INPUT event.");
		UINT dwSize;

	    GetRawInputData((HRAWINPUT)lParam, RID_INPUT, NULL, &dwSize, sizeof(RAWINPUTHEADER));
	    LPBYTE lpb = new BYTE[dwSize];
	    if (lpb == NULL) {
	        return 0;
	    }

	    if (GetRawInputData((HRAWINPUT)lParam, RID_INPUT, lpb, &dwSize,
	    		sizeof(RAWINPUTHEADER)) != dwSize ) {
			context->log("Keyboards", "WinKeyboardManager", 9, "GetRawInputData doesn't return correct size.  Expected: "
							 + QString::number(dwSize) + ", found " + QString::number(sizeof(RAWINPUTHEADER)));
	    }

	    RAWINPUT* raw = (RAWINPUT*)lpb;

	    if (raw->header.dwType == RIM_TYPEKEYBOARD) {
	    	if (raw->data.keyboard.Message == WM_KEYDOWN) {
	    		long deviceAsLong = (long)raw->header.hDevice;
	    		qlonglong device = deviceAsLong;
	    		QVariant number = convertToNumber(raw->data.keyboard.VKey);
	    		if (number.isValid()) {
					context->log("Keyboards", "WinKeyboardManager", 4, "Received key code " + QString::number(raw->data.keyboard.VKey)
							  + " from keyboard " + QString::number((qlonglong)(raw->header.hDevice)) + "/" + QString::number(device) + ", converted to number "
							  + QString::number(number.toInt()));
	    			emit numericKeyPressed(device, number.toInt());
	    		} else {
					context->log("Keyboards", "WinKeyboardManager", 1, "Received key code " + QString::number(raw->data.keyboard.VKey)
							  + " from keyboard " + QString::number((qlonglong)(raw->header.hDevice)) + "/" + QString::number(device) + ", could not convert it to a number.");
				}
	    	}
	    }
	} else {
		context->log("Keyboards", "WinKeyboardManager", 1, "Received some other event.");
	}

	return DefWindowProc(hwnd, message, wParam, lParam);
}

void WinKeyboardManager::registerKeyboards() {
	context->log("Keyboards", "WinKeyboardManager", 7, "Starting to register keyboards.");

	if (!keyboardsAreRegistered) {
		keyboardsAreRegistered = true;
	} else {
		// register only once
		return;
	}

	// Note: Documentation for the Raw Input API used here is located at http://msdn.microsoft.com/en-us/library/ms645598(VS.85).aspx

	HINSTANCE hInstance = qWinAppInst();

    //Registering the Window Class
	WNDCLASSEX  WindowClassEx;
    ZeroMemory(&WindowClassEx, sizeof(WNDCLASSEX));
    WindowClassEx.cbSize        = sizeof(WNDCLASSEX);
    WindowClassEx.style = 0;
    //WindowClassEx.lpfnWndProc   = testWindowProc; //MainWndProc;
    //WindowClassEx.lpfnWndProc = &WinKeyboardManager::keyboardCallbackProc;
    WindowClassEx.lpfnWndProc = keyboardCallbackProc;
    WindowClassEx.cbClsExtra = 0;
    WindowClassEx.cbWndExtra = 0;
    WindowClassEx.hInstance     = hInstance;
    WindowClassEx.hIcon = 0;
    WindowClassEx.hCursor = 0;
    WindowClassEx.hbrBackground = 0;
    WindowClassEx.lpszMenuName = 0;
    WindowClassEx.lpszClassName = L"KeyboardListener";
    WindowClassEx.hIconSm = 0;

    if (RegisterClassEx(&WindowClassEx) == 0) {
		context->log("Keyboards", "WinKeyboardManager", 9, "RegisterClassEx failed.");
    	return;
    }

    // Now create a pseudo-window for receiving messages
    HWND hwnd = CreateWindowEx(0, L"KeyboardListener", NULL, 0, 0, 0, 0, 0, HWND_MESSAGE, NULL, hInstance, NULL);
	if (hwnd == NULL) {
		context->log("Keyboards", "WinKeyboardManager", 9, "CreateWindowEx failed.");
		return;
	} else {
		context->log("Keyboards", "WinKeyboardManager", 4, "CreateWindowEx succeeded.");
	}

	// and finally register for receiving raw input
	RAWINPUTDEVICE Rid[1];
	Rid[0].usUsagePage = 0x01;
	Rid[0].usUsage = 0x06;
	Rid[0].dwFlags = 0;
	Rid[0].hwndTarget = hwnd;

	if (RegisterRawInputDevices(Rid, 1, sizeof(Rid[0])) == FALSE) {
		context->log("Keyboards", "WinKeyboardManager", 9, "Registration failed.");
		//registration failed. Call GetLastError for the cause of the error.
	} else {
		context->log("Keyboards", "WinKeyboardManager", 4, "Registration succeeded.");
	}

	context->log("Keyboards", "WinKeyboardManager", 7, "Registering keyboards finished.");
}

void WinKeyboardManager::printDeviceInformation() {
	context->log("Keyboards", "WinKeyboardManager", 4, "Starting to print device information.");

	UINT anzahlDevices;
	context->log("Keyboards", "WinKeyboardManager", 4, "  Note: Running under version " + QString(_WIN32_WINNT));

	PRAWINPUTDEVICELIST pRawInputDeviceList;
	if (GetRawInputDeviceList(NULL, &anzahlDevices, sizeof(RAWINPUTDEVICELIST)) != 0) {
		context->log("Keyboards", "WinKeyboardManager", 9, "Error when detecting keyboards.");
	}
	pRawInputDeviceList = PRAWINPUTDEVICELIST(malloc(sizeof(RAWINPUTDEVICELIST) * anzahlDevices));
	if (pRawInputDeviceList == NULL) {
		context->log("Keyboards", "WinKeyboardManager", 9, "Error when allocating memory");
	}
	if (GetRawInputDeviceList(pRawInputDeviceList, &anzahlDevices, sizeof(RAWINPUTDEVICELIST)) > 1000) {
		context->log("Keyboards", "WinKeyboardManager", 9, "Error when getting devices");
	}

	context->log("Keyboards", "WinKeyboardManager", 4, "  We have found " + QString::number(anzahlDevices) + " devices.");

	for (unsigned int n = 0; n < anzahlDevices; n++) {
		tagRAWINPUTDEVICELIST info = pRawInputDeviceList[n];
		if (info.dwType == RIM_TYPEHID) {
			context->log("Keyboards", "WinKeyboardManager", 4, "  Device " + QString::number((qlonglong)(info.hDevice)) + " has type HID.");
		} else if (info.dwType == RIM_TYPEKEYBOARD) {
			context->log("Keyboards", "WinKeyboardManager", 4, "  Device " + QString::number((qlonglong)(info.hDevice)) + " is a keyboard.");
		} else if (info.dwType == RIM_TYPEMOUSE) {
			context->log("Keyboards", "WinKeyboardManager", 4, "  Device " + QString::number((qlonglong)(info.hDevice)) + " is a mouse.");
		} else {
			context->log("Keyboards", "WinKeyboardManager", 4, "  Device " + QString::number((qlonglong)(info.hDevice)) + " is unknown.");
		}
	}

	free(pRawInputDeviceList);
	context->log("Keyboards", "WinKeyboardManager", 4, "Finished printing device information.");
}


QVariant WinKeyboardManager::convertToNumber(USHORT key) {
	  // see http://msdn.microsoft.com/en-us/library/ms645540(VS.85).aspx for key codes

	if (key == 0x31 || key == 0x61 || key == 0x23) {
		return QVariant(1);
	} else if (key == 0x32 || key == 0x62 || key == 0x28) {
		return QVariant(2);
	} else if (key == 0x33 || key == 0x63 || key == 0x22) {
		return QVariant(3);
	} else if (key == 0x34 || key == 0x64 || key == 0x25) {
		return QVariant(4);
	} else if (key == 0x35 || key == 0x65 || key == 0x0C) {
		return QVariant(5);
	} else if (key == 0x36 || key == 0x66 || key == 0x27) {
		return QVariant(6);
	} else if (key == 0x37 || key == 0x67 || key == 0x24) {
		return QVariant(7);
	} else if (key == 0x38 || key == 0x68 || key == 0x26) {
		return QVariant(8);
	} else if (key == 0x39 || key == 0x69 || key == 0x21) {
		return QVariant(9);
	} else if (key == 0x30 || key == 0x60 || key == 0x2D) {
		return QVariant(10);
	} else {
		return QVariant();
	}
}


#endif
