#include "LicenseManager.h"

#include <QByteArray>
#include <QNetworkInterface>
#include <QSettings>
#include <QVariant>

#include "DataHelper.h"
#include "DummyRSAManager.h"
#include "Exception.h"
#include "LicenseKey.h"
#include "LicenseKeyController.h"
#include "RSAManager.h"

QString LicenseManager::getHardwareKey(DatabaseContext *databaseContext) {
/*
#ifdef JUROR_WINDOWS_VERSION
	QSettings normalSettings("HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Cryptography", QSettings::NativeFormat);
	QVariant normalSettingVariant = normalSettings.value("MachineGuid");
	QString normalSettingsString = normalSettingVariant.toString();

	if (normalSettingsString.length() == 0) {
		QSettings settings6432("HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Cryptography", QSettings::NativeFormat);
		QVariant settingVariant6432 = settings6432.value("MachineGuid");
		normalSettingsString = settingVariant6432.toString();
	}

	return normalSettingsString;

#endif
#ifdef JUROR_LINUX_VERSION
*/
	QList<QNetworkInterface> interfaces = QNetworkInterface::allInterfaces();
	databaseContext->log("Permissions", "LicenseManager", 7, "Found " + QString::number(interfaces.size()) + " network interfaces.");

	std::vector<QNetworkInterface> preChosenInterfaces = std::vector<QNetworkInterface>();

	// First find out which interfaces and addresses exist
	QStringList allMacAddresses = QStringList();
	for (int n = 0; n < interfaces.size(); n++) {
		QNetworkInterface currInterface = interfaces.at(n);
		QString currHardwareAddress = currInterface.hardwareAddress();
		QString name = currInterface.name();
		QString humanReadableName = currInterface.humanReadableName();
		QNetworkInterface::InterfaceFlags flags = currInterface.flags();
		if (!(flags & QNetworkInterface::IsLoopBack)) {
			databaseContext->log("Permissions", "LicenseManager", 7, "Step 1: Choosing hardware address " + currHardwareAddress + " with flags " + flags
																			   + " for name " + name + " and human readable name " + humanReadableName);
			preChosenInterfaces.push_back(currInterface);
			allMacAddresses.push_back(currHardwareAddress);
		} else {
			databaseContext->log("Permissions", "LicenseManager", 5, "Step 1: Ignoring hardware address " + currHardwareAddress + " with flags " + flags
																			   + "for name " + name + " and human readable name " + humanReadableName);
		}
	}

	// now inspect them in the specified order, if we have a match, choose it
	std::vector<QString> deviceNameOrder = getDeviceNameOrder();
	for (std::vector<QString>::const_iterator it = deviceNameOrder.begin(); it != deviceNameOrder.end(); it++) {
		QString searchedName = *it;

		QStringList macAddresses = QStringList();
		for (std::vector<QNetworkInterface>::const_iterator it2 = preChosenInterfaces.begin(); it2 != preChosenInterfaces.end(); it2++) {
			QNetworkInterface currInterface = *it2;
			if (currInterface.humanReadableName() == searchedName) {
				QString currHardwareAddress = currInterface.hardwareAddress();
				QString name = currInterface.name();
				QString humanReadableName = currInterface.humanReadableName();
				QNetworkInterface::InterfaceFlags flags = currInterface.flags();
				macAddresses.push_back(currHardwareAddress);
				databaseContext->log("Permissions", "LicenseManager", 3, "Step 2: Choosing hardware address " + currHardwareAddress + " with flags " + flags
																				   + " for name " + name + " and human readable name " + humanReadableName);
			}
		}

		if (macAddresses.size() > 0) {
			macAddresses.sort();

			for (int n = 0; n < macAddresses.size(); n++) {
				if (!macAddresses.at(n).startsWith("00:00:00:00:00:00")) {
					QString chosenMacAddress = macAddresses.at(n);
					databaseContext->log("Permissions", "LicenseManager", 7, "Choosing address " + chosenMacAddress);
					return chosenMacAddress;
				}
			}
		}
	}

	allMacAddresses.sort();
	for (int n = 0; n < allMacAddresses.size(); n++) {
		if (!allMacAddresses.at(n).startsWith("00:00:00:00:00:00")) {
			QString chosenMacAddress = allMacAddresses.at(n);
			databaseContext->log("Permissions", "LicenseManager", 9, "Was not able to choose address in steps 1 and 2, now choosing address " + chosenMacAddress);
			return chosenMacAddress;
		}
	}

	return QString();
}

/** Returns the order of device names for the search for the MAC-Address to use for license processing.
 *  See also http://dbis-bugzilla.uibk.ac.at/bugzilla/show_bug.cgi?id=282#c4 (Comment 4 in Bugzilla 282).
 *
 *  Important: The different versions of a particular name (e.g. in different languages) must form one
 *  continuous block in this list, to ensure that the result stays stable even in the face of OS language
 *  changes etc.
 */
std::vector<QString> LicenseManager::getDeviceNameOrder() {
	std::vector<QString> deviceNames = std::vector<QString>();

	deviceNames.push_back("LAN-Verbindung");

	deviceNames.push_back("Bluetooth-Netzwerkverbindung");

	deviceNames.push_back("Drahtlosnetzwerkverbindung");

	return deviceNames;
}

LicenseKey* LicenseManager::loadLicenseKey(DatabaseContext *context) {
	LicenseKey *retValue = new LicenseKey();
/*
	LicenseKeyController controller = LicenseKeyController();
	std::vector<LicenseKey*> licenseKeyVector = controller.get(context, std::map<QString, QVariant>());
	if (licenseKeyVector.size() != 1) {
		return NULL;
	} else {
		retValue = licenseKeyVector[0];
	}
	*/

#ifdef JUROR_WINDOWS_VERSION
	QSettings settings("HKEY_LOCAL_MACHINE\\Software\\Juror", QSettings::NativeFormat);
	QVariant publicKeyVariant = settings.value("PublicKey", QVariant());
	if (!publicKeyVariant.isValid()) {
		retValue->deletePublicKey();
	} else {
		retValue->setPublicKey(publicKeyVariant.toString());
	}
	QVariant licenseKeyVariant = settings.value("LicenseKey", QVariant());
	if (!licenseKeyVariant.isValid()) {
		retValue->deleteLicenseKey();
	} else {
		retValue->setLicenseKey(licenseKeyVariant.toString());
	}

#endif
	return retValue;
}

QDate LicenseManager::getExpireDate(LicenseKey *key) {
	if (key->hasLicenseKey() && key->hasPublicKey()) {
		RSAManager *rsaManager = new DummyRSAManager();
		PublicRSAKey *publicKey = rsaManager->loadPublicKey(key->getPublicKey());

		QString licenseKey = key->getLicenseKey();

		QString decryptedString = rsaManager->decryptFromHex(publicKey, licenseKey);

		QString expireDateString = decryptedString.left(10);
		if (expireDateString == "NN.NN.NNNN") {
			return QDate::fromString("31.12.2999", "dd.MM.yyyy");
		} else {
			QDate expireDate = QDate::fromString(expireDateString, "dd.MM.yyyy");
			if (expireDate.isValid()) {
				return expireDate;
			} else {
				return QDate();
			}
		}
	} else {
		return QDate();
	}
}

QDate LicenseManager::getExpireWettbewerbDate(LicenseKey *key) {
	if (key->hasLicenseKey() && key->hasPublicKey()) {
		RSAManager *rsaManager = new DummyRSAManager();
		PublicRSAKey *publicKey = rsaManager->loadPublicKey(key->getPublicKey());

		QString licenseKey = key->getLicenseKey();

		QString decryptedString = rsaManager->decryptFromHex(publicKey, licenseKey);

		QString expireDateString = decryptedString.mid(10, 10);
		if (expireDateString == "NN.NN.NNNN") {
			return QDate::fromString("31.12.2999", "dd.MM.yyyy");
		} else {
			QDate expireDate = QDate::fromString(expireDateString, "dd.MM.yyyy");
			if (expireDate.isValid()) {
				return expireDate;
			} else {
				return QDate();
			}
		}
	} else {
		return QDate();
	}
}

void LicenseManager::checkLicenseKey(DatabaseContext *context, LicenseKey *key) {
	context->log("License", "LicenseManager", 7, "Starting to check license key");

	if (key->hasLicenseKey() && key->hasPublicKey()) {
		QDate expireDate = getExpireDate(key);
		if (expireDate.isValid()) {

			QDate currentDate = QDate::currentDate();
			QDate unlimitedDate = QDate::fromString("31.12.2999", "dd.MM.yyyy");
			if (expireDate < currentDate && expireDate < unlimitedDate) {
				context->log("License", "LicenseManager", 7, "Check said false because expirement date " + expireDate.toString() + " is in the past.");
				throw Exception(QObject::trUtf8("Ablaufdatum in der Vergangenheit"), QObject::trUtf8("Das in der Lizenz verschluesselte Ablaufdatum liegt in der Vergangenheit"));
			}

			key->setAblaufDatum(expireDate);
		} else {
			throw Exception(QObject::trUtf8("Ablaufdatum ungueltig"), QObject::trUtf8("Das im Lizensierungsschluessel kodierte Ablaufdatum ist ungueltig."));
		}

		QDate expireWettbewerbDate = getExpireWettbewerbDate(key);
		context->setExpireDateWettbewerbsModul(expireWettbewerbDate);

		RSAManager *rsaManager = new DummyRSAManager();
		PublicRSAKey *publicKey = rsaManager->loadPublicKey(key->getPublicKey());

		QString licenseKey = key->getLicenseKey();

		QString decryptedString = rsaManager->decryptFromHex(publicKey, licenseKey);
		QString decryptedLicenseKeyOldFashion = decryptedString.right(decryptedString.length() - 10);
		QString decryptedLicenseKeyNewFashion = decryptedString.right(decryptedString.length() - 20);
		QString hardwareString = getHardwareKey(context);

// Only if debugging is needed
//		context->log("License", "LicenseManager", "decryptedOldFashion: " + decryptedLicenseKeyOldFashion);
//		context->log("License", "LicenseManager", "decryptedNewFashion: " + decryptedLicenseKeyNewFashion);
//		context->log("License", "LicenseManager", "hardwareString: " + hardwareString);

		bool result = decryptedLicenseKeyOldFashion == hardwareString || decryptedLicenseKeyNewFashion == hardwareString;
		if (result) {
			context->log("License", "LicenseManager", 7, "Check said true.");
		} else {
			context->log("License", "LicenseManager", 7, "Check said false because keys did not match.");
			throw Exception(QObject::trUtf8("Lizenz ungueltig"), QObject::trUtf8("Die Lizenz ist fuer diesen Rechner nicht gueltig."));
		}
	} else if (!key->hasPublicKey()) {
		context->log("License", "LicenseManager", 7, "Check said false because there is no public key");
		throw Exception(QObject::trUtf8("Keine Lizenz vorhanden"), QObject::trUtf8("Der zum entschluesseln benoetigte Schluessel ist nicht registriert."));
	} else if (!key->hasLicenseKey()) {
		context->log("License", "LicenseManager", 7, "Check said false because there is no license key.");
		throw Exception(QObject::trUtf8("Keine Lizenz vorhanden"), QObject::trUtf8("Es ist keine Lizenz vorhanden."));

	}
}

void LicenseManager::checkLicenseKey(DatabaseContext *context) {
	LicenseKey *licenseKey = loadLicenseKey(context);
	if (licenseKey == NULL) {
		throw Exception(QObject::trUtf8("Keine Lizenz vorhanden"), QObject::trUtf8("Es ist keine Lizenz vorhanden."));
	} else {
		checkLicenseKey(context, licenseKey);
	}
}
