
#include "EventLogger.h"

#include <vector>

#include "EventLog.h"
#include "EventLogCategoryController.h"
#include "EventLogController.h"
#include "EventLogSettingsController.h"

EventLogger::EventLogger(DatabaseContext *context) {
	this->context = context;
	this->enabled = false;

	categoryNameToCategory = std::map<QString, EventLogCategory*>();
	settingsIdToSettings = std::map<qlonglong, EventLogSettings*>();
	categoryIdToSettings = std::multimap<qlonglong, EventLogSettings*>();
	locationToSettings = std::multimap<QString, EventLogSettings*>();
	withoutCategoryAndLocation = std::vector<EventLogSettings*>();
}

EventLogger::~EventLogger() {
	deleteConfigObjects();
}

void EventLogger::disable() {
	enabled = false;
}

void EventLogger::reloadConfig(DatabaseContext *context) {
	context->log("Log", "EventLogger", 7, "Starting to reload EventLog-configuration...");

	// clean up old data, if config was already loaded before
	deleteConfigObjects();
	categoryNameToCategory = std::map<QString, EventLogCategory*>();
	settingsIdToSettings = std::map<qlonglong, EventLogSettings*>();
	categoryIdToSettings = std::multimap<qlonglong, EventLogSettings*>();
	locationToSettings = std::multimap<QString, EventLogSettings*>();
	withoutCategoryAndLocation = std::vector<EventLogSettings*>();

	// load all categories
	EventLogCategoryController categoryController = EventLogCategoryController();
	std::vector<EventLogCategory*> categories = categoryController.get(context, std::map<QString, QVariant>());
	for (std::vector<EventLogCategory*>::const_iterator it = categories.begin(); it != categories.end(); it++) {
		EventLogCategory *category = *it;
		categoryNameToCategory[category->getName()] = category;
	}

	// load all settings
	EventLogSettingsController settingsController = EventLogSettingsController();
	std::vector<EventLogSettings*> settings = settingsController.get(context, std::map<QString, QVariant>());
	for (std::vector<EventLogSettings*>::const_iterator it = settings.begin(); it != settings.end(); it++) {
		EventLogSettings *settings = *it;

		// store settings id
		qlonglong settingsId = settings->getId();
		settingsIdToSettings[settingsId] = settings;

		// map to category if present
		if (settings->hasCategory()) {
			qlonglong categoryId = settings->getCategory();
			categoryIdToSettings.insert(std::multimap<qlonglong, EventLogSettings*>::value_type(categoryId, settings));
		}

		// map to location if present
		if (settings->hasLocation()) {
			QString location = settings->getLocation();
			locationToSettings.insert(std::multimap<QString, EventLogSettings*>::value_type(location, settings));
		}

		// record if neither category nor location is present
		if (!settings->hasCategory() && !settings->hasLocation()) {
			withoutCategoryAndLocation.push_back(settings);
		}
	}

	this->enabled = true;
	context->log("Log", "EventLogger", 7, "Finished reloading EventLog-configuration, logger now enabled.");
}

void EventLogger::log(QString categoryName, QString location, int level, QString text) {
	if (!enabled) {
		return;
	}

	// The database column we log to only has 255 characters.  Thus, we have to split longer messages.
	int z = 1;
	if (text.length() > 255) {
		for (int n = 0; n < text.length(); n += 225) {
			QString currText = text.mid(n, 225);
			currText = "Multi-Line Msg, Part " + QString::number(z) + ": " + currText;
			log(categoryName, location, level, currText);
			z++;
		}
		return;
	}

	bool doOutputToDB = false;
	bool doOutputToStdOut = false;

	EventLogCategory *category = NULL;
	if (!categoryName.isNull() && categoryNameToCategory.find(categoryName) != categoryNameToCategory.end()) {
		category = categoryNameToCategory[categoryName];
	}

	// use the multimaps for pre-filtering which settings objects might be interesting,
	// and then decide using the detail information.
	if (category != NULL) {
		qlonglong categoryId = category->getId();
		for (std::multimap<qlonglong, EventLogSettings*>::iterator it = categoryIdToSettings.find(categoryId);
		  it != categoryIdToSettings.end() && it != categoryIdToSettings.upper_bound(categoryId); it++) {
			EventLogSettings *settings = it->second;
			if (output(settings, category, location, level)) {
				doOutputToDB |= outputToDB(settings);
				doOutputToStdOut |= outputToStdOut(settings);
			}
		}
	}

	if (!location.isNull()) {
		for (std::multimap<QString, EventLogSettings*>::const_iterator it = locationToSettings.find(location);
			it != locationToSettings.end() && it != locationToSettings.upper_bound(location); it++) {
			EventLogSettings *settings = it->second;
			if (output(settings, category, location, level)) {
				doOutputToDB |= outputToDB(settings);
				doOutputToStdOut |= outputToStdOut(settings);
			}
		}
	}

	for (std::vector<EventLogSettings*>::const_iterator it = withoutCategoryAndLocation.begin();
		it != withoutCategoryAndLocation.end(); it++) {
		EventLogSettings *settings = *it;
		if (output(settings, category, location, level)) {
			doOutputToDB |= outputToDB(settings);
			doOutputToStdOut |= outputToStdOut(settings);
		}
	}

	//std::cout << "doOutputToDB: " << doOutputToDB << ", doOutputToStdOut: " << doOutputToStdOut << std::endl;

	if (doOutputToDB) {
		try {
			EventLogController logController = EventLogController();
			EventLog *eventLog = new EventLog();
			if (category == NULL) {
				eventLog->deleteCategory();
			} else {
				eventLog->setCategory(category->getId());
			}
			if (location.isNull()) {
				eventLog->deleteLocation();
			} else {
				eventLog->setLocation(location);
			}
			eventLog->setExecution_tist(QDateTime::currentDateTime());
			eventLog->setLevel(level);
			eventLog->setText(text);

			Session *session = context->getSession();
			if (session == NULL) {
				eventLog->deleteSession();
			} else {
				eventLog->setSession(session->getId());
			}

			logController.insertOrUpdate(context, eventLog);

			delete eventLog;
		} catch (PersistenceException e) {
			std::cout << "Unable to write the log.  Reason: " << qPrintable(e.getShortText()) << "; Details: " << qPrintable(e.getLongText()) << std::endl;
			// Do nothing here except writing a message to stdout.  Being unable to write the log shouldn´t be a reason for 
			//crashing the application, but we can´t do anything actually senseful here either.
		}
	}
	if (doOutputToStdOut) {
		if (category != NULL) {
			std::cout << "[" << qPrintable(categoryName) << "]";
		} else {
			std::cout << "[NoCategory]";
		}
		std::cout << "[" << level << "] ";
		if (!location.isNull()) {
			std::cout << qPrintable(location) << " says: ";
		} else {
			std::cout << "Some module says: ";
		}
		std::cout << qPrintable(text) << std::endl;
	}
}

bool EventLogger::output(EventLogSettings* settings, EventLogCategory *category, QString location, int level) {
	//std::cout << "Starting output." << std::endl;

	// if none of the criteria objects, we log
	bool output = true;
	if (category != NULL && settings->hasCategory()) {
		// If either as parameter for the log message or in the settings object,
		// no category name is given, we don't filter for a category.
		qlonglong categoryId = category->getId();
		//std::cout << "Inspecting categoryId " << categoryId << std::endl;
		if (categoryId != settings->getCategory()) {
			// If a category is given, but does not match the one in the settings object,
			// this settings object does not allow as to log
			//std::cout << "Setting output to false since Settings-Category is " << settings->getCategory() << std::endl;
			output = false;
		}
	}

	if (!location.isNull() && settings->hasLocation() && settings->getLocation().trimmed().length() > 0) {
		// If either as parameter of the log message, or in the settings object,
		// no location is given, we don't filter for a location
		//std::cout << "Inspecting location " << qPrintable(location) << std::endl;
		if (location != settings->getLocation()) {
		    //std::cout << "Setting output to false since Settings-Location is " 
			//		  << qPrintable(settings->getLocation()) << std::endl;
			output = false;
		}
	}

	if (settings->hasLevel()) {
		// only filter for level, if in the settings object a level is given
		int settingsLevel = settings->getLevel();

		// if no level operand is given, assume >=
		QString settingsOperand = (settings->hasLevelOperand() ? settings->getLevelOperand() : ">=");
		//std::cout << "Inspecting level " << settingsLevel << " and settingsOperand " << qPrintable(settingsOperand) 
		//		  << std::endl;
		if (settingsOperand == ">=" && !(level >= settingsLevel)) {
			//std::cout << "Case >= - Setting output to false since settingsLevel is " << settingsLevel << std::endl;
			output = false;
		} else if (settingsOperand == "<=" && !(level <= settingsLevel)) {
			//std::cout << "Case <= - Setting output to false since settingsLevel is " << settingsLevel << std::endl;
			output = false;
		} else if (settingsOperand == "=" && !(level == settingsLevel)) {
			//std::cout << "Case = - Setting output to false since settingsLevel is " << settingsLevel << std::endl;
			output = false;
		} else if (settingsOperand == ">" && !(level > settingsLevel)) {
			//std::cout << "Case > - Setting output to false since settingsLevel is " << settingsLevel << std::endl;
			output = false;
		} else if (settingsOperand == "<" && !(level < settingsLevel)) {
			//std::cout << "Case < - Setting output to false since settingsLevel is " << settingsLevel << std::endl;
			output = false;
		}
	}
	return output;
}

bool EventLogger::outputToDB(EventLogSettings* settings) {
	return settings->hasToDB() && settings->getToDB();
}

bool EventLogger::outputToStdOut(EventLogSettings *settings) {
	return settings->hasToStdOut() && settings->getToStdOut();
}

void EventLogger::deleteConfigObjects() {
	for (std::map<QString, EventLogCategory*>::const_iterator it = categoryNameToCategory.begin();
		it != categoryNameToCategory.end(); it++) {
		delete it->second;
	}
	for (std::map<qlonglong, EventLogSettings*>::const_iterator it = settingsIdToSettings.begin();
		it != settingsIdToSettings.end(); it++) {
		delete it->second;
	}
	// categoryIdToSettings, locationToSettings contain subsets of settingsIdToSettings, 
	// so we don't need to delete anything there
}
