#include "QueryLogger.h"

#include "QueryLog.h"
#include "QueryLogController.h"
#include "QueryLogSettingsController.h"

#include <QStringList>

QueryLogger::QueryLogger() {
	std::map<qlonglong, QueryLogSettings*> settingsIdToSettings;
	std::multimap<QString, QueryLogSettings*> entityCodeNameToSettings;
	std::vector<QueryLogSettings*> settingsWithoutEntity;
	this->context = NULL;
	this->enabled = false;
}

QueryLogger::~QueryLogger() {
	deleteConfigObjects();
}

void QueryLogger::disable() {
	this->enabled = false;
}

void QueryLogger::initialize(DatabaseContext *context) {
	context->log("Log", "QueryLogger", 7, "Starting to reload QueryLog-configuration...");

	this->context = context;

	// clean up old data, if config was already loaded before
	deleteConfigObjects();
	settingsIdToSettings = std::map<qlonglong, QueryLogSettings*>();
	entityCodeNameToSettings = std::multimap<QString, QueryLogSettings*>();
	settingsWithoutEntity = std::vector<QueryLogSettings*>();

	// load all settings
	QueryLogSettingsController settingsController = QueryLogSettingsController();
	std::vector<QueryLogSettings*> settings = settingsController.get(context, std::map<QString, QVariant>());
	for (std::vector<QueryLogSettings*>::const_iterator it = settings.begin(); it != settings.end(); it++) {
		QueryLogSettings *settings = *it;

//		std::cout << "Read settings object for output to db: " << settings->getToDB() << std::endl;
//		std::cout << "Read settings object for output to stdout:" << settings->getToStdOut() << std::endl;

		// store settings id
		qlonglong settingsId = settings->getId();
		settingsIdToSettings[settingsId] = settings;

		// map to entity if present
		if (settings->hasEntity()) {
			// TODO: unknown entities ignored for now
			QString entityCodeName = settings->getEntity();
			entityCodeNameToSettings.insert(std::multimap<QString, QueryLogSettings*>
											::value_type(entityCodeName, settings));
		} else {
			settingsWithoutEntity.push_back(settings);
		}
	}

	enabled = true;
	context->log("Log", "QueryLogger", 7, "Finished reloading QueryLog-configuration, logger now enabled.");
}

void QueryLogger::deleteConfigObjects() {
	// TODO: Find out wether this code is ok - according to STL documentation,
	// clear also destroys the objects.  If this is true, how do the other
	// containers behave afterwards (settinsIdToSettings contains everything)?
	settingsIdToSettings.clear();
	entityCodeNameToSettings.clear();
	settingsWithoutEntity.clear();

	for (std::map<qlonglong, QueryLogSettings*>::const_iterator it = settingsIdToSettings.begin();
		it != settingsIdToSettings.end(); it++) {
		delete it->second;
	}
}

void QueryLogger::log(QString location, int numberOfAffectedRows, QString queryType,
					  QString info, QString queryText, QString entities, bool dbLoggingEnabled) {
	if (!enabled) {
		// do nothing, call initialize first
		return;
	}

	CodeManager *codeManager = context->getCodeManager();
	CodeValue *queryTypeCodeValue = codeManager->getCodeValueByTypeAndName("QueryType", queryType);
	log(location, numberOfAffectedRows, queryTypeCodeValue, info, queryText, entities, dbLoggingEnabled);
	delete queryTypeCodeValue;
}

void QueryLogger::log(QString location, int numberOfAffectedRows, CodeValue* queryType,
					  QString info, QString queryText, QString entities, bool dbLoggingEnabled) {

	if (!enabled) {
		// do nothing, call initialize first
		return;
	}

	CodeManager *codeManager = context->getCodeManager();
	bool doOutputToDB = false;
	bool doOutputToStdOut = false;

	// first check settings without entitiy
	for (std::vector<QueryLogSettings*>::const_iterator it = settingsWithoutEntity.begin();
		  it != settingsWithoutEntity.end(); it++) {
		QueryLogSettings *settings = *it;

		// No Entitiy given, so we are content with any Entity.
		// Thus the only thing we have to check here is the QueryType.
		if (queryTypeMatches(codeManager, settings, queryType)) {
			doOutputToDB |= outputToDB(settings);
			doOutputToStdOut |= outputToStdOut(settings);
		}
	}

//	std::cout << "Step 1: " << doOutputToDB << "/" << doOutputToStdOut << std::endl;

	QStringList entityList = entities.split(",");
	for (int n = 0; n < entityList.size(); n++) {
		QString entityCodeName = entityList.at(n);
		for (std::multimap<QString, QueryLogSettings*>::const_iterator it = entityCodeNameToSettings.find(entityCodeName);
			it != entityCodeNameToSettings.end() && it != entityCodeNameToSettings.upper_bound(entityCodeName); it++) {
			QueryLogSettings *settings = it->second;

			if (queryTypeMatches(codeManager, settings, queryType)) {
				doOutputToDB |= outputToDB(settings);
				doOutputToStdOut |= outputToStdOut(settings);
			}
		}
	}

//	std::cout << "Step 2: " << doOutputToDB << "/" << doOutputToStdOut << std::endl;

	// We may not log entries for the log table QueryLog in the QueryLog.
	// This would result in an endless loop.
	if (dbLoggingEnabled && doOutputToDB && !entities.contains("QueryLog")) {
		try {
			QueryLogController logController = QueryLogController();
			QueryLog *queryLog = new QueryLog();

			// TODO: Timestamp from DB would be better...
			queryLog->setExecution_tist(QDateTime::currentDateTime());
			queryLog->setNumberOfRows(numberOfAffectedRows);
			queryLog->setInfo(info);
			queryLog->setQuery(queryText);
			queryLog->setQueryType(queryType->getId());
			queryLog->setEntities(entities);
			queryLog->setLocation(location);

			Session *session = context->getSession();
			if (session == NULL) {
				queryLog->deleteSession();
			} else {
				queryLog->setSession(session->getId());
			}

			logController.insertOrUpdate(context, queryLog);
		} catch (PersistenceException e) {
			std::cout << "Unable to write the log.  Reason: " << qPrintable(e.getShortText()) << "; Details: " << qPrintable(e.getLongText()) << std::endl;
			std::cout << "Info: queryText was: " << qPrintable(queryText) << std::endl;
			// Do nothing here except writing a message to stdout.  Being unable to write the log shouldn´t be a reason for 
			//crashing the application, but we can´t do anything actually senseful here either.
		}
	}

//	std::cout << "Step 3: " << doOutputToDB << "/" << doOutputToStdOut << std::endl;

	if (doOutputToStdOut) {
		if (!location.isNull()) {
			std::cout << qPrintable(location) << " executed query: ";
		} else {
			std::cout << "Some module executed query: ";
		}
		QString queryOutputText = queryText.replace("\\\"", "\"");
		std::cout << qPrintable(queryOutputText) << " ";
		if (!info.isNull() && info.length() > 0) {
			std::cout << "(info: " << qPrintable(info) << ") ";
		}
		std::cout << "--- Affected rows: " << numberOfAffectedRows << std::endl;
	}
}

bool QueryLogger::queryTypeMatches(CodeManager *codeManager, QueryLogSettings *settings, CodeValue *queryType) {
	// In detail:
	// ALL in the QueryLogSettings matches anything
	// READ matches SELECT
	// WRITE matches INSERT, UPDATE, DELETE
	// SELECT, INSERT, UPDATE, DELETE match themself

	if (!settings->hasQueryType()) {
		return true;
	} else {
		return (queryType->getName() == "SELECT" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "ALL")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "READ")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "SELECT")))
			|| (queryType->getName() == "INSERT" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "ALL")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "WRITE")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "INSERT")))
			|| (queryType->getName() == "UPDATE" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "ALL")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "WRITE")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "UPDATE")))
			|| (queryType->getName() == "DELETE" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "ALL")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "WRITE")
				  || codeManager->isCodeValue(settings->getQueryType(), "QueryType", "DELETE")))
			|| (queryType->getName() == "USER" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "USER")))
			|| (queryType->getName() == "GRANT" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "GRANT")))
			|| (queryType->getName() == "ATTRIBUTEPATH" &&
				  (  codeManager->isCodeValue(settings->getQueryType(), "QueryType", "ATTRIBUTEPATH")));
	}
}

bool QueryLogger::outputToDB(QueryLogSettings* settings) {
	return settings->hasToDB() && settings->getToDB();
}

bool QueryLogger::outputToStdOut(QueryLogSettings *settings) {
	return settings->hasToStdOut() && settings->getToStdOut();
}
